package it.gotoandplay.smartfoxserver;

import it.gotoandplay.smartfoxserver.config.ConfigData;
import it.gotoandplay.smartfoxserver.data.User;
import it.gotoandplay.smartfoxserver.lib.ChannelQueue;
import it.gotoandplay.smartfoxserver.lib.ChannelWrap;
import it.gotoandplay.smartfoxserver.lib.OutgoingServerEvent;
import it.gotoandplay.smartfoxserver.lib.ServerEvent;
import java.io.IOException;
import java.net.InetAddress;
import java.net.Socket;
import java.nio.ByteBuffer;
import java.nio.channels.SelectionKey;
import java.nio.channels.SocketChannel;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.logging.Logger;

public class EventWriter extends ChannelWrap
{
  private int dropCount = 0;
  private int dropThreshold = 300;
  private static final int MAX_EVENT_SIZE = 32768;
  private static final char EOM = 0;
  private static SmartFoxServer sf;
  private List failedChannels;
  private List droppableChannels;
  static Logger log = SmartFoxServer.log;

  public EventWriter(SmartFoxServer sf, int nThreads)
  {
    sf = sf;

    this.failedChannels = new LinkedList();
    this.droppableChannels = new ArrayList();

    initWrap(nThreads);
  }

  public void run()
  {
    ByteBuffer writeBuffer = ByteBuffer.allocateDirect(32768);

    this.running = true;

    label58: while (this.running)
    {
      try
      {
        SocketChannel ch;
        if ((ch = this.eventQueue.deQueue()) == null)
          break label58;

        processEvent(ch, writeBuffer);
      }
      catch (InterruptedException e)
      {
        if (SmartFoxServer.IS_SHUTTING_DOWN) break label58;
        log.warning("Problems during event processing");
      }
      catch (Throwable t)
      {
        t.printStackTrace();
      }
    }
  }

  public void addNewTask(SocketChannel ch)
  {
    if (ch != null)
      this.eventQueue.enQueue(ch);
  }

  public void handleEvent(ServerEvent event)
  {
    LinkedList ll = event.getRecipients();

    if (ll.size() > 0)
    {
      String msg = event.getStrMsg();
      SocketChannel sender = event.getSenderChannel();

      for (Iterator i = ll.iterator(); i.hasNext(); )
      {
        SocketChannel chan = (SocketChannel)i.next();
        LinkedList q = SmartFoxServer.getInstance().getChannelQueue(chan);

        if (q == null)
          break label333;
        synchronized (q)
        {
          boolean wasEmpty = q.size() == 0;

          if (q.size() < ConfigData.MAX_CHANNEL_QUEUE)
          {
            q.addLast(new OutgoingServerEvent(msg, sender));

            label333: if (wasEmpty) {
              this.eventQueue.enQueue(chan);
            }

          }
          else
          {
            ConfigData.outGoingDroppedMessages += 1L;

            User u = SmartFoxServer.getInstance().getUserByChannel(chan);

            if ((u != null) && (ConfigData.MAX_DROPPED_PACKETS > -1))
            {
              if (!(u.isMarkedForRemoval())) {
                u.increaseDroppedPackets();
              }

              if ((u.getDroppedPackets() > ConfigData.MAX_DROPPED_PACKETS) && (!(u.isMarkedForRemoval())))
              {
                synchronized (this.failedChannels)
                {
                  this.droppableChannels.add(chan);
                }
                u.setMarkedForRemoval(true);

                log.info("Too many dropped messages. Client will be disconnected: " + u.getName() + ", " + u.getDroppedPackets());
              }
            }

            if (ConfigData.debugInfo)
            {
              String ipadd = (chan.isOpen()) ? chan.socket().getInetAddress().getHostAddress() : "Not available";
              SmartFoxServer.log.fine("Dropped outgoing message. User: " + u + ", IP: " + ipadd);
            }
          }
        }
      }

    }

    disconnectDroppableChannels();
  }

  public void handleEvent(ServerEvent event, boolean b)
  {
  }

  protected void processEvent(SocketChannel ch)
  {
  }

  protected void processEvent(SocketChannel ch, ByteBuffer writeBuffer)
  {
    LinkedList q;
    try
    {
      q = sf.getChannelQueue(ch);

      if ((q == null) || 
        (q.size() <= 0)) {
        return;
      }

      synchronized (q)
      {
        OutgoingServerEvent ose = (OutgoingServerEvent)q.getFirst();

        writeBuffer.clear();

        if (ose.bb == null)
        {
          String msg = ose.getStrMsg() + ';
          byte[] msgBytes = msg.getBytes();

          if (writeBuffer.capacity() < msgBytes.length)
          {
            SmartFoxServer.log.fine("Allocating new write buffer, size >> " + msgBytes.length);
            writeBuffer = ByteBuffer.allocateDirect(msgBytes.length);
          }

          writeBuffer.put(msgBytes);
        }
        else
        {
          if (writeBuffer.capacity() < ose.bb.length)
          {
            SmartFoxServer.log.fine("Allocating new bufferSize (pending op.) to >> " + ose.bb.length);
            writeBuffer = ByteBuffer.allocateDirect(ose.bb.length);
          }

          writeBuffer.put(ose.bb);
        }

        writeBuffer.flip();

        long toWrite = writeBuffer.remaining();
        try
        {
          long bytesWritten = ch.write(writeBuffer);

          ConfigData.dataOUT += bytesWritten;

          if (bytesWritten < toWrite)
          {
            byte[] bb = new byte[writeBuffer.remaining()];
            writeBuffer.get(bb);
            writeBuffer.rewind();

            ose.bb = bb;

            SelectionKey sk = SmartFoxServer.getInstance().getChannelKey(ch);

            if (sk == null)
              break label386;

            sk.interestOps(5); break label386:
          }

          q.removeFirst();

          if (q.size() <= 0) break label386;
          label386: this.eventQueue.enQueue(ch);
        }
        catch (IOException ioe)
        {
          log.fine("failed writing to channel: " + ch);

          if (ConfigData.DEAD_CHANNELS_POLICY <= 0)
            break label386;
          synchronized (this.failedChannels)
          {
            this.failedChannels.add(ch);
          }

        }
        catch (Exception e)
        {
          log.warning("Generic exception during write: " + e);
        }

      }

    }
    catch (Exception e)
    {
      log.warning("Generic Exception during write operations: " + e);
      e.printStackTrace();
    }
  }

  public void cleanFailedChannels()
  {
    dropListOfClients(this.failedChannels);
  }

  private void disconnectDroppableChannels()
  {
    dropListOfClients(this.droppableChannels);
  }

  private void dropListOfClients(List listOfClients)
  {
    if (listOfClients.size() > 0)
    {
      int i;
      Object[] chArray = (Object[])null;
      SocketChannel sc = null;

      synchronized (listOfClients)
      {
        chArray = listOfClients.toArray();
        listOfClients.clear();

        log.fine("Disconnecting " + chArray.length + " dead channels.");
      }

      for (int i = 0; i < chArray.length; ++i)
      {
        sf.lostConnection((SocketChannel)chArray[i]);
      }
    }
  }
}