package it.gotoandplay.smartfoxserver;

import it.gotoandplay.smartfoxserver.config.ConfigData;
import it.gotoandplay.smartfoxserver.controllers.BlueBoxHandler;
import it.gotoandplay.smartfoxserver.controllers.ExtensionHandler;
import it.gotoandplay.smartfoxserver.controllers.SystemHandler;
import it.gotoandplay.smartfoxserver.data.BannedUser;
import it.gotoandplay.smartfoxserver.data.KickedUser;
import it.gotoandplay.smartfoxserver.data.Moderator;
import it.gotoandplay.smartfoxserver.data.ModeratorManager;
import it.gotoandplay.smartfoxserver.data.Room;
import it.gotoandplay.smartfoxserver.data.User;
import it.gotoandplay.smartfoxserver.data.Zone;
import it.gotoandplay.smartfoxserver.data.buddylist.BuddyListManager;
import it.gotoandplay.smartfoxserver.data.buddylist.persistence.AbstractBuddyPersister;
import it.gotoandplay.smartfoxserver.data.buddylist.persistence.DefaultBuddyListPersister;
import it.gotoandplay.smartfoxserver.data.buddylist.persistence.IBuddyListPersister;
import it.gotoandplay.smartfoxserver.data.buddylist.tasks.BuddyListTaskHandler;
import it.gotoandplay.smartfoxserver.data.buddylist.tasks.BuddyTasks;
import it.gotoandplay.smartfoxserver.db.DbManager;
import it.gotoandplay.smartfoxserver.exceptions.ConfigurationException;
import it.gotoandplay.smartfoxserver.exceptions.CreateRoomException;
import it.gotoandplay.smartfoxserver.exceptions.DbManagerException;
import it.gotoandplay.smartfoxserver.exceptions.LoginException;
import it.gotoandplay.smartfoxserver.extensions.ExtensionHelper;
import it.gotoandplay.smartfoxserver.extensions.ExtensionItem;
import it.gotoandplay.smartfoxserver.lib.BadWordsFilter;
import it.gotoandplay.smartfoxserver.lib.ConfigReader;
import it.gotoandplay.smartfoxserver.lib.ConnectionCleanerTask;
import it.gotoandplay.smartfoxserver.lib.ExtensionAutoReloader;
import it.gotoandplay.smartfoxserver.lib.Password;
import it.gotoandplay.smartfoxserver.lib.ServerEvent;
import it.gotoandplay.smartfoxserver.lib.ShutdownManager;
import it.gotoandplay.smartfoxserver.lib.SmartFoxLib;
import it.gotoandplay.smartfoxserver.py.IPyClass;
import it.gotoandplay.smartfoxserver.py.PyHelper;
import it.gotoandplay.smartfoxserver.util.Attachment;
import it.gotoandplay.smartfoxserver.util.SmartFoxShortLogFormatter;
import it.gotoandplay.smartfoxserver.util.SmartFoxSimpleLogFormatter;
import it.gotoandplay.smartfoxserver.util.scheduling.ITaskHandler;
import it.gotoandplay.smartfoxserver.util.scheduling.Scheduler;
import it.gotoandplay.smartfoxserver.util.scheduling.Task;
import it.gotoandplay.smartfoxserver.webserver.WebServer;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.io.PrintStream;
import java.net.InetAddress;
import java.net.InetSocketAddress;
import java.net.NetworkInterface;
import java.net.ServerSocket;
import java.net.Socket;
import java.net.SocketException;
import java.nio.ByteBuffer;
import java.nio.CharBuffer;
import java.nio.channels.ClosedChannelException;
import java.nio.channels.SelectionKey;
import java.nio.channels.Selector;
import java.nio.channels.ServerSocketChannel;
import java.nio.channels.SocketChannel;
import java.nio.charset.Charset;
import java.nio.charset.CharsetDecoder;
import java.nio.charset.MalformedInputException;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.Enumeration;
import java.util.HashMap;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.Timer;
import java.util.logging.ConsoleHandler;
import java.util.logging.FileHandler;
import java.util.logging.Logger;
import net.n3.nanoxml.IXMLElement;
import net.n3.nanoxml.IXMLParser;
import net.n3.nanoxml.IXMLReader;
import net.n3.nanoxml.StdXMLReader;
import net.n3.nanoxml.XMLException;
import net.n3.nanoxml.XMLParserFactory;
import org.h2.tools.Server;
import org.json.JSONObject;

public class SmartFoxServer extends Thread
{
  public static boolean IS_SHUTTING_DOWN = false;
  private static int $_TYP = -1;
  private static String $_CLI = null;
  private static List $_ADD = null;
  private static List $_BB = null;
  private static int $_MAX = -1;
  private static SmartFoxServer instance;
  private static final String currVersion = "1.6.2";
  private static String ADMIN_LOGIN;
  private static String ADMIN_PASSWORD;
  private long serverStartTime;
  private ServerSocketChannel sSockChan;
  private Selector acceptSelector;
  private Selector readSelector;
  private SelectionKey selectKey;
  private SelectionKey acceptKey;
  boolean running;
  boolean cfg;
  private long guestCount;
  public List ALLOWED_ADMIN_ADDRESSES;
  public boolean isAdminRestricted;
  private LinkedList clients;
  private LinkedList newClients;
  private LinkedList bannedUsers;
  private HashMap kickedUsers;
  private CharsetDecoder asciiDecoder;
  private EventReader eventReader;
  private EventWriter eventWriter;
  private ConfigReader configReader;
  private Timer connectionCleanerTimer;
  public static Logger log = Logger.getLogger("it.gotoandplay.smartfoxserver.SmartFoxServer");
  private static FileHandler fh;
  private static ConsoleHandler ch;
  private HashMap eventControllers;
  private HashMap zones;
  private HashMap usersByChannel;
  private HashMap usersById;
  private HashMap ChannelQueues;
  private HashMap SelectionKeysMap;
  private Map extensions;
  private SystemHandler sysHandler;
  private ExtensionHandler xtHandler;
  private BlueBoxHandler bbHandler;
  private HashMap ipFilter;
  IXMLParser xmlParser;
  IXMLReader xmlReader;
  private String crossDomainPolicy;
  ShutdownManager shutDownManager;
  private Thread extAutoReloader;
  private ExtensionAutoReloader autoReloader;
  private Scheduler taskScheduler;
  private PyHelper pyHelper;
  private String rinfo = null;
  private Server h2Engine;
  private DbManager h2Connection;
  private StringBuffer bbc;
  private int bbc_ref;

  public static void main(String[] args)
  {
    if (args.length == 1)
    {
      ConfigReader.xmlFile = args[0];
    }

    instance = new SmartFoxServer();
    getInstance().start();
  }

  public SmartFoxServer()
  {
    super("SmartFoxServer");
    this.running = false;
    this.isAdminRestricted = true;

    ConfigData.apiMinVersion = 130;

    this.clients = new LinkedList();
    this.newClients = new LinkedList();

    this.eventControllers = new HashMap();
    this.usersByChannel = new HashMap();
    this.usersById = new HashMap();
    this.zones = new HashMap();

    this.ChannelQueues = new HashMap();
    this.SelectionKeysMap = new HashMap();

    this.bannedUsers = loadBannedUserList();

    this.kickedUsers = new HashMap();
    this.extensions = new HashMap();
    this.ipFilter = new HashMap();

    this.guestCount = 0L;

    ConfigData.dataIN = 0L;
    ConfigData.dataOUT = 0L;
    ConfigData.inComingDroppedMessages = 0L;
    ConfigData.outGoingDroppedMessages = 0L;
    ConfigData.maxSimultanousConnections = 0;

    this.extAutoReloader = null;

    this.asciiDecoder = Charset.forName("UTF-8").newDecoder();

    if (ConfigData.restartCount == 0)
    {
      this.shutDownManager = new ShutdownManager();
      Runtime.getRuntime().addShutdownHook(this.shutDownManager);
    }

    this.configReader = ConfigReader.getInstance();

    this.rinfo = Password.getInstance().decodePassword("1E324DE657963702112756F06E7267169EC6562A6FD637AB3011F9F0522B96464CA2496B00E0FCA81F190A299D0F4DFA42AC1D81E1");

    this.pyHelper = new PyHelper();
    this.pyHelper.init(null);
  }

  private void welcomeMessage()
  {
    System.out.println("|::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::|");
    System.out.println("|                                                            |");
    System.out.println("|              ...:::  SmartFoxServer :::...                 |");
    System.out.println("|                Multiplayer Socket Server                   |");
    System.out.println("|                      version 1.6.2                         |");
    System.out.println("|                           ---                              |");
    System.out.println("|              (c) 2004 - 2007 gotoAndPlay()                 |");
    System.out.println("|                  www.smartfoxserver.com                    |");
    System.out.println("|                    www.gotoandplay.it                      |");
    System.out.println("|                                                            |");
    System.out.println("|::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::|\n");
  }

  private void initServerSocket()
  {
    showSystemInfo();
    try
    {
      this.configReader.runHelper();

      System.out.println("\n\n--- [ Licence Loaded ] ------------------------------------\n");

      this.pyHelper.getPyClass("taskGenHelper").call("scanByTag", new Object[0]);

      this.pyHelper.getPyClass("taskGenHelper").call("scanByObj", new Object[0]);

      this.cfg = this.configReader.readZoneConfig();
    }
    catch (ConfigurationException ce)
    {
      System.out.println("Problems in config file: " + ce.getMessage());
      ce.printStackTrace();
      dieOnConfig();
    }
    catch (DbManagerException dme)
    {
      System.out.println("\tDbManager problems: " + dme.getMessage());
    }
    catch (CreateRoomException cre)
    {
      System.out.println("Could not create room on server boot!");
    }
    catch (Exception e)
    {
      System.out.println("Errors in configuration file! " + e);
      dieOnConfig();
    }

    if (!(this.cfg))
    {
      dieOnConfig();
    }

    try
    {
      this.xmlParser = XMLParserFactory.createDefaultXMLParser();
    }
    catch (Exception e)
    {
      log.severe("Can't instantiate XML Parser! Server quits");
      System.exit(1);
    }

    try
    {
      this.sSockChan = ServerSocketChannel.open();
      this.sSockChan.configureBlocking(false);

      InetAddress addr = InetAddress.getByName(ConfigData.SERVER_ADDR);

      this.sSockChan.socket().bind(new InetSocketAddress(ConfigData.SERVER_PORT));
      this.sSockChan.socket().setReuseAddress(true);

      this.readSelector = Selector.open();
      this.acceptSelector = Selector.open();

      this.acceptKey = this.sSockChan.register(this.acceptSelector, 16);

      System.out.println("\n\n--- [ Server Starting ] ---------------------------------\n");
      System.out.println("Server address: " + addr.getHostAddress());
      System.out.println("Server port   : " + ConfigData.SERVER_PORT + "\n");

      if (!(ConfigData.EXTERNAL_CROSSDOAMIN_POLICY))
      {
        if ((ConfigData.POLICY_ALLOWED_DOMAINS != null) && (ConfigData.POLICY_ALLOWED_DOMAINS.size() > 0))
        {
          this.crossDomainPolicy = "<cross-domain-policy>";

          for (int i = 0; i < ConfigData.POLICY_ALLOWED_DOMAINS.size(); ++i) {
            SmartFoxServer tmp408_407 = this; tmp408_407.crossDomainPolicy = tmp408_407.crossDomainPolicy + "<allow-access-from domain='" + ((String)ConfigData.POLICY_ALLOWED_DOMAINS.get(i)) + "' to-ports='" + ConfigData.SERVER_PORT + "' />";
          }
          this.crossDomainPolicy += "</cross-domain-policy>";
        }
        else {
          this.crossDomainPolicy = "<cross-domain-policy><allow-access-from domain='*' to-ports='" + ConfigData.SERVER_PORT + "' /></cross-domain-policy>";
        }
      }

      this.running = true;

      this.serverStartTime = System.currentTimeMillis();
    }
    catch (Exception addr)
    {
      log.severe("\n\n[ --> FATAL ERROR <-- ]: Error initializing server.\n\nCheck if server address and port are properly configured.\n");
      this.running = false;
      System.out.println("Exception caught in " + Thread.currentThread().getName());
      e.printStackTrace();

      System.exit(1);
    }
  }

  private void initLogger()
  {
    if (ConfigData.restartCount == 0)
    {
      try
      {
        ch = new ConsoleHandler();

        String logPath = SmartFoxLib.getFullPath("logs/");

        fh = new FileHandler(logPath + "smartfox_log_%g.txt", ConfigData.LOG_MAX_SIZE, ConfigData.LOG_MAX_FILES);
      }
      catch (Exception e)
      {
        e.printStackTrace();
        System.out.println("FATAL ERROR: Could not create log files.");
        System.exit(1);
      }

      ch.setLevel(ConfigData.CONSOLE_LOGGIN_LEVEL);
      fh.setLevel(ConfigData.FILE_LOGGIN_LEVEL);

      log.addHandler(ch);
      log.addHandler(fh);

      fh.setFormatter(new SmartFoxSimpleLogFormatter());
      ch.setFormatter(new SmartFoxShortLogFormatter());
    }
  }

  private void bootH2Engine()
  {
    try
    {
      log.info("Starting h2 engine...");
      this.h2Engine = Server.createTcpServer(ConfigData.H2_ARGS).start();

      this.h2Connection = 
        new DbManager("org.h2.Driver", 
        ConfigData.H2_CONN_STRING, 
        "sa", 
        "", 
        "h2db", 
        10, 
        10, 
        "grow", 
        5000);

      log.fine("Connection --> " + this.h2Connection);
    }
    catch (Exception e)
    {
      log.warning("Cannot start H2 Engine: " + e);
      e.printStackTrace();
    }
  }

  public void run()
  {
    welcomeMessage();
    try
    {
      this.configReader.readServerConfig();
    }
    catch (ConfigurationException ce)
    {
      System.out.println("Problems in config file: " + ce.getMessage());
      ce.printStackTrace();
      dieOnConfig();
    }

    initLogger();

    bootH2Engine();
    try
    {
      createAdminZone();
    }
    catch (CreateRoomException cre)
    {
      log.warning("Could not create admin zone: " + cre.getMessage());
    }

    this.eventWriter = new EventWriter(this, ConfigData.OUT_QUEUE_THREADS);

    this.sysHandler = new SystemHandler(this, this.eventWriter);
    this.sysHandler.init(this, ConfigData.SYS_HANDLER_THREADS);

    this.bbHandler = new BlueBoxHandler();
    this.bbHandler.init(this, 1);

    this.xtHandler = new ExtensionHandler(this, this.eventWriter);
    this.xtHandler.init(this, ConfigData.EXT_HANDLER_THREADS);

    initServerSocket();

    this.eventReader = new EventReader(this);

    this.connectionCleanerTimer = new Timer();
    this.connectionCleanerTimer.schedule(new ConnectionCleanerTask(ConfigData.MAX_USER_IDLETIME), ConfigData.CONNECTION_CLEANER_INTERVAL * 1000, ConfigData.CONNECTION_CLEANER_INTERVAL * 1000);

    this.autoReloader = new ExtensionAutoReloader();

    if (ConfigData.EXTENSIONS_AUTORELOAD)
    {
      this.extAutoReloader = new Thread(this.autoReloader, "extensionReloader");
      this.extAutoReloader.start();
    }

    setupTaskScheduler();

    log.info("[BlueBoxHandler]... started :: ");
    log.info("[EventWriter] ..... started :: " + ConfigData.OUT_QUEUE_THREADS + " thread(s)");
    log.info("[SystemHandler] ... started :: " + ConfigData.SYS_HANDLER_THREADS + " thread(s)");
    log.info("[ExtensionHandler]. started :: " + ConfigData.EXT_HANDLER_THREADS + " thread(s)");
    log.info("[DeadChannelsPolicy: " + ConfigData.CHANNEL_POLICIES[ConfigData.DEAD_CHANNELS_POLICY] + "]");
    log.info("Server is up and running!");

    IS_SHUTTING_DOWN = false;
    this.running = true;
    int numReady = 0;

    while ((this.running) && (!(IS_SHUTTING_DOWN)))
    {
      acceptNewConnections();
      try
      {
        Thread.sleep(5L);
      }
      catch (InterruptedException ie)
      {
        log.warning("Error during Accept task");
      }
    }
  }

  private void setupTaskScheduler()
  {
    this.taskScheduler = new Scheduler(ConfigData.SCHEDULER_RESOLUTION);
    Task buddyListAutoSaveTask = new Task(BuddyTasks.AUTO_SAVE);
    buddyListAutoSaveTask.parameters.put("zones", this.zones);

    Task buddyListPermissionCleaner = new Task(BuddyTasks.CLEAN_PERMISSIONS);
    buddyListPermissionCleaner.parameters.put("zones", this.zones);

    ITaskHandler buddyTaskHandler = new BuddyListTaskHandler();
    this.taskScheduler.addScheduledTask(buddyListAutoSaveTask, ConfigData.BUDDY_LIST_AUTOSAVE_INTERVAL, true, buddyTaskHandler);
    this.taskScheduler.addScheduledTask(buddyListPermissionCleaner, ConfigData.BUDDY_LIST_PERMISSION_CLEAN_INTERVAL, true, buddyTaskHandler);

    this.taskScheduler.startService();
    log.info("[TaskScheduler] ... started ::");
  }

  private void acceptNewConnections()
  {
    try
    {
      this.acceptSelector.select();

      Set readyKeys = this.acceptSelector.selectedKeys();

      Iterator i = readyKeys.iterator();

      while (i.hasNext())
      {
        SelectionKey key = (SelectionKey)i.next();
        i.remove();

        ServerSocketChannel ssChannel = (ServerSocketChannel)key.channel();

        SocketChannel clientChannel = ssChannel.accept();

        addNewClient(clientChannel);
      }

    }
    catch (IOException ioe)
    {
      log.warning("error during accept(): " + ioe);
    }
    catch (Exception e)
    {
      log.warning("Generic Exception in acceptNewConnections():" + e);
      e.printStackTrace();
    }
  }

  public void addChannelQueue(SocketChannel chan)
  {
    if (chan != null)
    {
      synchronized (this.ChannelQueues)
      {
        this.ChannelQueues.put(chan, new LinkedList());
      }
    }
  }

  private void removeChannelQueue(SocketChannel sc)
  {
    synchronized (this.ChannelQueues)
    {
      this.ChannelQueues.remove(sc);
    }
  }

  public LinkedList getChannelQueue(SocketChannel chan)
  {
    return ((LinkedList)this.ChannelQueues.get(chan));
  }

  public int getGlobalQueueSize()
  {
    int i;
    Object[] qList = (Object[])null;
    int qSize = 0;

    synchronized (this.ChannelQueues)
    {
      qList = this.ChannelQueues.values().toArray();
    }

    for (int i = 0; i < qList.length; ++i)
    {
      LinkedList q = (LinkedList)qList[i];

      if (q != null)
        qSize += q.size();
    }

    return qSize;
  }

  public LinkedList getChannels()
  {
    return this.clients;
  }

  public void setAdminLogin(String login)
  {
    ADMIN_LOGIN = login;
  }

  public void setAdminPassword(String password)
  {
    ADMIN_PASSWORD = password;
  }

  public void setAdminAddresses(String ip)
  {
    if (this.ALLOWED_ADMIN_ADDRESSES == null)
      this.ALLOWED_ADMIN_ADDRESSES = new ArrayList();

    if (this.ALLOWED_ADMIN_ADDRESSES.size() <= 15)
    {
      this.ALLOWED_ADMIN_ADDRESSES.add(ip);

      if (ip.equals("*.*.*.*"))
        this.isAdminRestricted = false;
    }
    else {
      log.warning("Too many Admin Addresses! 15 is the maximum allowed");
    }
  }

  public boolean getAdminAddress(String ip) {
    if (this.isAdminRestricted)
      return this.ALLOWED_ADMIN_ADDRESSES.contains(ip);

    return true;
  }

  public String getAdminLogin()
  {
    return ADMIN_LOGIN;
  }

  public String getAdminPassword()
  {
    return ADMIN_PASSWORD;
  }

  public void setCrossDomainPolicy(String policy)
  {
    this.crossDomainPolicy = policy;
  }

  public void readIncomingMessages()
  {
    try
    {
      this.readSelector.selectNow();

      Set readyKeys = this.readSelector.selectedKeys();

      if (readyKeys.size() <= 0) { return;
      }

      Iterator i = readyKeys.iterator();

      while (i.hasNext())
      {
        SelectionKey key = (SelectionKey)i.next();

        i.remove();

        if (!(key.isValid()))
          continue;

        if (key.isWritable())
        {
          key.interestOps(1);

          label512: this.eventWriter.addNewTask((SocketChannel)key.channel());
        }
        else
        {
          SocketChannel channel = (SocketChannel)key.channel();
          Attachment att = (Attachment)key.attachment();

          User u = getUserByChannel(channel);
          try
          {
            att.readBuffer.clear();

            long nbytes = channel.read(att.readBuffer);

            if (nbytes == -1L)
            {
              lostConnection(channel); break label512:
            }

            att.readBuffer.flip();

            String str = this.asciiDecoder.decode(att.readBuffer).toString();

            att.readBuffer.clear();

            att.msg.append(str);

            String line = att.msg.toString();

            int msgCount = 0;

            while (line.indexOf(0) != -1)
            {
              int posEOM = line.indexOf(0);

              String fullMsg = line.substring(0, posEOM).trim();

              int msgSize = fullMsg.length();

              if (((msgSize > 0) && (msgSize <= ConfigData.MAX_MSG_LEN)) || ((msgSize > ConfigData.MAX_MSG_LEN) && (u.isAdmin())))
              {
                ConfigData.dataIN += msgSize;

                dispatchEvent(fullMsg, channel, (u == null) ? false : u.isAdmin());
              }
              else
              {
                log.warning("Message exceed allowed max length! (" + msgSize + " bytes) from > " + channel.socket().getInetAddress());
                ConfigData.inComingDroppedMessages += 1L;
              }

              att.msg.delete(0, posEOM + 1);
              line = att.msg.toString();

              ++msgCount;
            }

            if (msgCount <= 100) break label512;
            log.warning("Read queue under heavy load (" + msgCount + " msgs)");
          }
          catch (MalformedInputException mie)
          {
            log.warning("Unexpected / Malformed input: " + mie);
            mie.printStackTrace();
          }
          catch (IOException ioe)
          {
            lostConnection(channel);
          }
          catch (Exception e)
          {
            log.warning("Exception during msg reading: " + e);
            e.printStackTrace();
          }

        }

      }

    }
    catch (IOException ioe)
    {
      log.warning("I/O problems while reading socket > " + ioe);
    }
    catch (Throwable t)
    {
      log.warning("Generic Exception: " + t);
      t.printStackTrace();
    }
  }

  public void dispatchEvent(String msg, SocketChannel ch, boolean isAdmin)
  {
    if (IS_SHUTTING_DOWN)
      return;

    String type = "";
    ServerEvent se = null;
    try
    {
      if (msg.charAt(0) == '<')
      {
        this.xmlReader = StdXMLReader.stringReader(msg);
        this.xmlParser.setReader(this.xmlReader);

        IXMLElement header = (IXMLElement)this.xmlParser.parse();

        type = header.getAttribute("t", "");

        se = new ServerEvent(header.getFirstChildNamed("body"), ch);
      }
      else if (msg.charAt(0) == ConfigData.STR_DELIMITER.charAt(0))
      {
        int msgBody = msg.indexOf(ConfigData.STR_DELIMITER, 1);
        type = msg.substring(1, msgBody);

        String body = msg.substring(msgBody + 1, msg.length());

        se = new ServerEvent(body, ch);
      }
      else if (msg.charAt(0) == '{')
      {
        HashMap jso = new JSONObject(msg).getData();
        type = (String)jso.get("t");

        JSONObject body = (JSONObject)jso.get("b");
        se = new ServerEvent(body, ch);
      }
      else if (msg.equals(this.rinfo))
      {
        getRInfo(ch);
      }

      if (se == null) return;

      if (type.equals("sys")) {
        this.sysHandler.handleEvent(se, isAdmin); return;
      }
      if (type.equals("xt")) {
        this.xtHandler.handleEvent(se, isAdmin); return;
      }
      if (!(type.equals("bb"))) return;
      this.bbHandler.handleEvent(se);
    }
    catch (XMLException xe)
    {
      log.warning("Malformed XML request, MSG: " + msg + "\n From: " + ch.socket().getInetAddress());
      log.warning("exception > " + xe);
      ConfigData.inComingDroppedMessages += 1L;
    }
    catch (Exception e)
    {
      log.warning("Generic Error while dispatching message: " + e);
      ConfigData.inComingDroppedMessages += 1L;
      e.printStackTrace();
    }
  }

  public void addNewClient(SocketChannel chan)
  {
    synchronized (this.newClients)
    {
      this.newClients.addLast(chan);
      addChannelQueue(chan);
    }

    if (ConfigData.AUTO_CROSSDOMAIN)
    {
      LinkedList ll = new LinkedList();
      ll.add(chan);
      this.eventWriter.handleEvent(new ServerEvent(this.crossDomainPolicy, null, ll));
    }
  }

  public void checkNewConnections()
  {
    if (this.newClients.size() > 0)
    {
      try
      {
        SocketChannel chan;
        synchronized (this.newClients)
        {
          chan = (SocketChannel)this.newClients.removeFirst();
        }

        if (chan == null) return;

        synchronized (this.clients)
        {
          this.clients.add(chan);
        }

        chan.configureBlocking(false);

        SelectionKey key = chan.register(this.readSelector, 1, new Attachment());

        this.SelectionKeysMap.put(chan, key);
      }
      catch (ClosedChannelException cce)
      {
        log.warning("checkNewConnection Problem: closed channel.");
      }
      catch (IOException ioe)
      {
        log.warning("checkNewConnection Problem: I/O exception: " + ioe);
      }
      catch (Exception e)
      {
        log.warning("checkNewConnection Generic exception: " + e);
      }
    }
  }

  public Selector getReadSelector()
  {
    return this.readSelector;
  }

  public void removeUser(User u)
  {
    if (u != null)
    {
      Zone zn = getZone(u.getZone());

      synchronized (this.usersByChannel)
      {
        this.usersByChannel.remove(u.getChannel());
      }

      synchronized (this.usersById)
      {
        this.usersById.remove(new Integer(u.getUserId()));
      }

      if (zn != null)
      {
        zn.removeName(u.getName(), u);
        zn.destroyVariables(u);
      }

      log.fine("_User [ " + u.getName() + " ] removed");
    }
  }

  public void lostConnection(Integer id)
  {
    User u = getUserById(id);

    if (u != null)
    {
      lostConnection(u.getChannel());
    }
  }

  public void lostConnection(SocketChannel ch)
  {
    Integer userId = null;
    String userName = null;
    try
    {
      User u = (User)this.usersByChannel.get(ch);
      String ip = null;

      if ((ch.socket() != null) && (ch.socket().isConnected()))
      {
        ip = ch.socket().getInetAddress().toString();
        ch.close();
      }

      removeChannelQueue(ch);

      synchronized (this.clients)
      {
        this.clients.remove(ch);
      }

      this.SelectionKeysMap.remove(ch);

      if (u == null) {
        break label276;
      }

      removeIpFromFilter(ip);

      Zone zn = getZone(u.getZone());

      int[] roomsConnected = u.getRoomsConnected();

      int[] playerIndexes = new int[roomsConnected.length];

      for (int i = 0; i < roomsConnected.length; ++i)
      {
        Room tempRoom = zn.getRoom(roomsConnected[i]);

        if (tempRoom != null)
          playerIndexes[i] = u.getPlayerIndex(tempRoom);
        else
          playerIndexes[i] = 0;

      }

      this.sysHandler.broadcastUserLost(u, roomsConnected, playerIndexes);

      u.exitAllRooms();

      userId = new Integer(u.getUserId());
      userName = u.getName();

      if (zn == null)
        break label276;
      zn.removeName(userName, u);
      zn.destroyVariables(u);
    }
    catch (Exception u)
    {
      log.warning("Exception during client disconnection: " + ???);
      ???.printStackTrace();
    }

    synchronized (this.usersByChannel)
    {
      label276: this.usersByChannel.remove(ch);
    }

    synchronized (this.usersById)
    {
      this.usersById.remove(userId);
    }

    log.fine("User [ " + ((userName != null) ? userName : ch.toString()) + " ] removed");
  }

  public void setupZone(IXMLElement zone, boolean reloading)
    throws Exception
  {
    Zone z;
    boolean isNew = false;

    String zoneName = zone.getAttribute("name", "");
    String customLogin = zone.getAttribute("customLogin", "");
    String countUpdate = zone.getAttribute("uCountUpdate", "");
    String z_maxUsers = zone.getAttribute("maxUsers", "");
    String z_maxRooms = zone.getAttribute("maxRooms", "");
    String buddyList = zone.getAttribute("buddyList", "");
    String z_emptyNames = zone.getAttribute("emptyNames", "");
    String z_maxRoomsPerUser = zone.getAttribute("maxRoomsPerUser", "5");
    String z_sendVars = zone.getAttribute("roomListVars", "false");
    String z_roomUpdates = zone.getAttribute("roomUpdates", "true");

    String z_userNameAvoidCh = (zone.getFirstChildNamed("UserNameAvoidChars") != null) ? zone.getFirstChildNamed("UserNameAvoidChars").getContent() : null;
    String z_userRoomAvoidCh = (zone.getFirstChildNamed("RoomNameAvoidChars") != null) ? zone.getFirstChildNamed("RoomNameAvoidChars").getContent() : null;

    String z_maxRoomNameLen = (zone.getFirstChildNamed("MaxRoomNamesLen") != null) ? zone.getFirstChildNamed("MaxRoomNamesLen").getContent() : null;
    String z_maxUserNameLen = (zone.getFirstChildNamed("MaxUserNamesLen") != null) ? zone.getFirstChildNamed("MaxUserNamesLen").getContent() : null;

    String z_autoReloadExt = (zone.getFirstChildNamed("AutoReloadExtensions") != null) ? zone.getFirstChildNamed("AutoReloadExtensions").getContent() : "false";

    String z_maxBvars = (zone.getFirstChildNamed("MaxBuddyVars") != null) ? zone.getFirstChildNamed("MaxBuddyVars").getContent() : null;

    if (reloading)
    {
      z = getInstance().getZone(zoneName);

      if (z == null)
      {
        z = createZone(zoneName, customLogin);
        isNew = true;
        log.info("New zone created: " + zoneName);
      }
      else {
        log.info("Settings for Zone: " + zoneName + " reloaded.");
      }
    } else {
      z = createZone(zoneName, customLogin);
    }

    if (countUpdate.length() > 0)
    {
      if (countUpdate.equalsIgnoreCase("true"))
        z.setCountUpdate(true);
      else
        z.setCountUpdate(false);

    }

    z.userNameAvoidChars = z_userNameAvoidCh;
    z.roomNameAvoidChars = z_userRoomAvoidCh;

    if (z_maxUsers.length() > 0) {
      z.setMaxUsers(Integer.parseInt(z_maxUsers));
    }

    if (z_maxRooms.length() > 0) {
      z.setMaxRooms(Integer.parseInt(z_maxRooms));
    }

    if (z_maxRoomNameLen != null) {
      z.setMaxRoomNameLen(Integer.parseInt(z_maxRoomNameLen));
    }

    if (z_maxUserNameLen != null) {
      z.setMaxUserNameLen(Integer.parseInt(z_maxUserNameLen));
    }

    if ((buddyList.length() > 0) && 
      (!(buddyList.equalsIgnoreCase("false"))))
    {
      IBuddyListPersister persister = new DefaultBuddyListPersister();
      ((AbstractBuddyPersister)persister).setZoneName(zoneName);
      persister.init(null);

      z.initBuddyList(Integer.parseInt(buddyList), persister);

      if (z_maxBvars != null) {
        z.getBuddyManager().setMaxBuddyVars(Integer.parseInt(z_maxBvars));
      }

    }

    z.setVarsOnRoomList(!(z_sendVars.equalsIgnoreCase("false")));

    if (z_emptyNames.length() > 0)
    {
      if (z_emptyNames.equalsIgnoreCase("false"))
        z.setEmptyNames(false);
      else
        z.setEmptyNames(true);

    }

    if (z_maxRoomsPerUser.length() > 0) {
      z.setMaxRoomsPerUser(Integer.parseInt(z_maxRoomsPerUser));
    }

    if (z_roomUpdates.equalsIgnoreCase("false")) {
      z.setRoomUpdates(false);
    }

    if (z_autoReloadExt.equalsIgnoreCase("true")) {
      z.setAutoReloadExtensions(true);
    }

    z.clearDisabledEvents();

    IXMLElement dse = zone.getFirstChildNamed("DisabledSysEvents");

    if (dse != null)
    {
      Enumeration evtEnum = dse.enumerateChildren();

      while (evtEnum.hasMoreElements())
      {
        IXMLElement evt = (IXMLElement)evtEnum.nextElement();
        String evtName = evt.getContent();

        z.addDisabledEvent(evtName);
      }

    }

    z.clearDisabledSysActions();

    IXMLElement dsa = zone.getFirstChildNamed("DisabledSysActions");

    if (dsa != null)
    {
      Enumeration actionEnum = dsa.enumerateChildren();

      while (actionEnum.hasMoreElements())
      {
        IXMLElement action = (IXMLElement)actionEnum.nextElement();
        String actionName = action.getContent();

        z.addDisabledSysAction(actionName);
      }

    }

    IXMLElement dbm = zone.getFirstChildNamed("DatabaseManager");

    if ((dbm != null) && (get_TYP() == 1) && 
      (dbm.getAttribute("active", "").equalsIgnoreCase("true")))
    {
      String driverName = dbm.getFirstChildNamed("Driver").getContent();
      String connStr = dbm.getFirstChildNamed("ConnectionString").getContent();
      String testSQL = dbm.getFirstChildNamed("TestSQL").getContent();

      String userName = dbm.getFirstChildNamed("UserName").getContent();
      String password = dbm.getFirstChildNamed("Password").getContent();

      if (userName == null)
        userName = "";

      if (password == null)
        password = "";

      int maxActive = Integer.parseInt(dbm.getFirstChildNamed("MaxActive").getContent());
      int maxIdle = Integer.parseInt(dbm.getFirstChildNamed("MaxIdle").getContent());

      String exhaustedAction = dbm.getFirstChildNamed("OnExhaustedPool").getContent().toLowerCase();
      int blockTime = Integer.parseInt(dbm.getFirstChildNamed("BlockTime").getContent());

      if (("".equals(driverName)) || (driverName == null))
      {
        throw new DbManagerException("Missing Driver Name");
      }

      if (("".equals(connStr)) || (connStr == null))
      {
        throw new DbManagerException("Missing connection string");
      }

      if (("".equals(testSQL)) || (testSQL == null))
      {
        throw new DbManagerException("Missing Test SQL statement");
      }

      if (!(testSQL.toLowerCase().startsWith("select")))
      {
        throw new DbManagerException("The Test SQL statement should start with a SELECT Sql command!");
      }

      if ((!(exhaustedAction.equals("fail"))) && (!(exhaustedAction.equals("grow"))) && (!(exhaustedAction.equals("wait"))))
      {
        exhaustedAction = "fail";
      }

      if (!(reloading)) {
        System.out.println("\tDB Manager Activated ( " + driverName + " )");
      }

      z.dbManager = 
        new DbManager(driverName, connStr, userName, 
        password, zoneName, maxActive, maxIdle, 
        exhaustedAction, blockTime);

      ArrayList al = z.dbManager.executeQuery(testSQL);

      if (al == null) {
        throw new DbManagerException("The Test SQL statement failed! Please check your configuration.");
      }

    }

    IXMLElement buddyListNode = zone.getFirstChildNamed("BuddyList");

    if ((buddyListNode != null) && (buddyList.length() == 0))
    {
      String isBuddyListActive = buddyListNode.getAttribute("active", "false");
      String maxBuddyVars = (buddyListNode.getFirstChildNamed("maxBuddyVariables") != null) ? buddyListNode.getFirstChildNamed("maxBuddyVariables").getContent() : null;
      String buddyListSize = (buddyListNode.getFirstChildNamed("size") != null) ? buddyListNode.getFirstChildNamed("size").getContent() : null;
      String buddyListPersister = (buddyListNode.getFirstChildNamed("persisterClass") != null) ? buddyListNode.getFirstChildNamed("persisterClass").getContent() : null;
      String useOfflineVars = (buddyListNode.getFirstChildNamed("offLineBuddyVariables") != null) ? buddyListNode.getFirstChildNamed("offLineBuddyVariables").getContent() : "false";
      String securityMode = (buddyListNode.getFirstChildNamed("mode") != null) ? buddyListNode.getFirstChildNamed("mode").getContent() : "basic";
      String mutualAdd = (buddyListNode.getFirstChildNamed("mutualAddBuddy") != null) ? buddyListNode.getFirstChildNamed("mutualAddBuddy").getContent() : "true";
      String mutualRemove = (buddyListNode.getFirstChildNamed("mutualRemoveBuddy") != null) ? buddyListNode.getFirstChildNamed("mutualRemoveBuddy").getContent() : "true";
      String permTimeout = (buddyListNode.getFirstChildNamed("permissionTimeOut") != null) ? buddyListNode.getFirstChildNamed("permissionTimeOut").getContent() : "60";
      String usePermission = (buddyListNode.getFirstChildNamed("addBuddyPermission") != null) ? buddyListNode.getFirstChildNamed("addBuddyPermission").getContent() : "true";

      if (isBuddyListActive.equalsIgnoreCase("true"))
      {
        int blistSize = 50;
        int blistMaxBVars = 10;

        if (buddyListSize != null)
          blistSize = Integer.parseInt(buddyListSize);

        if (maxBuddyVars != null)
          blistMaxBVars = Integer.parseInt(maxBuddyVars);

        if (blistSize < 1) {
          throw new ConfigurationException("Invalid BuddyList size for Zone: " + zoneName);
        }

        String persisterClass = (buddyListPersister == null) ? 
          "it.gotoandplay.smartfoxserver.data.buddylist.persistence.DefaultBuddyListPersister" : (securityMode.equalsIgnoreCase("advanced")) ? 
          "it.gotoandplay.smartfoxserver.data.buddylist.persistence.AdvancedBuddyListPersister" : buddyListPersister;

        IBuddyListPersister persister = (IBuddyListPersister)Class.forName(persisterClass).newInstance();
        ((AbstractBuddyPersister)persister).setZoneName(zoneName);
        persister.init(null);

        z.initBuddyList(blistSize, persister);

        BuddyListManager bManager = z.getBuddyManager();

        bManager.setMaxBuddyVars(blistMaxBVars);

        bManager.setUseOfflineVariables(useOfflineVars.equalsIgnoreCase("true"));

        bManager.setSecurityMode((securityMode.equalsIgnoreCase("advanced")) ? 1 : 0);

        bManager.setMutualAddBuddy(mutualAdd.equalsIgnoreCase("true"));

        bManager.setMutualRemoveBuddy(mutualRemove.equalsIgnoreCase("true"));

        bManager.setAddBuddyPermission(usePermission.equalsIgnoreCase("true"));
        try
        {
          bManager.setBuddyPermissionTimeout(Integer.parseInt(permTimeout));
        }
        catch (NumberFormatException localNumberFormatException)
        {
        }
      }

    }

    IXMLElement mods = zone.getFirstChildNamed("Moderators");

    if (mods != null)
    {
      z.modManager.setIsActive(mods.getAttribute("status", "").equalsIgnoreCase("on"));

      Enumeration sub = mods.enumerateChildren();

      z.modManager.clearList();

      while (sub.hasMoreElements())
      {
        IXMLElement mod = (IXMLElement)sub.nextElement();

        String name = mod.getAttribute("name", "");
        String pwd = mod.getAttribute("pwd", "");

        z.modManager.addMod(new Moderator(name, pwd));
      }

    }

    if ((!(reloading)) || ((reloading) && (isNew)))
    {
      IXMLElement roomsNode = zone.getFirstChildNamed("Rooms");
      Enumeration roomNodes = roomsNode.enumerateChildren();
      StringBuffer roomDesc = new StringBuffer();

      if (!(reloading))
        System.out.println("Zone: " + zoneName + "\n");

      while (roomNodes.hasMoreElements())
      {
        int maxUsers = 40;
        int maxSpectators = 0;

        IXMLElement room = (IXMLElement)roomNodes.nextElement();

        String roomName = room.getAttribute("name", "");

        String maxU = room.getAttribute("maxUsers", "");
        if (!(maxU.equals(""))) {
          maxUsers = Integer.parseInt(maxU);
        }

        String maxSpec = room.getAttribute("maxSpectators", "");
        if (!(maxSpec.equals("")))
          maxSpectators = Integer.parseInt(maxSpec);

        boolean isPrivate = Boolean.valueOf(room.getAttribute("isPrivate", "")).booleanValue();
        boolean isTemp = Boolean.valueOf(room.getAttribute("isTemp", "")).booleanValue();
        boolean isGame = Boolean.valueOf(room.getAttribute("isGame", "")).booleanValue();
        boolean autoJoin = Boolean.valueOf(room.getAttribute("autoJoin", "")).booleanValue();
        String uCnt = room.getAttribute("uCountUpdate", "");
        String limbo = room.getAttribute("limbo", "false");
        String password = room.getAttribute("pwd", "");

        Room rm = createRoom(zoneName, roomName, maxUsers, maxSpectators, isTemp, isGame, isPrivate, password);

        if (uCnt.equalsIgnoreCase("true"))
          rm.setUcountUpdates(true);
        else if (uCnt.equalsIgnoreCase("false")) {
          rm.setUcountUpdates(false);
        }

        if (limbo.equalsIgnoreCase("true")) {
          rm.setLimbo(true);
        }

        roomDesc.delete(0, roomDesc.length());
        roomDesc.append("\t");

        if (roomName.length() > 18)
          roomDesc.append(roomName.substring(0, 15)).append("...");
        else {
          roomDesc.append(roomName);
        }

        char[] spaces = new char[20 - roomDesc.length()];
        Arrays.fill(spaces, ' ');

        roomDesc.append(spaces);
        roomDesc.append("(id: ").append(rm.getId()).append(", max: ").append(maxUsers).append(", pass:").append((password.equals("")) ? "N" : "Y").append(")");

        if (!(reloading)) {
          System.out.println(roomDesc.toString());
        }

        if (autoJoin) {
          z.setAutoJoinRoom(rm.getId());
        }

        IXMLElement vars = room.getFirstChildNamed("Vars");

        if (vars != null)
        {
          Enumeration varNodes = vars.enumerateChildren();

          while (varNodes.hasMoreElements())
          {
            IXMLElement roomVar = (IXMLElement)varNodes.nextElement();

            String vName = roomVar.getAttribute("name", "");
            String vType = roomVar.getAttribute("type", "");
            String vValue = roomVar.getContent();
            boolean priv = roomVar.getAttribute("private", "").equals("true");
            boolean persist = roomVar.getAttribute("persistent", "").equals("true");

            if (!(vName.equals("")))
              rm.setVariable(vName, vType, vValue, priv, persist, null, true);

          }

        }

        IXMLElement roomExtNode = room.getFirstChildNamed("Extensions");

        if (roomExtNode != null) {
          this.xtHandler.parse_Extensions(roomExtNode, z.getName(), rm.getId());
        }

      }

    }

    IXMLElement extNode = zone.getFirstChildNamed("Extensions");

    if (extNode != null)
      this.xtHandler.parse_Extensions(extNode, z.getName(), -1);

    if (!(reloading))
      System.out.println("");
  }

  public Zone createZone(String zoneName, String customLogin)
  {
    if (!(zoneName.equals("$dmn")))
    {
      Zone z = new Zone(zoneName, customLogin);
      this.zones.put(zoneName, z);

      return z;
    }

    return null;
  }

  private void createAdminZone()
    throws CreateRoomException
  {
    Zone adminZ = new Zone("$dmn", "true");
    this.zones.put("$dmn", adminZ);

    Room adminR = new Room("adminRoom", 5, false, false, false, "", "$dmn");

    adminZ.addRoom(adminR);

    adminZ.setAutoJoinRoom(adminR.getId());
  }

  public Zone getZone(String zoneName)
  {
    return ((Zone)this.zones.get(zoneName));
  }

  public LinkedList getAllZones()
  {
    LinkedList list;
    synchronized (this.zones)
    {
      list = new LinkedList(this.zones.values());
    }

    return list;
  }

  public int getZoneNumber()
  {
    return this.zones.size();
  }

  public int getRoomNumber()
  {
    int howMany = 0;
    LinkedList zoneList = null;

    synchronized (this.zones)
    {
      zoneList = new LinkedList(this.zones.values());
    }

    for (Iterator i = zoneList.iterator(); i.hasNext(); )
    {
      Zone z = (Zone)i.next();

      howMany += z.getRoomCount();
    }

    return howMany;
  }

  public User getUserByChannel(SocketChannel usrChannel)
  {
    synchronized (this.usersByChannel)
    {
      return ((User)this.usersByChannel.get(usrChannel));
    }
  }

  public User getUserById(Integer i)
  {
    return ((User)this.usersById.get(i));
  }

  public List<User> getBlueBoxedUsers()
  {
    List blueBoxedUsers = new ArrayList();

    synchronized (this.usersByChannel)
    {
      Collection users = this.usersByChannel.values();

      for (Iterator it = users.iterator(); it.hasNext(); )
      {
        User user = (User)it.next();

        if (!(user.isBlueBoxed())) break label63;
        label63: blueBoxedUsers.add(user);
      }
    }

    return blueBoxedUsers;
  }

  public long getServerStartTime()
  {
    return this.serverStartTime;
  }

  public Room createRoom(String zoneName, String name, int maxUsers, int maxSpectators, boolean isTemp, boolean isGame, boolean isPrivate, String password)
    throws CreateRoomException
  {
    if ((ConfigData.BADWORDS_ACTIVE) && (ConfigData.BADWORDS_ROOM_NAMES) && 
      (BadWordsFilter.getInstance().containsBadWords(name))) {
      throw new CreateRoomException("Room name contains bad words!");
    }

    Room rm = new Room(name, maxUsers, isTemp, isGame, isPrivate, password, zoneName);

    if (isGame) {
      rm.setMaxSpectators(maxSpectators);
    }

    Zone zn = (Zone)this.zones.get(zoneName);

    if (zn.isActive())
      zn.addRoom(rm);
    else
      throw new CreateRoomException("Zone is not active!");

    return rm;
  }

  public User canLogin(String name, String password, SocketChannel ch, String zone)
    throws LoginException
  {
    if (ch == null)
      throw new LoginException("Socket connection is null");

    if (!(ch.isConnected())) {
      throw new LoginException("Socket connection is closed");
    }

    User usr = (User)this.usersByChannel.get(ch);

    if (usr != null) {
      throw new LoginException("User already logged in");
    }

    Zone zn = (Zone)this.zones.get(zone);
    if (zn == null) {
      throw new LoginException("Zone does not exist!");
    }

    if (!(zn.isActive())) {
      throw new LoginException("This Zone is not active!");
    }

    if (zn.validateUserName(name)) {
      throw new LoginException("This user name is already taken.");
    }

    if ((ConfigData.BADWORDS_ACTIVE) && (ConfigData.BADWORDS_USER_NAMES) && 
      (BadWordsFilter.getInstance().containsBadWords(name))) {
      throw new LoginException("Your username contains badwords!");
    }

    if (!(addIpToFilter(ch.socket().getInetAddress().toString()))) {
      throw new LoginException("Too many connections coming from your ip address.");
    }

    boolean b = this.clients.size() >= 0;

    boolean c = this.usersByChannel.size() < get_MAX();

    boolean d = (b) && (c);

    boolean e = this.usersByChannel == null;

    User u = null;

    if ((d) && (!(e)) && (b))
    {
      u = new User(ch, name, zone);

      if (zn.modManager.isModerator(name, password))
        u.setAsModerator();

      synchronized (this.usersByChannel)
      {
        this.usersByChannel.put(ch, u);
      }

      Integer uid = new Integer(u.getUserId());

      synchronized (this.usersById)
      {
        this.usersById.put(uid, u);
      }

      zn.addName(name, uid);

      if (this.usersById.size() > ConfigData.maxSimultanousConnections)
        ConfigData.maxSimultanousConnections = this.usersById.size();

      log.fine("User [ " + name + " ] logged in");
    }
    else
    {
      throw new LoginException(ShutdownManager.mesg);
    }
    return u;
  }

  public void addExtension(String name, ExtensionItem eItem)
  {
    this.extensions.put(name, eItem);
  }

  public Set getExtensions()
  {
    return this.extensions.keySet();
  }

  public ExtensionItem getExtension(String itemKey)
  {
    return ((ExtensionItem)this.extensions.get(itemKey));
  }

  public String getGuestName()
  {
    if (this.guestCount >= 9223372036854775807L)
      this.guestCount = 0L;

    return "guest_" + (this.guestCount++);
  }

  public int getGlobalUserCount()
  {
    return this.usersByChannel.size();
  }

  public EventWriter getEW()
  {
    return this.eventWriter;
  }

  private void dieOnConfig()
  {
    System.out.println("");
    System.out.println("Errors found in configuration files.\nUnable to start the server.");
    System.exit(1);
  }

  private static void showSystemInfo()
  {
    List props = new ArrayList();

    System.out.println("\n\n--- [ System Info ] ------------------------------------------\n");
    props.add("os.name");
    props.add("os.arch");
    props.add("os.version");
    props.add("java.version");
    props.add("java.vendor");
    props.add("java.vendor.url");
    props.add("java.vm.specification.version");
    props.add("java.vm.version");
    props.add("java.vm.vendor");
    props.add("java.vm.name");

    Runtime rt = Runtime.getRuntime();

    System.out.println("System CPU(s): " + rt.availableProcessors());

    System.out.println("VM Max memory: " + (rt.maxMemory() / 1000000L) + " MB");
    System.out.println("");

    for (Iterator i = props.iterator(); i.hasNext(); )
    {
      String prop = (String)i.next();
      System.out.println(prop + ": " + System.getProperty(prop));
    }

    System.out.println("\n\n--- [ Network Cards ] -----------------------------------------\n");
    try
    {
      Enumeration list = NetworkInterface.getNetworkInterfaces();

      while (list.hasMoreElements())
      {
        NetworkInterface iFace = (NetworkInterface)list.nextElement();
        System.out.println("Card:" + iFace.getDisplayName());

        Enumeration addresses = iFace.getInetAddresses();

        while (addresses.hasMoreElements())
        {
          InetAddress adr = (InetAddress)addresses.nextElement();
          System.out.println(" -> " + adr.getHostAddress());
        }
      }
    }
    catch (SocketException se)
    {
      System.out.println("Failed discovering network cards!");
      System.out.println("Error: " + se);
    }
  }

  public SystemHandler getSysHandler()
  {
    return this.sysHandler;
  }

  private LinkedList loadBannedUserList()
  {
    LinkedList ll = null;
    try
    {
      FileInputStream fis = new FileInputStream("bannedUsers.sfs");
      ObjectInputStream ois = new ObjectInputStream(fis);

      ll = (LinkedList)ois.readObject();
      ois.close();

      if (ll != null)
        break label58;
      ll = new LinkedList();
    }
    catch (Exception e)
    {
      ll = new LinkedList();
    }

    label58: return ll;
  }

  public void saveBannedUserList()
  {
    FileOutputStream fos;
    try
    {
      fos = new FileOutputStream("bannedUsers.sfs");
      ObjectOutputStream oos = new ObjectOutputStream(fos);

      oos.writeObject(this.bannedUsers);
      oos.flush();

      oos.close();
    }
    catch (Exception e)
    {
      log.warning("Could not save banned user list to file: bannedUsers.sfs");
    }
  }

  public void addBannedUser(BannedUser bu)
  {
    boolean found = false;

    if ((bu.getIp() != null) && ((($_BB.contains(bu.getIp())) || (bu.getIp().equals("127.0.0.1")))))
      return;

    synchronized (this.bannedUsers)
    {
      for (Iterator i = this.bannedUsers.iterator(); i.hasNext(); )
      {
        BannedUser bbu = (BannedUser)i.next();
        if (!(bbu.equals(bu)))
          break label82;
        found = true;
        label82: break;
      }

      if (!(found))
      {
        this.bannedUsers.add(bu);

        if (this.bannedUsers.size() > 500)
          log.warning("The BanList contains more than 500 items! It's time to do some cleanup!");
      }
    }
  }

  public void removeBannedUser(BannedUser bu)
  {
    synchronized (this.bannedUsers)
    {
      for (Iterator i = this.bannedUsers.iterator(); i.hasNext(); )
      {
        if (!(bu.equals((BannedUser)i.next())))
          break label43;
        i.remove();
        label43: break;
      }
    }
  }

  public void saveBanList()
  {
    FileOutputStream fos;
    try
    {
      fos = new FileOutputStream("bannedUsers.sfs");
      ObjectOutputStream oos = new ObjectOutputStream(fos);

      oos.writeObject(this.bannedUsers);
      oos.flush();
      oos.close();
    }
    catch (Exception e)
    {
      log.severe("Can't write to banlist file: bannedUsers.sfs");
      e.printStackTrace();
    }
  }

  public boolean isBanned(String nick, SocketChannel sc)
  {
    boolean result = false;

    if (sc == null)
      return result;

    long now = System.currentTimeMillis();

    String ipAdr = sc.socket().getInetAddress().toString().substring(1);

    synchronized (this.bannedUsers)
    {
      for (Iterator i = this.bannedUsers.iterator(); i.hasNext(); )
      {
        BannedUser bu = (BannedUser)i.next();

        String bannedIp = bu.getIp();
        String bannedName = bu.getName();
        long banTime = bu.getBanDate();

        if (now > banTime + ConfigData.BAN_DURATION * 1000L)
        {
          i.remove();
          break label147:
        }

        if ((bannedIp != null) && (ipAdr.equals(bannedIp)))
        {
          result = true;
          break;
        }

        if ((bannedName == null) || (!(nick.equals(bannedName))))
          break label147;
        result = true;
        label147: break;
      }

    }

    return result;
  }

  public LinkedList getBanList()
  {
    return this.bannedUsers;
  }

  public SelectionKey getChannelKey(SocketChannel sc)
  {
    return ((SelectionKey)this.SelectionKeysMap.get(sc));
  }

  public ExtensionHandler getExtensionHandler()
  {
    return this.xtHandler;
  }

  public void addKickedUser(User u, int reason)
  {
    String ip = u.getIpAddress();

    synchronized (this.kickedUsers)
    {
      KickedUser ku = (KickedUser)this.kickedUsers.get(ip);

      if (ku == null)
      {
        ku = new KickedUser(ip);
        this.kickedUsers.put(ip, ku);
      }

      if (reason == 1)
        ku.addKick();
      else if (reason == 2) {
        ku.addBadWordKick();
      }

      if (ku.isBannable())
      {
        if (reason == 1)
          this.sysHandler.sendAdminMessage(ConfigData.ANTIFLOOD_BAN_MSG, u, "dmn");
        else if (reason == 2) {
          this.sysHandler.sendAdminMessage(ConfigData.BADWORDS_BAN_MSG, u, "dmn");
        }

        addBannedUser(new BannedUser(null, ip));
        log.warning("Banning user " + u.getName() + " with IP: " + ip);
      }
    }
  }

  public HashMap getKickedUsers()
  {
    return this.kickedUsers;
  }

  public LinkedList getGlobalUserList()
  {
    LinkedList res = null;

    synchronized (this.usersByChannel)
    {
      res = new LinkedList((Collection)this.usersByChannel.keySet());
    }

    return res;
  }

  public void closeAllBuddyManagers()
  {
    Zone z = null;

    if (this.zones != null)
    {
      synchronized (this.zones)
      {
        Collection items = this.zones.values();

        for (Iterator it = items.iterator(); it.hasNext(); )
        {
          z = (Zone)it.next();
          z.destroy(null);
        }
      }
    }
  }

  private boolean addIpToFilter(String ip)
  {
    if (ConfigData.MAX_USERS_PER_IP <= 0)
      return true;

    boolean b = false;

    synchronized (this.ipFilter)
    {
      Integer cnt = (Integer)this.ipFilter.get(ip);

      if (cnt != null)
      {
        int c = cnt.intValue();

        if (c < ConfigData.MAX_USERS_PER_IP)
        {
          ++c;
          this.ipFilter.put(ip, new Integer(c));
          b = true;
        }
      }
      else
      {
        this.ipFilter.put(ip, new Integer(1));
        b = true;
      }
    }

    return b;
  }

  private void removeIpFromFilter(String ip)
  {
    if (ConfigData.MAX_USERS_PER_IP > 0)
    {
      synchronized (this.ipFilter)
      {
        Integer cnt = (Integer)this.ipFilter.get(ip);

        if (cnt != null)
        {
          int c = cnt.intValue();

          --c;
          this.ipFilter.put(ip, new Integer(c));

          if (c <= 0)
            this.ipFilter.remove(ip);
        }
        else {
          log.warning("Could not remove ip address from IpFilter: " + ip);
        }
      }
    }
  }

  public void logOutUser(User user, boolean broadcastInternalEvent)
  {
    Zone zone = getZone(user.getZone());

    if (zone != null)
    {
      int[] roomsConnected = user.getRoomsConnected();

      int[] playerIndexes = new int[roomsConnected.length];

      for (int i = 0; i < roomsConnected.length; ++i)
      {
        Room tempRoom = zone.getRoom(roomsConnected[i]);

        if (tempRoom != null)
          playerIndexes[i] = user.getPlayerIndex(tempRoom);
        else {
          playerIndexes[i] = 0;
        }

      }

      if (broadcastInternalEvent) {
        this.sysHandler.broadCastUserLogout(user, user.getZone(), roomsConnected, playerIndexes);
      }

      user.exitAllRooms();

      Integer i = new Integer(user.getUserId());
      String name = user.getName();

      zone.removeName(name, user);
      zone.destroyVariables(user);

      synchronized (this.usersByChannel)
      {
        this.usersByChannel.remove(user.getChannel());
      }

      synchronized (this.usersById)
      {
        this.usersById.remove(i);
      }

      log.fine("User [ " + name + " ] logged out");
    }
  }

  public void restart(User who)
  {
    if ((!(IS_SHUTTING_DOWN)) && 
      (who.isAdmin()))
    {
      closeAllBuddyManagers();

      System.out.println("\n\nServer Restarting ...\n\n");

      if (shutDown())
      {
        System.out.println("\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n");

        instance = new SmartFoxServer();
        getInstance().start();
      }
      else
      {
        System.out.println("Could not restart server!");
      }
    }
  }

  public void halt(User who)
  {
    if ((!(IS_SHUTTING_DOWN)) && 
      (who.isAdmin()))
    {
      closeAllBuddyManagers();
      saveBanList();

      System.out.println("\n\nHalting Server ...\n\n");

      shutDown();
    }
  }

  private boolean shutDown()
  {
    boolean res = false;
    try
    {
      IS_SHUTTING_DOWN = true;

      int cn = 0;

      synchronized (this.zones)
      {
        for (Iterator it = this.zones.values().iterator(); it.hasNext(); )
          ((Zone)it.next()).destroy(null);
      }

      synchronized (this.clients)
      {
        for (Iterator i = this.clients.iterator(); i.hasNext(); )
        {
          SocketChannel ch = (SocketChannel)i.next();
          try
          {
            ch.close();
            ++cn;
            i.remove();
          }
          catch (IOException e)
          {
            log.warning("IOException during clients shutdown: " + e.getMessage());
            e.printStackTrace();
          }
          catch (Exception e)
          {
            log.warning("Exception while shutting down all connections: " + e.getMessage());
            e.printStackTrace();
          }

        }

      }

      this.running = false;

      this.taskScheduler.destroy(null);
      this.sysHandler.shutdown();
      this.eventWriter.shutdown();
      this.bbHandler.shutdown();

      clearAllExtensions();
      this.xtHandler.shutdown();

      this.connectionCleanerTimer.cancel();

      ExtensionHelper.destroy();

      saveBannedUserList();

      this.acceptSelector.wakeup();
      this.acceptSelector.close();
      this.readSelector.close();

      this.acceptSelector = null;
      this.readSelector = null;
      this.acceptSelector = null;
      this.sSockChan.socket().close();
      this.sSockChan.close();
      this.sSockChan = null;

      Room.resetRoomStaticData();
      User.resetUserStaticData();

      ConfigData.inComingDroppedMessages = 0L;
      ConfigData.outGoingDroppedMessages = 0L;
      ConfigData.maxSimultanousConnections = 0;
      ConfigData.dataIN = 0L;
      ConfigData.dataOUT = 0L;

      this.clients = null;
      this.newClients = null;

      this.eventControllers = null;
      this.usersByChannel = null;
      this.usersById = null;
      this.zones = null;
      this.ChannelQueues = null;
      this.SelectionKeysMap = null;

      this.bannedUsers = null;
      this.kickedUsers = null;

      if (this.h2Engine.isRunning())
      {
        this.h2Engine.stop();
        this.h2Connection.shutDown();
      }

      if (ConfigData.USE_WEBSERVER)
      {
        WebServer.getInstance().stop();
        WebServer.getInstance().destroy();
      }

      if (this.extAutoReloader != null)
      {
        this.autoReloader.halt();
        this.extAutoReloader = null;
      }

      System.gc();

      Thread.sleep(2000L);

      ConfigData.restartCount += 1;

      res = true;
    }
    catch (IOException ioe)
    {
      log.warning("Could not close connections during shutdown!!");
    }
    catch (InterruptedException ie)
    {
      log.warning("Thread interrupted during shutdown!");
    }
    catch (Exception e)
    {
      log.warning("Generic exception while shutting down: " + e);
      e.printStackTrace();
    }

    return res;
  }

  private void clearAllExtensions()
  {
    synchronized (this.zones)
    {
      for (Iterator i = this.zones.values().iterator(); i.hasNext(); )
      {
        Zone z = (Zone)i.next();

        if (z == null) break label41;
        label41: z.clearAllExtensions();
      }
    }
  }

  private void getRInfo(SocketChannel ch)
  {
    StringBuffer sb = new StringBuffer();
    sb.append(getType()).append(", ").append(get_MAX()).append(", ").append(getVersion());
    LinkedList ll = new LinkedList();
    ll.add(ch);

    this.eventWriter.handleEvent(new ServerEvent(sb.toString(), null, ll));
  }

  public String getVersion()
  {
    return "1.6.2";
  }

  public int getType()
  {
    return get_TYP();
  }

  public static void set_TYP(int lC_TYP)
  {
    if ($_TYP < 0)
      $_TYP = lC_TYP;
  }

  public static int get_TYP()
  {
    return $_TYP;
  }

  public static void set_CLI(String lC_CLI)
  {
    $_CLI = lC_CLI;
  }

  public static String get_CLI()
  {
    return $_CLI;
  }

  public static void set_ADD(List lC_ADD)
  {
    if ($_ADD == null)
      $_ADD = lC_ADD;
  }

  public static void set_BB(List lC_BB)
  {
    if ($_BB == null)
    {
      $_BB = lC_BB;
    }
  }

  public static String[] get_BB()
  {
    if ($_BB.size() == 0)
      return null;

    String[] bbs = new String[$_BB.size()];

    int c = 0;
    for (Iterator it = $_BB.iterator(); it.hasNext(); )
    {
      String s = (String)it.next();
      bbs[(c++)] = s;
    }

    return bbs;
  }

  public boolean bbOk(String source)
  {
    boolean b = false;

    if ($_BB != null)
    {
      return $_BB.contains(source);
    }

    return b;
  }

  public static List get_ADD()
  {
    return $_ADD;
  }

  public static void set_MAX(int lC_MAX)
  {
    if ($_MAX < 0)
      $_MAX = lC_MAX;
  }

  public static int get_MAX()
  {
    return $_MAX;
  }

  public static int getMAX_ROOM_VARS()
  {
    return ConfigData.MAX_ROOM_VARS;
  }

  public static void setMAX_ROOM_VARS(int max_room_vars)
  {
    if (ConfigData.MAX_ROOM_VARS == -1)
      ConfigData.MAX_ROOM_VARS = max_room_vars;
  }

  public static int getMAX_USER_VARS()
  {
    return ConfigData.MAX_USER_VARS;
  }

  public static void setMAX_USER_VARS(int max_user_vars)
  {
    ConfigData.MAX_USER_VARS = max_user_vars;
  }

  public static int getMAX_INCOMING_QUEUE()
  {
    return ConfigData.MAX_INCOMING_QUEUE;
  }

  public static void setMAX_INCOMING_QUEUE(int max)
  {
    if (max > 500)
      ConfigData.MAX_INCOMING_QUEUE = max;
  }

  public static SmartFoxServer getInstance()
  {
    return instance;
  }

  public void initWebServer()
  {
    WebServer ws = WebServer.getInstance();
    try
    {
      ws.start();
    }
    catch (Exception e)
    {
      log.warning("Could not start webserver: " + e);
    }
  }

  public DbManager getCoreDbManager()
  {
    return this.h2Connection;
  }

  public PyHelper getPyHelper()
  {
    return this.pyHelper;
  }

  public Map getDebugData()
  {
    Map result = null;

    if (ConfigData.debugInfo)
    {
      result = new HashMap();
      result.put("channels", this.clients);
      result.put("usersByChannel", this.usersByChannel);
      result.put("usersById", this.usersById);
      result.put("zones", this.zones);
      result.put("queues", this.ChannelQueues);
    }

    return result;
  }
}