package it.gotoandplay.smartfoxserver.controllers;

import it.gotoandplay.smartfoxserver.EventWriter;
import it.gotoandplay.smartfoxserver.SmartFoxServer;
import it.gotoandplay.smartfoxserver.config.ConfigData;
import it.gotoandplay.smartfoxserver.lib.EventQueue;
import it.gotoandplay.smartfoxserver.lib.ServerEvent;
import java.net.InetAddress;
import java.net.Socket;
import java.nio.channels.SocketChannel;
import java.util.ArrayList;
import java.util.LinkedList;
import java.util.List;
import java.util.logging.Logger;
import net.n3.nanoxml.IXMLElement;

public class BlueBoxHandler extends MessageHandler
{
  private List<String> allowedSources;
  private EventWriter eventWriter;

  protected void setupHandler()
  {
    this.eventWriter = SmartFoxServer.getInstance().getEW();
  }

  public void run()
  {
    this.running = true;

    label43: while (this.running)
    {
      try
      {
        ServerEvent event;
        if ((event = this.eventQueue.deQueue()) == null)
          break label43;
        processEvent(event);
      }
      catch (InterruptedException e)
      {
        if (!(SmartFoxServer.IS_SHUTTING_DOWN))
          SmartFoxServer.log.warning("Problems during event processing");
      }
    }
  }

  protected void processEvent(ServerEvent event)
  {
    try
    {
      if (this.allowedSources == null)
        initAllowedSources();

      action = null;
      SocketChannel channel = event.getSenderChannel();

      if (!(this.allowedSources.contains(channel.socket().getInetAddress().getHostAddress()))) { return;
      }

      IXMLElement bodyTag = event.getXmlMsg();

      action = bodyTag.getAttribute("action", "");

      if (!(action.equals("hs"))) return;
      handleHandShake(channel);
    }
    catch (java.lang.Exception action)
    {
    }
  }

  private void initAllowedSources()
  {
    this.allowedSources = new ArrayList();
    String[] bbs = SmartFoxServer.get_BB();

    if (bbs == null) {
      this.allowedSources.add("127.0.0.1");
    }
    else
    {
      for (String s : bbs)
        this.allowedSources.add(s);
    }
  }

  private void handleHandShake(SocketChannel channel)
  {
    String[] bbs = SmartFoxServer.get_BB();
    StringBuilder response = new StringBuilder("hs_").append((bbs == null) ? 0 : 1);
    response.append(",").append(ConfigData.MAX_USER_IDLETIME);

    LinkedList recipients = new LinkedList();
    recipients.add(channel);

    this.eventWriter.handleEvent(new ServerEvent(response.toString(), channel, recipients));
  }
}