package it.gotoandplay.smartfoxserver.controllers;

import it.gotoandplay.smartfoxserver.EventWriter;
import it.gotoandplay.smartfoxserver.SmartFoxServer;
import it.gotoandplay.smartfoxserver.config.ConfigData;
import it.gotoandplay.smartfoxserver.data.Room;
import it.gotoandplay.smartfoxserver.data.User;
import it.gotoandplay.smartfoxserver.data.Zone;
import it.gotoandplay.smartfoxserver.exceptions.MissingExtensionException;
import it.gotoandplay.smartfoxserver.extensions.AbstractExtension;
import it.gotoandplay.smartfoxserver.extensions.AdminExtension;
import it.gotoandplay.smartfoxserver.extensions.ExtensionManager;
import it.gotoandplay.smartfoxserver.extensions.ISmartFoxExtension;
import it.gotoandplay.smartfoxserver.extensions.JavascriptExtension;
import it.gotoandplay.smartfoxserver.extensions.PythonExtension;
import it.gotoandplay.smartfoxserver.lib.ASObjectSerializer;
import it.gotoandplay.smartfoxserver.lib.ActionscriptObject;
import it.gotoandplay.smartfoxserver.lib.EventQueue;
import it.gotoandplay.smartfoxserver.lib.ServerEvent;
import java.io.File;
import java.io.PrintStream;
import java.net.Socket;
import java.net.URL;
import java.net.URLClassLoader;
import java.nio.channels.SocketChannel;
import java.util.Enumeration;
import java.util.Map;
import java.util.logging.Logger;
import net.n3.nanoxml.IXMLElement;
import org.json.JSONObject;
import org.mozilla.javascript.EvaluatorException;

public class ExtensionHandler extends MessageHandler
{
  SmartFoxServer sf;
  EventWriter eventWriter;
  Zone adminZone;
  private final String JAVA_EXTENSIONS_PATH = "javaExtensions/";
  private Map extensions;

  public ExtensionHandler(SmartFoxServer sf, EventWriter ew)
  {
    this.sf = sf;
    this.eventWriter = ew;
    this.adminZone = SmartFoxServer.getInstance().getZone("$dmn");
  }

  public void setupHandler()
  {
    loadAdminExtension();
  }

  public ISmartFoxExtension loadExtension(String xtName, String className, int xtType)
  {
    ISmartFoxExtension xtObj = null;

    if (xtType == 1)
    {
      try
      {
        File f = new File("javaExtensions/");
        ClassLoader loader = new URLClassLoader(new URL[] { f.toURL() });

        Class cl = loader.loadClass(className);

        if (!(AbstractExtension.class.isAssignableFrom(cl)))
        {
          SmartFoxServer.log.warning("Extension: " + className + " does not extend AbstractExtension.");
        }

        xtObj = (ISmartFoxExtension)cl.newInstance();
      }
      catch (Exception e)
      {
        SmartFoxServer.log.warning("Problems loading extension called: " + className);
      }

    }
    else if (xtType == 2)
    {
      try
      {
        JavascriptExtension jsExt = new JavascriptExtension();

        jsExt.setScriptFile(className);

        xtObj = jsExt;
      }
      catch (Exception jsExt)
      {
        if (e instanceof EvaluatorException)
        {
          logCompileError(xtName, e); break label274:
        }

        SmartFoxServer.log.warning("Problems loading script: " + className);
        System.out.println(e);
        e.printStackTrace();
      }

    }
    else if (xtType == 3)
    {
      try
      {
        PythonExtension pyExt = new PythonExtension();

        pyExt.setScriptFile(className);

        xtObj = pyExt;
      }
      catch (Exception pyExt)
      {
        SmartFoxServer.log.warning("Problems loading python script: " + className);
        System.out.println(e);
        e.printStackTrace();
      }
    }

    label274: return xtObj;
  }

  public void createExtension(String xtName, String className, int xtType, String zoneOwner, int roomOwner)
  {
    createExtension(xtName, className, xtType, zoneOwner, roomOwner, true);
  }

  public void createExtension(String xtName, String className, int xtType, String zoneOwner, int roomOwner, boolean init)
  {
    if ((SmartFoxServer.get_TYP() & 0x1) == 0)
      return;

    ISmartFoxExtension xtObj = loadExtension(xtName, className, xtType);

    Zone z = SmartFoxServer.getInstance().getZone(zoneOwner);
    Room r = null;

    if (xtObj != null)
    {
      if (z != null)
      {
        if (roomOwner == -1)
        {
          z.getExtManager().add(xtName, xtObj);
          SmartFoxServer.log.info("Zone Extension [ " + xtName + " ] created, for zone:" + z.getName());
        }
        else
        {
          r = z.getRoom(roomOwner);

          if (r != null)
          {
            r.getExtManager().add(xtName, xtObj);
            SmartFoxServer.log.info("Room Extension [ " + xtName + " ] created!");
          }
          else {
            SmartFoxServer.log.warning("Could not assign extension [ " + xtName + " ] to Room id= " + roomOwner);
          }
        }
      }
      else { SmartFoxServer.log.warning("Could not assign extension [ " + xtName + " ] to Zone = " + zoneOwner);
      }

      xtObj.registerForEvents(zoneOwner, (r == null) ? null : r.getName());

      if (init)
        xtObj.init();
    }
  }

  private void logCompileError(String extName, Exception e)
  {
    String msg = e.toString();
    String error = e.getMessage();

    int globalLineNumber = -1;
    int localLineNumber = -1;

    int lnPos = msg.lastIndexOf("#");
    try
    {
      if (lnPos == -1)
        break label68;

      ++lnPos;
      lineNumStr = msg.substring(lnPos, msg.lastIndexOf(")"));

      globalLineNumber = Integer.parseInt(lineNumStr);
      localLineNumber = globalLineNumber - JavascriptExtension.MAINLIB_LINES;
    }
    catch (NumberFormatException lineNumStr)
    {
    }
    label68: String atLine = null;

    if (localLineNumber != -1)
    {
      atLine = " Line: " + localLineNumber;
      if (localLineNumber < 1)
        atLine = atLine + " (MainLib line: " + globalLineNumber + ")";
    }
    else {
      atLine = " (Line number not available)";
    }

    StringBuffer errorMessage = new StringBuffer("Error compiling extension [ ");
    errorMessage.append(extName).append(" ]: ").append(error).append(atLine);

    SmartFoxServer.log.warning(errorMessage.toString());
  }

  public ISmartFoxExtension getExtension(String key, String zoneName, int roomId)
    throws MissingExtensionException
  {
    ISmartFoxExtension sfe = null;

    Zone z = SmartFoxServer.getInstance().getZone(zoneName);

    if (z != null)
    {
      sfe = z.getExtManager().get(key);
      if ((sfe == null) && (roomId != -1))
      {
        Room r = z.getRoom(roomId);

        if (r != null)
          sfe = r.getExtManager().get(key);
      }
    }

    if (sfe == null)
      throw new MissingExtensionException("Unrecognized extension name: [ " + key + " ]! Zone: " + zoneName + ", Room: " + roomId);

    return sfe;
  }

  public void run()
  {
    this.running = true;

    label44: while (this.running)
    {
      try
      {
        ServerEvent event;
        if ((event = this.eventQueue.deQueue()) == null)
          break label44;
        processEvent(event);
      }
      catch (InterruptedException e)
      {
        if (!(SmartFoxServer.IS_SHUTTING_DOWN))
          SmartFoxServer.log.warning("Problems during event processing");
      }
    }
  }

  protected void processEvent()
  {
  }

  protected void processEvent(ServerEvent event)
  {
    SocketChannel ch = event.getSenderChannel();

    User u = this.sf.getUserByChannel(ch);

    if (u != null)
    {
      try
      {
        if (event.getIsXml())
        {
          IXMLElement bodyTag = event.getXmlMsg();

          String action = bodyTag.getAttribute("action", "");

          int fromRoom = -1;
          try
          {
            fromRoom = Integer.parseInt(bodyTag.getAttribute("r", ""));
          }
          catch (NumberFormatException nfe)
          {
            SmartFoxServer.log.warning("Extension Call has invalid RoomID. User = " + u.getName() + " - " + u.getChannel().socket().getInetAddress());
            return;
          }

          if ((fromRoom == -1) && (!(ConfigData.ALLOW_ANON_XT_MESSAGES))) {
            return;
          }

          String dataObj = bodyTag.getContent();

          if (action.equals("xtReq"))
          {
            if (dataObj != null)
            {
              String zoneName;
              ActionscriptObject ao = ASObjectSerializer.deserialize(dataObj);

              String xtName = (String)ao.get("name");
              String xtCmd = (String)ao.get("cmd");
              ActionscriptObject param = (ActionscriptObject)ao.get("param");

              if (xtName.equals("$dmn"))
                zoneName = "$dmn";
              else
                zoneName = u.getZone();

              ISmartFoxExtension sfe = getExtension(xtName, zoneName, fromRoom);

              u.updateMessageTime();

              if (!(sfe.isActive())) return;
              sfe.handleRequest(xtCmd, param, u, fromRoom); return;
            }

            SmartFoxServer.log.warning("Extension request arguments = null , action = " + action); return;
          }

          SmartFoxServer.log.warning("Unrecognized ExtensionHandler action: " + action); return;
        }

        if (event.getIsJson())
        {
          JSONObject jso = event.getJsonMsg();

          String xtName = (String)jso.get("x");

          String xtCmd = (String)jso.get("c");

          int fromRoom = jso.getInt("r");

          JSONObject jsoParams = (JSONObject)jso.get("p");

          sfe = getExtension(xtName, u.getZone(), fromRoom);

          u.updateMessageTime();

          if (!(sfe.isActive())) return;
          sfe.handleRequest(xtCmd, jsoParams, u, fromRoom); return;
        }

        String[] params = splitParameters(event.getStrMsg());

        String[] xParams = new String[params.length - 3];

        String xtName = params[0];
        String xtCmd = params[1];
        int fromRoom = Integer.parseInt(params[2]);

        System.arraycopy(params, 3, xParams, 0, params.length - 3);

        ISmartFoxExtension sfe = getExtension(xtName, u.getZone(), fromRoom);

        u.updateMessageTime();

        if (!(sfe.isActive())) return;
        sfe.handleRequest(xtCmd, xParams, u, fromRoom);
      }
      catch (ArrayIndexOutOfBoundsException aiobe)
      {
        SmartFoxServer.log.warning("String message in wrong format: " + aiobe);
      }
      catch (NumberFormatException nfe)
      {
        SmartFoxServer.log.warning("Bad value in String message: " + nfe);
        nfe.printStackTrace();
      }
      catch (MissingExtensionException mee)
      {
        SmartFoxServer.log.warning(mee.toString());
      }
      catch (Exception e)
      {
        SmartFoxServer.log.warning("Exception in ExtensionHandler: " + e);
        System.out.println("Exception type: " + e.getClass().getName());
        e.printStackTrace();
      }
    }
  }

  private void loadAdminExtension()
  {
    Class cl;
    try
    {
      cl = Class.forName("it.gotoandplay.smartfoxserver.extensions.AdminExtension");

      if (!(AbstractExtension.class.isAssignableFrom(cl)))
      {
        SmartFoxServer.log.warning("Admin Extension is not valid! Unloading item!"); return;
      }

      ISmartFoxExtension xtObj = (ISmartFoxExtension)cl.newInstance();

      Zone z = SmartFoxServer.getInstance().getZone("$dmn");
      z.getExtManager().add("$dmn", xtObj);

      xtObj.registerForEvents(null, null);

      xtObj.init();
    }
    catch (Exception e)
    {
      SmartFoxServer.log.severe("Problems loading Admin Extension: " + e);
    }
  }

  private String makeHeader(String type, int room, String action)
  {
    return "<msg t='" + type + "'><body action='" + action + "' r='" + room + "'>";
  }

  private String closeHeader()
  {
    return "</body></msg>";
  }

  private String[] splitParameters(String str)
  {
    return str.split("\\" + ConfigData.STR_DELIMITER);
  }

  public AdminExtension getAdminExtension()
  {
    return ((AdminExtension)this.adminZone.getExtManager().get("$dmn"));
  }

  public void parse_Extensions(IXMLElement node, String zoneName, int roomId)
  {
    Enumeration subNodes = node.enumerateChildren();
    Zone z = SmartFoxServer.getInstance().getZone(zoneName);
    ExtensionManager em = z.getExtManager();

    while (subNodes.hasMoreElements())
    {
      IXMLElement xt = (IXMLElement)subNodes.nextElement();

      if (xt.getName().equalsIgnoreCase("extension"))
      {
        String xtName = xt.getAttribute("name", "");

        if (!(em.contains(xtName)))
        {
          String type = xt.getAttribute("type", "");

          int xtType = 1;

          if (type.equalsIgnoreCase("script")) {
            xtType = 2;
          }
          else if (type.equalsIgnoreCase("python")) {
            xtType = 3;
          }

          String xtClass = xt.getAttribute("className", "");

          createExtension(xtName, xtClass, xtType, zoneName, roomId);
        }
      }
    }
  }
}