package it.gotoandplay.smartfoxserver.controllers;

import it.gotoandplay.smartfoxserver.SmartFoxServer;
import it.gotoandplay.smartfoxserver.data.Room;
import it.gotoandplay.smartfoxserver.data.Zone;
import it.gotoandplay.smartfoxserver.events.IEventDispatcher;
import it.gotoandplay.smartfoxserver.events.IEventListener;
import it.gotoandplay.smartfoxserver.events.InternalEventObject;
import it.gotoandplay.smartfoxserver.extensions.ISmartFoxExtension;
import it.gotoandplay.smartfoxserver.lib.Wrap;
import java.util.ArrayList;

public abstract class MessageHandler extends Wrap
  implements IEventDispatcher
{
  protected SmartFoxServer sf;
  protected ArrayList listeners;

  public final void init(SmartFoxServer sf, int numThreads)
  {
    this.sf = sf;

    this.listeners = new ArrayList();

    initWrap(numThreads);

    setupHandler();
  }

  protected abstract void setupHandler();

  public void addListener(IEventListener o)
  {
    synchronized (this.listeners)
    {
      this.listeners.add(o);
    }
  }

  public void removeListener(IEventListener o)
  {
    synchronized (this.listeners)
    {
      this.listeners.remove(o);
    }
  }

  public void dispatchEvent(InternalEventObject ieo, String ownerZone, String ownerRoom)
  {
    if (this.listeners.size() < 1)
      return;

    Object[] allListeners = (Object[])null;

    synchronized (this.listeners)
    {
      allListeners = this.listeners.toArray();
    }

    if (allListeners != null)
    {
      for (int i = 0; i < allListeners.length; ++i)
      {
        IEventListener target = (IEventListener)allListeners[i];
        ISmartFoxExtension ext = (ISmartFoxExtension)target;

        String zn = target.getOwnerZone();
        String rn = target.getOwnerRoom();

        if ((zn == null) && (rn == null))
        {
          target.handleInternalEvent(ieo);
        }
        else if ((zn != null) && (rn == null) && (ownerZone.equals(zn)) && (ext.isActive()))
        {
          target.handleInternalEvent(ieo);
        }
        else if ((zn != null) && (rn != null) && (ownerRoom != null) && (ext.isActive()) && 
          (ownerZone.equals(zn)) && (ownerRoom.equals(rn)))
        {
          target.handleInternalEvent(ieo);
        }
      }
    }
  }

  public void dispatchEvent(InternalEventObject ieo, String ownerZone, String dummy, int[] roomIds)
  {
    if (this.listeners.size() < 1)
      return;

    Object[] allListeners = (Object[])null;

    synchronized (this.listeners)
    {
      allListeners = this.listeners.toArray();
    }

    if (allListeners != null)
    {
      for (int i = 0; i < allListeners.length; ++i)
      {
        IEventListener target = (IEventListener)allListeners[i];
        ISmartFoxExtension ext = (ISmartFoxExtension)target;

        String zn = target.getOwnerZone();
        String rn = target.getOwnerRoom();

        if ((zn == null) && (ext.isActive()))
        {
          target.handleInternalEvent(ieo);
        }
        else if ((zn != null) && (ownerZone.equals(zn)) && (ext.isActive()))
        {
          target.handleInternalEvent(ieo);
        }
        else if ((zn != null) && (rn != null) && (ownerZone.equals(zn)) && (ext.isActive()))
        {
          Zone z = SmartFoxServer.getInstance().getZone(zn);

          if (z != null)
          {
            for (int n = 0; n < roomIds.length; ++n)
            {
              Room room = z.getRoom(roomIds[n]);

              if ((room != null) && 
                (rn.equals(room.getName())))
              {
                target.handleInternalEvent(ieo);
              }
            }
          }
        }
      }
    }
  }
}