package it.gotoandplay.smartfoxserver.controllers;

import it.gotoandplay.smartfoxserver.EventWriter;
import it.gotoandplay.smartfoxserver.SmartFoxServer;
import it.gotoandplay.smartfoxserver.admin.KickUser;
import it.gotoandplay.smartfoxserver.config.ConfigData;
import it.gotoandplay.smartfoxserver.crypto.SecretKeyGenerator;
import it.gotoandplay.smartfoxserver.data.Room;
import it.gotoandplay.smartfoxserver.data.User;
import it.gotoandplay.smartfoxserver.data.Zone;
import it.gotoandplay.smartfoxserver.data.buddylist.BuddyListManager;
import it.gotoandplay.smartfoxserver.data.buddylist.LoadableBuddyItem;
import it.gotoandplay.smartfoxserver.data.buddylist.LoadableBuddyList;
import it.gotoandplay.smartfoxserver.data.buddylist.PermissionStatus;
import it.gotoandplay.smartfoxserver.events.InternalEventObject;
import it.gotoandplay.smartfoxserver.exceptions.BuddyListException;
import it.gotoandplay.smartfoxserver.exceptions.CreateRoomException;
import it.gotoandplay.smartfoxserver.exceptions.JoinRoomException;
import it.gotoandplay.smartfoxserver.exceptions.LoginException;
import it.gotoandplay.smartfoxserver.extensions.AdminExtension;
import it.gotoandplay.smartfoxserver.lib.ASObjectSerializer;
import it.gotoandplay.smartfoxserver.lib.ActionscriptObject;
import it.gotoandplay.smartfoxserver.lib.BadWordsFilter;
import it.gotoandplay.smartfoxserver.lib.EventQueue;
import it.gotoandplay.smartfoxserver.lib.ServerEvent;
import it.gotoandplay.smartfoxserver.lib.SmartFoxLib;
import it.gotoandplay.smartfoxserver.util.Attachment;
import java.net.InetAddress;
import java.net.Socket;
import java.nio.channels.SelectionKey;
import java.nio.channels.SocketChannel;
import java.util.Enumeration;
import java.util.HashMap;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;
import java.util.logging.Logger;
import net.n3.nanoxml.IXMLElement;
import net.n3.nanoxml.IXMLParser;
import net.n3.nanoxml.IXMLReader;
import net.n3.nanoxml.StdXMLReader;
import net.n3.nanoxml.XMLParserFactory;

public class SystemHandler extends MessageHandler
{
  public static final String RINFO = "1E324DE657963702112756F06E7267169EC6562A6FD637AB3011F9F0522B96464CA2496B00E0FCA81F190A299D0F4DFA42AC1D81E1";
  public static final String EVENT_JOIN = "onJoinRoom";
  public static final String EVENT_USER_ENTER_ROOM = "onUserEnterRoom";
  public static final String EVENT_ROOM_ADDED = "onRoomAdded";
  public static final String EVENT_ROOM_DELETED = "onRoomDeleted";
  public static final String EVENT_SPECTATOR_SWITCHED = "onSpectatorSwitched";
  public static final String EVENT_USER_COUNT_CHANGE = "onUserCountChange";
  public static final String EVENT_USER_LEAVE_ROOM = "onUserLeaveRoom";
  public static final String EVENT_ROOM_LEFT = "onRoomLeft";
  public static final String EVENT_CREATE_ROOM_ERROR = "onRoomCreateError";
  public static final String EVENT_JOIN_ROOM_ERROR = "onJoinRoomError";
  public static final String EVENT_ROUND_TRIP = "onRoundTripResponse";
  public static final String EVENT_LOGOUT = "onLogout";
  public static final String MESSAGE_TYPE_MOD = "mod";
  public static final String MESSAGE_TYPE_DMN = "dmn";
  public static final String ACTION_JOIN = "joinRoom";
  public static final String ACTION_AUTOJOIN = "autoJoin";
  public static final String ACTION_ADD_BUDDY = "addBuddy";
  public static final String ACTION_CLEAR_BUDDYLIST = "clearBuddyList";
  public static final String ACTION_CREATE_ROOM = "createRoom";
  public static final String ACTION_GET_ROOMLIST = "getRoomList";
  public static final String ACTION_LEAVE_ROOM = "leaveRoom";
  public static final String ACTION_LOAD_BUDDYLIST = "loadBuddyList";
  public static final String ACTION_REMOVE_BUDDY = "removeBuddy";
  public static final String ACTION_ROUNDTRIP = "roundTripBench";
  public static final String ACTION_SEND_OBJECT = "sendObject";
  public static final String ACTION_SEND_OBJ2GROUP = "sendObjectToGroup";
  public static final String ACTION_SEND_PUBLIC_MSG = "sendPublicMessage";
  public static final String ACTION_SEND_PRIVATE_MSG = "sendPrivateMessage";
  public static final String ACTION_SET_ROOM_VARS = "setRoomVariables";
  public static final String ACTION_SET_USER_VARS = "setUserVariables";
  public static final String ACTION_SWITCH_SPECTATOR = "switchSpectator";
  public static final String ACTION_GET_BUDDY_ROOM = "getBuddyRoom";
  private SmartFoxServer sfs;
  private EventWriter eventWriter;

  public SystemHandler(SmartFoxServer sf, EventWriter ew)
  {
    this.sfs = sf;
    this.eventWriter = ew;
  }

  public void setupHandler()
  {
  }

  public void run()
  {
    this.running = true;

    label57: while (this.running)
    {
      try
      {
        ServerEvent event;
        if ((event = this.eventQueue.deQueue()) == null)
          break label57;
        processEvent(event);
      }
      catch (InterruptedException e)
      {
        if (!(SmartFoxServer.IS_SHUTTING_DOWN))
          SmartFoxServer.log.warning("Problems during event processing > " + e);
      }
    }
  }

  protected void processEvent()
  {
  }

  protected void processEvent(ServerEvent event)
  {
    String action = null;
    try
    {
      IXMLElement bodyTag = event.getXmlMsg();

      SocketChannel ch = event.getSenderChannel();

      User u = SmartFoxServer.getInstance().getUserByChannel(ch);
      Zone z = null;

      if (u != null)
      {
        z = SmartFoxServer.getInstance().getZone(u.getZone());
        if (!(z.isActive()))
          return;

      }

      action = bodyTag.getAttribute("action", "");
      int fromRoom = Integer.parseInt(bodyTag.getAttribute("r", ""));

      if (action.equals("login"))
      {
        handleLoginRequest(bodyTag, event.getSenderChannel()); return;
      }

      if ((action.equals("getRmList")) && (isValidRequest(z, "getRoomList")))
      {
        handleGetRoomList(event.getSenderChannel()); return;
      }

      if ((action.equals("autoJoin")) && (isValidRequest(z, "autoJoin")))
      {
        autoJoinDefaultRoom(event.getSenderChannel()); return;
      }

      if (action.equals("delUser"))
      {
        notifyUserRemoval(bodyTag, fromRoom); return;
      }

      if ((action.equals("pubMsg")) && (isValidRequest(z, "sendPublicMessage")))
      {
        handlePublicMessage(bodyTag, fromRoom, event.getSenderChannel()); return;
      }

      if ((action.equals("prvMsg")) && (isValidRequest(z, "sendPrivateMessage")))
      {
        handlePrivateMessage(bodyTag, fromRoom, event.getSenderChannel()); return;
      }

      if (action.equals("modMsg"))
      {
        handleModMessage(bodyTag, fromRoom, event.getSenderChannel()); return;
      }

      if ((action.equals("joinRoom")) && (isValidRequest(z, "joinRoom")))
      {
        handleJoinRequest(bodyTag, fromRoom, event.getSenderChannel()); return;
      }

      if ((action.equals("asObj")) && (isValidRequest(z, "sendObject")))
      {
        handleASObject(bodyTag, fromRoom, event.getSenderChannel()); return;
      }

      if ((action.equals("asObjG")) && (isValidRequest(z, "sendObjectToGroup")))
      {
        handleASObjectToGroup(bodyTag, fromRoom, event.getSenderChannel()); return;
      }

      if ((action.equals("setUvars")) && (isValidRequest(z, "setUserVariables")))
      {
        handleSetUserVariables(bodyTag, fromRoom, event.getSenderChannel()); return;
      }

      if ((action.equals("setRvars")) && (isValidRequest(z, "setRoomVariables")))
      {
        handleSetRoomVariables(bodyTag, fromRoom, event.getSenderChannel(), false); return;
      }

      if ((action.equals("createRoom")) && (isValidRequest(z, "createRoom")))
      {
        handleCreateRoom(bodyTag, fromRoom, event.getSenderChannel()); return;
      }
      if ((action.equals("leaveRoom")) && (isValidRequest(z, "leaveRoom")))
      {
        handleLeaveRoom(bodyTag, fromRoom, event.getSenderChannel()); return;
      }

      if ((action.equals("roundTrip")) && (isValidRequest(z, "roundTripBench")))
      {
        handleRoundTripRequest(fromRoom, event.getSenderChannel()); return;
      }
      if (action.equals("verChk"))
      {
        check_API_version(bodyTag, event.getSenderChannel()); return;
      }
      if ((action.equals("swSpec")) && (isValidRequest(z, "switchSpectator")))
      {
        handleSpectatorSwitch(fromRoom, event.getSenderChannel()); return;
      }
      if ((action.equals("loadB")) && (isValidRequest(z, "loadBuddyList")))
      {
        handleLoadBuddyList(event.getSenderChannel()); return;
      }
      if ((action.equals("addB")) && (isValidRequest(z, "addBuddy")))
      {
        handleAddBuddy(bodyTag, event.getSenderChannel()); return;
      }
      if ((action.equals("remB")) && (isValidRequest(z, "removeBuddy")))
      {
        handleRemoveBuddy(bodyTag, event.getSenderChannel()); return;
      }
      if ((action.equals("clearB")) && (isValidRequest(z, "clearBuddyList")))
      {
        handleClearBuddyList(event.getSenderChannel()); return;
      }
      if ((action.equals("roomB")) && (isValidRequest(z, "getBuddyRoom")))
      {
        handleGetBuddyRoom(bodyTag, event.getSenderChannel()); return;
      }
      if (action.equals("rndK"))
      {
        handleRandomKeyRequest(event.getSenderChannel()); return;
      }
      if (action.equals("setBvars"))
      {
        handleSetBuddyVars(bodyTag, event.getSenderChannel()); return;
      }
      if (action.equals("setB"))
      {
        handleSetBuddyBlockStatus(bodyTag, event.getSenderChannel()); return;
      }
      if (action.equals("zInfo"))
      {
        handleZoneInfo(bodyTag, event.getSenderChannel()); return;
      }
      if (action.equals("logout"))
      {
        handleLogout(event.getSenderChannel(), true, true); return;
      }
      if (action.equals("bPrm"))
      {
        handleBuddyPermission(bodyTag, event.getSenderChannel()); return;
      }

      StringBuffer sb = new StringBuffer("SystemHandler action refused: ");
      sb.append(action).append(", from: ");
      Socket socket = event.getSenderChannel().socket();

      if ((socket != null) && (socket.isConnected()))
        sb.append(event.getSenderChannel().socket().getInetAddress().getHostAddress());
      else
        sb.append(" - socket channel closed / not available - ");

      SmartFoxServer.log.fine(sb.toString());
    }
    catch (NumberFormatException nfe)
    {
      SmartFoxServer.log.warning("SysHandler -> Bad room id. Action: " + action + " >> " + nfe);
    }
    catch (Exception e)
    {
      SmartFoxServer.log.warning("Unexpected Exception: " + e);
      e.printStackTrace();
    }
  }

  private boolean isValidRequest(Zone z, String action)
  {
    return ((z != null) && (!(z.isSysActionDisabled(action))));
  }

  private void handleLoginRequest(IXMLElement xml, SocketChannel ch)
  {
    SelectionKey sk;
    try
    {
      StringBuffer response;
      LinkedList recipients;
      sk = ch.keyFor(SmartFoxServer.getInstance().getReadSelector());
      Attachment attachment = (Attachment)sk.attachment();
      attachment.blueBoxed = (xml.getAttribute("bb", "0").equals("1"));
      attachment.bbSessionId = xml.getAttribute("ssid", null);

      xml = xml.getFirstChildNamed("login");

      String zone = xml.getAttribute("z", "");
      String nick = xml.getFirstChildNamed("nick").getContent();
      String pass = xml.getFirstChildNamed("pword").getContent();

      if (pass == null) {
        pass = "";
      }

      Zone z = SmartFoxServer.getInstance().getZone(zone);

      if (z == null) return;

      int nameMaxLen = z.getMaxUserNameLen();

      nick = SmartFoxLib.cleanUpString(nick, z.userNameAvoidChars, nameMaxLen);

      if (SmartFoxServer.getInstance().isBanned(nick, ch))
      {
        response = makeHeader("sys", 0, "logKO");
        response.append("<login e='").append(ConfigData.BANNED_LOGIN_MESSAGE).append("'/>");
        response.append(closeHeader());

        recipients = new LinkedList();
        recipients.add(ch);

        this.eventWriter.handleEvent(new ServerEvent(response.toString(), ch, recipients));
        return;
      }

      if (!(z.isThereRoom()))
      {
        response = makeHeader("sys", 0, "logKO");
        response.append("<login e='").append("Sorry, this zone is full!").append("'/>");
        response.append(closeHeader());

        recipients = new LinkedList();
        recipients.add(ch);

        this.eventWriter.handleEvent(new ServerEvent(response.toString(), ch, recipients));
        return;
      }

      if (z.hasCustomLogin())
      {
        InternalEventObject ioe = new InternalEventObject("loginRequest");

        ioe.addParam("zone", zone);
        ioe.addParam("nick", nick);
        ioe.addParam("pass", pass);
        ioe.addObject("chan", ch);

        dispatchEvent(ioe, zone, null); return;
      }

      defaultLogin(nick, pass, zone, ch);
    }
    catch (NullPointerException npe)
    {
      npe.printStackTrace();
      SmartFoxServer.log.warning("SystemHandler: Invalid login request > " + npe);
    }
  }

  public void handleLogout(SocketChannel ch, boolean fireClientEvt, boolean fireInternalEvt)
  {
    User requester = this.sfs.getUserByChannel(ch);

    if (requester != null)
    {
      Zone z = this.sfs.getZone(requester.getZone());

      this.sfs.logOutUser(requester, fireInternalEvt);
      StringBuffer response = makeHeader("sys", 0, "logout").append(closeHeader());

      if ((!(z.isEventDisabled("onLogout"))) && (fireClientEvt))
      {
        LinkedList recipients = new LinkedList();
        recipients.add(ch);
        this.eventWriter.handleEvent(new ServerEvent(response.toString(), ch, recipients));
      }
    }
  }

  private void check_API_version(IXMLElement node, SocketChannel ch)
  {
    node = node.getFirstChildNamed("ver");
    try
    {
      StringBuffer response;
      if (node == null) return;

      int apiVersion = Integer.parseInt(node.getAttribute("v", ""));

      if (apiVersion >= ConfigData.apiMinVersion)
      {
        response = makeHeader("sys", 0, "apiOK").append(closeHeader());
      }
      else
      {
        response = makeHeader("sys", 0, "apiKO").append(closeHeader());

        SmartFoxServer.log.warning("Client from ip: " + 
          ch.socket().getInetAddress().toString() + 
          " was refused because API are obsolete.");
      }

      LinkedList recipients = new LinkedList();
      recipients.add(ch);
      this.eventWriter.handleEvent(new ServerEvent(response.toString(), ch, recipients));
    }
    catch (Exception e)
    {
      SmartFoxServer.log.warning("Client from ip: " + 
        getUserIp(ch) + 
        " is trying to connect with an unrecognized API version");
    }
  }

  public void defaultLogin(String nick, String pass, String zone, SocketChannel ch)
  {
    StringBuffer response;
    boolean success = false;

    Zone z = SmartFoxServer.getInstance().getZone(zone);
    try
    {
      if ((!(z.getEmptyNames())) && (nick.equals("")))
      {
        throw new LoginException("You should provide a non empty user name!");
      }

      if (nick.equals(""))
      {
        nick = this.sfs.getGuestName();
      }

      User me = this.sfs.canLogin(nick, pass, ch, zone);

      response = makeHeader("sys", 0, "logOK");
      response.append("<login n='").append(nick).append("' id='").append(
        me.getUserId()).append("' mod='").append(
        (me.isModerator()) ? "1" : "0").append("'/>");
      response.append(closeHeader());
      success = true;
    }
    catch (LoginException le)
    {
      response = makeHeader("sys", 0, "logKO");
      response.append("<login e='").append(le.getMessage()).append("'/>");
      response.append(closeHeader());
    }

    LinkedList recipients = new LinkedList();
    recipients.add(ch);

    this.eventWriter.handleEvent(
      new ServerEvent(response.toString(), ch, 
      recipients));
  }

  public void requestRoomList(SocketChannel sc)
  {
    handleGetRoomList(sc);
  }

  private void handleGetRoomList(SocketChannel ch)
  {
    User u = this.sfs.getUserByChannel(ch);

    if (u != null)
    {
      Zone z = this.sfs.getZone(u.getZone());

      LinkedList rl = z.getRoomList();

      Iterator it = rl.iterator();

      StringBuffer response = makeHeader("sys", 0, "rmList");
      response.append("<rmList>");

      while (it.hasNext())
      {
        Room r = (Room)it.next();

        response.append("<rm id='").append(r.getId());
        response.append("' priv='").append((r.isPrivate()) ? "1" : "0");
        response.append("' temp='").append((r.isTemp()) ? "1" : "0");
        response.append("' game='").append((r.isGame()) ? "1" : "0");
        response.append("' ucnt='").append(r.getUserCount());

        if (r.isLimbo()) {
          response.append("' lmb='1");
        }

        if (r.isGame())
          response.append("' scnt='").append(r.getSpectatorCount());

        response.append("' maxu='").append(r.getMaxUsers());
        response.append("' maxs='").append(r.getMaxSpectator());
        response.append("'>");

        response.append("<n><![CDATA[").append(r.getName()).append("]]></n>");

        if (z.getVarsOnRoomList()) {
          response.append(r.getXmlVarList());
        }

        response.append("</rm>");
      }

      response.append("</rmList>").append(closeHeader());

      LinkedList recipients = new LinkedList();
      recipients.add(ch);

      this.eventWriter.handleEvent(new ServerEvent(response.toString(), ch, recipients));
    }
    else
    {
      SmartFoxServer.log.warning("Room list was requested by a non logged user! IP: " + getUserIp(ch));
    }
  }

  private void handleJoinRequest(IXMLElement xml, int fromRoom, SocketChannel ch)
  {
    xml = xml.getFirstChildNamed("room");

    User u = this.sfs.getUserByChannel(ch);

    if ((u != null) && (xml != null))
    {
      try
      {
        boolean isSpectator = false;
        int newRoom = Integer.parseInt(xml.getAttribute("id", ""));
        String spec = xml.getAttribute("spec", "");

        if (spec.equals("1"))
          isSpectator = true;

        int leaveRoom = Integer.parseInt(xml.getAttribute("leave", ""));
        int oldRoom = Integer.parseInt(xml.getAttribute("old", ""));
        String pword = xml.getAttribute("pwd", "");

        Zone z = this.sfs.getZone(u.getZone());

        Room r = z.getRoom(oldRoom);

        int oldRoomPlayerIndex = 0;

        if (r != null) {
          oldRoomPlayerIndex = u.getPlayerIndex(r);
        }

        startJoinTransaction(u, 
          r, 
          z, 
          newRoom, 
          leaveRoom, 
          oldRoomPlayerIndex, 
          pword, 
          isSpectator, 
          true);
      }
      catch (NumberFormatException nfe)
      {
        SmartFoxServer.log.warning("Bad arguments in JoinRoom Request > " + u.getName());
      }
      catch (Exception e)
      {
        e.printStackTrace();
        SmartFoxServer.log.warning("Generic exception in JoinRoom Request > " + u.getName() + " - " + e);
      }

    }
    else
      SmartFoxServer.log.warning("A non logged user is requesting a joinRoom " + getUserIp(ch));
  }

  public boolean startJoinTransaction(User u, Room r, Zone z, int newRoom, int leaveRoom, int oldRoomPlayerIndex, String pword, boolean isSpectator, boolean broadcast)
  {
    boolean success = joinRoom(newRoom, u.getChannel(), pword, isSpectator, broadcast);

    if (success)
    {
      if ((leaveRoom > 0) && 
        (r != null))
      {
        r.removeUser(u, true, true);

        broadcastUserRemoved(u, r, oldRoomPlayerIndex);

        checkEmptyRoom(r, u);
      }

      Room targetRoom = z.getRoom(newRoom);
      if (targetRoom != null) {
        targetRoom.updateUserCount();
      }

    }

    return success;
  }

  private void checkEmptyRoom(Room r, User u)
  {
    if ((r.howManyUsers() < 1) && (r.isTemp()))
    {
      if (!(r.isGame()))
      {
        u = this.sfs.getUserByChannel(r.getCreator());

        if (u == null)
        {
          SmartFoxServer.getInstance().getZone(r.getZone()).removeRoom(r, false);
          SmartFoxServer.log.info("Dynamic Room Deleted: " + r.getName());
        }

      }
      else
      {
        SmartFoxServer.getInstance().getZone(r.getZone()).removeRoom(r, false);
        SmartFoxServer.log.info("Game Room Deleted: " + r.getName());
      }
    }
  }

  private boolean joinRoom(int roomId, SocketChannel ch, String password, boolean isSpectator, boolean broadcast)
  {
    StringBuffer response = new StringBuffer();

    boolean success = false;

    User u = this.sfs.getUserByChannel(ch);

    if (u != null)
    {
      Zone z = this.sfs.getZone(u.getZone());

      Room r = z.getRoom(roomId);
      try
      {
        if (r == null)
        {
          SmartFoxServer.log.warning("User trying to join non existent room (" + 
            roomId + 
            ")! User: " + 
            u.getName() + 
            ", IP: " + 
            getUserIp(u.getChannel()));

          throw new JoinRoomException("Room does not exist or was just deleted (id=" + roomId + ")");
        }

        r.join(u, password, isSpectator);

        success = true;

        if (!(broadcast))
          break label581;

        response = makeHeader("sys", roomId, "joinOK");

        response.append("<pid id='").append(u.getPlayerIndex(r)).append("'/>");

        response.append(r.getXmlVarList());

        response.append("<uLs r='").append(roomId).append("'>");

        if (!(r.isLimbo()))
          response.append(r.getXmlUserList());

        response.append("</uLs>");

        if ((z.isEventDisabled("onUserEnterRoom")) || (r.isLimbo()))
          break label581;
        StringBuffer broadcastMsg = makeHeader("sys", roomId, "uER");
        broadcastMsg.append("<u i ='").append(u.getUserId());
        broadcastMsg.append("' m='").append((u.isModerator()) ? "1" : "0");

        if (r.isGame())
        {
          broadcastMsg.append("' s='").append((u.isSpectator()) ? "1" : "0");
          broadcastMsg.append("' p='").append(u.getPlayerIndex());
        }

        broadcastMsg.append("'>");
        broadcastMsg.append("<n><![CDATA[").append(u.getName()).append("]]></n>");
        broadcastMsg.append(u.getXmlVarList()).append("</u>");
        broadcastMsg.append(closeHeader());

        LinkedList allButMe = r.getChannellList();
        allButMe.remove(ch);

        this.eventWriter.handleEvent(new ServerEvent(broadcastMsg.toString(), ch, allButMe));
      }
      catch (JoinRoomException jre)
      {
        if (broadcast)
        {
          response = makeHeader("sys", 0, "joinKO");
          response.append("<error msg='").append(jre.getMessage()).append("' />");
        }

        SmartFoxServer.log.info("Room Join Error: " + 
          jre.getMessage() + ", User: " + u.getName() + 
          ", IP: " + u.getIpAddress());
      }

      if (broadcast)
      {
        label581: response.append(closeHeader());

        LinkedList recipients = new LinkedList();
        recipients.add(ch);

        if (!(z.isEventDisabled("onJoinRoom")))
          this.eventWriter.handleEvent(new ServerEvent(response.toString(), ch, recipients));
      }

      if (success)
      {
        InternalEventObject ioe = new InternalEventObject("userJoin");
        ioe.addParam("zone", u.getZone());
        ioe.addObject("user", u);
        ioe.addObject("room", r);

        ioe.addParam("uid", String.valueOf(u.getUserId()));
        ioe.addParam("roomName", r.getName());
        ioe.addParam("roomId", String.valueOf(r.getId()));
        ioe.addParam("roomIsGame", String.valueOf(r.isGame()));

        dispatchEvent(ioe, u.getZone(), r.getName());
      }
    }
    else
    {
      SmartFoxServer.log.warning("Non logged user trying to join a room. IP:" + getUserIp(ch));
    }

    return success;
  }

  private void autoJoinDefaultRoom(SocketChannel ch)
  {
    User u = this.sfs.getUserByChannel(ch);

    if (u != null)
    {
      Zone z = this.sfs.getZone(u.getZone());

      int autoId = z.getAutoJoinRoom();

      if (autoId > -1)
      {
        joinRoom(autoId, ch, "", false, true);
      }

      Room r = z.getRoom(autoId);

      if (r != null)
        r.updateUserCount();

    }
    else
    {
      SmartFoxServer.log.warning("Non logged user requested autojoin. IP: " + getUserIp(ch));
    }
  }

  private void notifyUserRemoval(IXMLElement xml, int roomId)
  {
    xml = xml.getFirstChildNamed("user");
    try
    {
      int uid = Integer.parseInt(xml.getAttribute("id", ""));

      Zone z = this.sfs.getZone(xml.getAttribute("z", ""));

      Room r = z.getRoom(roomId);

      if ((r == null) || (r.isLimbo()) || (z.isEventDisabled("onUserLeaveRoom")) || 
        (r.howManyUsers() <= 0)) {
        return;
      }

      StringBuffer response = makeHeader("sys", roomId, "userGone");
      response.append("<user id='" + uid + "' />").append(closeHeader());

      this.eventWriter.handleEvent(new ServerEvent(response.toString(), null, r.getChannellList()));
    }
    catch (NumberFormatException nfe)
    {
      SmartFoxServer.log.warning("Error occurred while removing user, bad usrID = " + xml.getAttribute("id", ""));
    }
  }

  public void notifyUserRemoval(int uid, int roomId, String zoneName)
  {
    Zone z;
    try
    {
      z = this.sfs.getZone(zoneName);

      Room r = z.getRoom(roomId);

      if ((r == null) || (r.isLimbo()) || (z.isEventDisabled("onUserLeaveRoom")) || 
        (r.howManyUsers() <= 0)) {
        return;
      }

      StringBuffer response = makeHeader("sys", roomId, "userGone");
      response.append("<user id='").append(uid).append("' />").append(closeHeader());

      this.eventWriter.handleEvent(new ServerEvent(response.toString(), null, r.getChannellList()));
    }
    catch (NumberFormatException nfe)
    {
      SmartFoxServer.log.warning("Error occurred while removing user, bad usrID");
    }
  }

  public void notifyRoomDeleted(Zone z, int roomId)
  {
    if (!(z.isEventDisabled("onRoomDeleted")))
    {
      StringBuffer sb = new StringBuffer("<msg t='sys'><body action='roomDel'><rm id='");
      sb.append(roomId).append("'/></body></msg>");

      this.eventWriter.handleEvent(new ServerEvent(sb.toString(), null, z.getAllUsersInZone()));
    }
  }

  private void handlePublicMessage(IXMLElement xml, int roomId, SocketChannel ch)
  {
    xml = xml.getFirstChildNamed("txt");
    String txtmsg = xml.getContent();

    User u = this.sfs.getUserByChannel(ch);

    if ((u != null) && (txtmsg != null))
    {
      Zone z = this.sfs.getZone(u.getZone());

      Room r = z.getRoom(roomId);

      if ((r != null) && (r.isLimbo())) {
        return;
      }

      if (!(u.isInRoom(roomId))) {
        return;
      }

      if (ConfigData.ANTIFLOOD_ACTIVE)
        applyFloodFilter(txtmsg, u);
      else {
        u.updateMessageTime();
      }

      if (ConfigData.BADWORDS_ACTIVE)
      {
        txtmsg = applyBadWordFilter(txtmsg, u);

        if (txtmsg.equals(""))
          return;

      }

      if (r != null)
      {
        if (z.isPubMsgInternalEventEnabled())
          broadcastPublicMessage(z, r, u, txtmsg);
        else {
          dispatchPublicMessage(txtmsg, r, u);
        }

      }
      else
      {
        SmartFoxServer.log.warning("Public Message Request for non-existent room. User: " + u.getName() + ", IP: " + getUserIp(u.getChannel()));
      }

    }
    else
    {
      SmartFoxServer.log.warning("Non logged user trying to send public message. IP: " + getUserIp(ch));
    }
  }

  public void dispatchPublicMessage(String msg, Room r, User u)
  {
    StringBuffer response = makeHeader("sys", r.getId(), "pubMsg");
    response.append("<user id='").append(u.getUserId()).append("' />");
    response.append("<txt><![CDATA[").append(msg).append("]]></txt>").append(closeHeader());

    this.eventWriter.handleEvent(new ServerEvent(response.toString(), u.getChannel(), r.getChannellList()));
  }

  private void handlePrivateMessage(IXMLElement xml, int roomId, SocketChannel ch)
  {
    xml = xml.getFirstChildNamed("txt");
    int recipientId = Integer.parseInt(xml.getAttribute("rcp", "-1"));

    if (recipientId == -1)
    {
      SmartFoxServer.log.warning("Bad recipient ID for private message! IP: " + getUserIp(ch));
      return;
    }

    String txtmsg = xml.getContent();

    User u = this.sfs.getUserByChannel(ch);

    User rec = this.sfs.getUserById(new Integer(recipientId));

    if ((u != null) && (txtmsg != null))
    {
      Zone currZone = this.sfs.getZone(u.getZone());

      if ((currZone.getBuddyManager() != null) && (currZone.getBuddyManager().isAdvancedMode()))
      {
        boolean validRecipient = currZone.getBuddyManager().validateRecipient(u, rec);

        if (!(validRecipient))
        {
          SmartFoxServer.log.info("Private message discarded. User: " + rec.getName() + " is blocked in " + u.getName() + " buddy list.");
          return;
        }

      }

      if (ConfigData.ANTIFLOOD_ACTIVE) {
        applyFloodFilter(txtmsg, u);
      }
      else {
        u.updateMessageTime();
      }

      if (ConfigData.BADWORDS_ACTIVE)
      {
        txtmsg = applyBadWordFilter(txtmsg, u);

        if (txtmsg.equals(""))
          return;

      }

      Zone z = this.sfs.getZone(u.getZone());

      Room r = z.getRoom(roomId);

      if (r != null)
      {
        if (z.isPrivMsgInternalEventEnabled())
        {
          broadcastPrivateMessage(z, r, u, rec, txtmsg);
        }
        else
        {
          if ((rec == null) && (ConfigData.PM_MISSING_RECIPIENT_ALERT))
          {
            StringBuffer response = makeHeader("sys", roomId, "prvMsg");
            response.append("<user id='-1' />");
            response.append("<txt><![CDATA[-1|Private message missing recipient with id:" + recipientId + "]]></txt>").append(closeHeader());

            LinkedList ll = new LinkedList();
            ll.add(u.getChannel());

            this.eventWriter.handleEvent(new ServerEvent(response.toString(), null, ll));
            return;
          }

          dispatchPrivateMessage(r, u, rec, txtmsg);
        }

      }
      else
      {
        SmartFoxServer.log
          .warning("Private Message Request from non-existent room. User: " + 
          u.getName() + ", IP: " + getUserIp(u.getChannel()));
      }

    }
    else
    {
      SmartFoxServer.log
        .warning("Non logged user trying to send a private message. IP: " + getUserIp(ch));
    }
  }

  public void dispatchPrivateMessage(Room room, User sender, User recipient, String message)
  {
    StringBuffer response = makeHeader("sys", room.getId(), "prvMsg");
    response.append("<user id='").append(sender.getUserId()).append("' />");
    response.append("<txt><![CDATA[").append(message).append("]]></txt>").append(closeHeader());

    SocketChannel recipientChannel = sender.getChannel();

    LinkedList ll = new LinkedList();

    if (recipient != null)
      ll.add(recipient.getChannel());

    ll.add(sender.getChannel());

    this.eventWriter.handleEvent(new ServerEvent(response.toString(), recipientChannel, ll));
  }

  private void handleModMessage(IXMLElement xml, int roomId, SocketChannel ch)
  {
    User u = SmartFoxServer.getInstance().getUserByChannel(ch);
    Zone z = SmartFoxServer.getInstance().getZone(u.getZone());

    xml = xml.getFirstChildNamed("txt");

    String type = xml.getAttribute("t", "r");

    String id = xml.getAttribute("id", "");

    String message = xml.getContent();

    if ((message != null) && (message.length() > 0))
    {
      if (type.equals("z"))
      {
        sendAdminMessage(u, roomId, message, z, "mod");
      }
      else
      {
        try
        {
          int theId = Integer.parseInt(id);

          if (type.equals("r"))
          {
            Room r = z.getRoom(theId);

            if (r == null) return;
            sendAdminMessage(u, roomId, message, r, "mod"); return;
          }

          if (!(type.equals("u"))) return;

          User recipient = SmartFoxServer.getInstance().getUserById(new Integer(theId));
          sendAdminMessage(u, roomId, message, recipient, "mod");
        }
        catch (NumberFormatException nfe)
        {
          SmartFoxServer.log.warning("Mod Message failed: bad id");
        }
      }
    }
  }

  public void sendAdminMessage(User sender, int roomId, String message, User recipient, String msgType)
  {
    if ((sender != null) && (recipient != null) && ((
      (sender.isAdmin()) || (sender.isModerator()))))
    {
      LinkedList ll = new LinkedList();
      ll.add(recipient.getChannel());

      sendAdminMessageToList(sender, roomId, message, ll, msgType);
    }
  }

  public void sendAdminMessage(String message, User recipient, String msgType)
  {
    if (recipient != null)
    {
      LinkedList ll = new LinkedList();
      ll.add(recipient.getChannel());

      sendAdminMessageToList(recipient, 0, message, ll, msgType);
    }
  }

  public void sendAdminMessage(User sender, int roomId, String message, Room recipient, String msgType)
  {
    if ((sender.isAdmin()) || (sender.isModerator()))
    {
      LinkedList ll = recipient.getChannellList();
      sendAdminMessageToList(sender, roomId, message, ll, msgType);
    }
  }

  public void sendAdminMessage(User sender, int roomId, String message, Zone recipient, String msgType)
  {
    if ((sender.isAdmin()) || (sender.isModerator()))
    {
      LinkedList ll = recipient.getChannelList();
      sendAdminMessageToList(sender, roomId, message, ll, msgType);
    }
  }

  public void sendAdminMessage(User sender, String message, String msgType)
  {
    if ((sender.isAdmin()) || (sender.isModerator()))
    {
      sendAdminMessageToList(sender, message, 
        SmartFoxServer.getInstance().getGlobalUserList(), msgType);
    }
  }

  private void sendAdminMessageToList(User sender, String message, LinkedList recipients, String type)
  {
    sendAdminMessageToList(sender, 0, message, recipients, type);
  }

  private void sendAdminMessageToList(User sender, int roomId, String message, LinkedList recipients, String type)
  {
    StringBuffer response = makeHeader("sys", roomId, type + "Msg");
    response.append("<user id='").append(sender.getUserId()).append("' />");
    response.append("<txt><![CDATA[").append(message).append("]]></txt>")
      .append(closeHeader());

    this.eventWriter.handleEvent(
      new ServerEvent(response.toString(), 
      null, 
      recipients));
  }

  private void handleASObject(IXMLElement xml, int roomId, SocketChannel ch)
  {
    User u = this.sfs.getUserByChannel(ch);

    if (u != null)
    {
      if (!(u.isInRoom(roomId)))
        return;

      u.updateMessageTime();

      Zone z = this.sfs.getZone(u.getZone());

      Room r = z.getRoom(roomId);

      if (r != null)
      {
        StringBuffer response = makeHeader("sys", roomId, "dataObj");
        response.append("<user id='").append(u.getUserId()).append(
          "' />");
        response.append("<dataObj><![CDATA[").append(xml.getContent())
          .append("]]></dataObj>").append(closeHeader());

        LinkedList allButMe = r.getChannellList();
        allButMe.remove(ch);

        this.eventWriter.handleEvent(new ServerEvent(response.toString(), ch, allButMe));
      }
      else
      {
        SmartFoxServer.log.warning("Object Message Request for non-existent room. IP: " + getUserIp(ch));
      }

    }
    else
    {
      SmartFoxServer.log.warning("Non logged user trying to send an object. IP: " + getUserIp(ch));
    }
  }

  private void handleASObjectToGroup(IXMLElement xml, int roomId, SocketChannel ch)
  {
    User u = this.sfs.getUserByChannel(ch);

    if (u != null)
    {
      if (!(u.isInRoom(roomId)))
        return;

      u.updateMessageTime();

      Zone z = this.sfs.getZone(u.getZone());

      String dataObj = xml.getContent();

      if (dataObj != null)
      {
        ActionscriptObject aso = ASObjectSerializer.deserialize(dataObj);

        String[] toList = aso.getString("_$$_").split(",");

        Object o = aso.removeElement("_$$_");

        dataObj = ASObjectSerializer.serialize(aso);

        LinkedList ll = new LinkedList();

        label157: for (int i = 0; i < toList.length; ++i)
        {
          try
          {
            Integer id = new Integer(toList[i]);
            User usr = SmartFoxServer.getInstance().getUserById(id);

            if (usr == null) break label157;
            ll.add(usr.getChannel());
          }
          catch (NumberFormatException localNumberFormatException)
          {
          }

        }

        if (ll.size() > 0)
        {
          StringBuffer response = makeHeader("sys", roomId, "dataObj");
          response.append("<user id='").append(u.getUserId()).append("' />");
          response.append("<dataObj><![CDATA[").append(dataObj).append("]]></dataObj>").append(closeHeader());

          this.eventWriter.handleEvent(new ServerEvent(response.toString(), ch, ll));
        }

      }

    }
    else
    {
      SmartFoxServer.log.warning("Non logged user is trying to send an object. IP: " + getUserIp(ch));
    }
  }

  private void handleSetUserVariables(IXMLElement xml, int roomId, SocketChannel ch)
  {
    xml = xml.getFirstChildNamed("vars");

    User u = this.sfs.getUserByChannel(ch);

    if (u != null)
    {
      u.updateMessageTime();

      Zone z = this.sfs.getZone(u.getZone());

      StringBuffer response = new StringBuffer("<vars>");

      Enumeration varNodes = xml.enumerateChildren();

      while (varNodes.hasMoreElements())
      {
        IXMLElement var = (IXMLElement)varNodes.nextElement();

        String vName = var.getAttribute("n", "");
        String vType = var.getAttribute("t", "");
        String vValue = var.getContent();

        response.append(setUserVariable(vName, vType, vValue, u));
      }

      response.append("</vars>");

      int[] roomIds = u.getRoomsConnected();

      for (int i = 0; i < roomIds.length; ++i)
      {
        Room r = z.getRoom(roomIds[i]);

        if ((r != null) && 
          (!(r.isLimbo())))
        {
          LinkedList allButMe = r.getChannellList();
          allButMe.remove(u.getChannel());

          StringBuffer message = makeHeader("sys", roomIds[i], "uVarsUpdate");
          message.append("<user id='").append(u.getUserId()).append("' />");
          message.append(response).append(closeHeader());

          this.eventWriter.handleEvent(new ServerEvent(message.toString(), u.getChannel(), allButMe));
        }
      }
    }
  }

  private void handleSetBuddyVars(IXMLElement xml, SocketChannel sc)
  {
    xml = xml.getFirstChildNamed("vars");

    User u = this.sfs.getUserByChannel(sc);

    if (u != null) {
      u.updateMessageTime();
    }

    Enumeration varNodes = xml.enumerateChildren();
    Map vars = new HashMap();

    while (varNodes.hasMoreElements())
    {
      IXMLElement var = (IXMLElement)varNodes.nextElement();
      vars.put(var.getAttribute("n", ""), var.getContent());
    }

    doSetBuddyVars(u, vars);
  }

  public void doSetBuddyVars(User user, Map<String, String> vars)
  {
    if (user != null)
    {
      Zone z = SmartFoxServer.getInstance().getZone(user.getZone());

      if (z.hasBuddyList())
      {
        user.updateMessageTime();

        List changedVars = z.getBuddyManager().setBuddyVariables(user, vars);

        if (changedVars.size() > 0)
        {
          if (!(user.buddyListChanged()))
            user.setBuddyListChanged();

          LinkedList recipients = z.getBuddyManager().getClientsToUpdate(user.getName());
          sendBuddyListUpdate(recipients, user.getName(), user.getUserId(), true, changedVars);
        }
      }
    }
  }

  private void handleSetBuddyBlockStatus(IXMLElement xml, SocketChannel sc)
  {
    xml = xml.getFirstChildNamed("n");

    User u = this.sfs.getUserByChannel(sc);

    if ((u != null) && (xml != null))
    {
      boolean status = xml.getAttribute("x", "0").equals("1");
      String buddyName = xml.getContent();

      setBuddyBlockStatus(u, buddyName, status);
    }
  }

  public void setBuddyBlockStatus(User u, String buddyName, boolean isBlocked)
  {
    this.sfs.getZone(u.getZone()).getBuddyManager().setBuddyBlockStatus(u, buddyName, isBlocked);
    SmartFoxServer.log.fine("Buddy block status set -> " + isBlocked + ", buddy: " + buddyName + ", user: " + u.getName());
  }

  private void handleSetRoomVariables(IXMLElement xml, int roomId, SocketChannel ch, boolean creatingRoom)
  {
    xml = xml.getFirstChildNamed("vars");
    String setOwnership = xml.getAttribute("so", "1");
    User u = this.sfs.getUserByChannel(ch);

    if (u != null)
    {
      if ((!(creatingRoom)) && (!(u.isInRoom(roomId)))) {
        return;
      }

      Zone z = this.sfs.getZone(u.getZone());

      Room r = z.getRoom(roomId);

      if ((r != null) && (r.isLimbo())) {
        return;
      }

      u.updateMessageTime();

      StringBuffer response = makeHeader("sys", roomId, "rVarsUpdate");

      Enumeration varNodes = xml.enumerateChildren();

      int changeCount = 0;

      response.append("<vars>");

      String str = null;

      while (varNodes.hasMoreElements())
      {
        IXMLElement var = (IXMLElement)varNodes.nextElement();

        String vName = var.getAttribute("n", "");
        String vType = var.getAttribute("t", "");
        String vValue = var.getContent();
        boolean vPrivate = var.getAttribute("pr", "").equals("1");
        boolean vPersist = var.getAttribute("pe", "").equals("1");

        SmartFoxServer.log.finer("VAR: " + vName + ", T:" + vType + ", V:" + vValue + ", PR:" + vPrivate + ", PE:" + vPersist);

        str = setRoomVariable(r, vName, vType, vValue, vPrivate, vPersist, u, !(setOwnership.equals("0")));
        if (str != null)
        {
          ++changeCount;
          response.append(str);
        }

      }

      response.append("</vars>");
      response.append(closeHeader());

      if (changeCount > 0)
      {
        this.eventWriter.handleEvent(new ServerEvent(response.toString(), ch, r.getChannellList()));
      }
    }
  }

  public String setRoomVariable(Room r, String vn, String vt, String vv, boolean pr, boolean ps, User u, boolean setOwnership)
  {
    StringBuffer result = new StringBuffer("");

    if (vt.equals("x"))
    {
      if (r.deleteVariable(vn, u))
      {
        result.append("<var n='").append(vn);
        result.append("' t='x' />");
      }

    }
    else if (r.setVariable(vn, vt, vv, pr, ps, u, setOwnership))
    {
      result.append(r.getXmlVariable(vn));
    }

    return ((result.length() > 0) ? result.toString() : null);
  }

  public String setUserVariable(String vn, String vt, String vv, User u)
  {
    StringBuffer result = new StringBuffer("");

    if (vt.equals("x"))
    {
      u.deleteVariable(vn);
      result.append("<var n='").append(vn);
      result.append("' t='x' />");
    }
    else if (u.setVariable(vn, vv, vt))
    {
      result.append("<var n='").append(vn);
      result.append("' t='").append(vt);
      result.append("'><![CDATA[").append(vv).append("]]></var>");
    }

    return result.toString();
  }

  private void handleCreateRoom(IXMLElement xml, int roomId, SocketChannel ch)
    throws CreateRoomException
  {
    String errorMsg = "";

    xml = xml.getFirstChildNamed("room");
    String updatable = xml.getAttribute("upd", "");
    String isTemp = xml.getAttribute("tmp", "");
    String isGame = xml.getAttribute("gam", "");
    String exit = xml.getAttribute("exit", "");

    String roomName = xml.getFirstChildNamed("name").getContent();
    String password = xml.getFirstChildNamed("pwd").getContent();
    String maxUsers = xml.getFirstChildNamed("max").getContent();
    String uCnt = null;

    IXMLElement xmlUCnt = xml.getFirstChildNamed("uCnt");

    if (xmlUCnt != null)
      uCnt = xmlUCnt.getContent();

    String maxSpectators = xml.getAttribute("spec", "");

    User u = this.sfs.getUserByChannel(ch);

    if (u != null)
    {
      u.updateMessageTime();

      Zone z = this.sfs.getZone(u.getZone());

      roomName = SmartFoxLib.cleanUpString(roomName, z.roomNameAvoidChars, z.getMaxRoomNameLen());

      if (roomName.equals(""))
      {
        errorMsg = "Bad or empty room name.";
      }
      else if (z.roomNameAlreadyExist(roomName))
      {
        errorMsg = "Room name already taken.";
      }
      else if (z.getRoomCount() >= 65536)
      {
        errorMsg = "Too many rooms.";
      }
      else if (u.getRoomsCreatedCount() >= z.getMaxRoomsPerUser())
      {
        errorMsg = "You have reached the max. amount of rooms that you can create at once.";
      }

      try
      {
        if (errorMsg.equals(""))
        {
          int maxU;
          int maxS;
          boolean bool1;
          if (isGame.equals("1"))
          {
            maxU = 16;
            maxS = 0;
          }
          else
          {
            maxU = 40;
            maxS = 0;
          }

          try
          {
            maxU = Integer.parseInt(maxUsers);
            maxS = Integer.parseInt(maxSpectators);
          }
          catch (NumberFormatException localNumberFormatException)
          {
          }

          boolean bool1 = !(isTemp.equals("0"));
          boolean game = isGame.equals("1");

          Room rm = this.sfs.createRoom(z.getName(), roomName, maxU, maxS, bool1, 
            game, (password.length() > 0) ? 1 : false, 
            password);

          rm.setCreator(ch);

          if (uCnt != null) {
            rm.setUcountUpdates(uCnt.equals("1"));
          }

          u.addCreatedRoom(rm);

          handleSetRoomVariables(xml, rm.getId(), ch, true);

          IXMLElement xtXml = xml.getFirstChildNamed("xt");

          if (xtXml != null)
          {
            String xtName = xtXml.getAttribute("n", "");
            String xtScript = xtXml.getAttribute("s", "");

            int type = 1;

            if ((xtScript.endsWith(".as")) || (xtScript.endsWith(".AS"))) {
              type = 2;
            }
            else if ((xtScript.endsWith(".py")) || (xtScript.endsWith(".PY")))
              type = 3;

            if (!(xtName.equals("")))
            {
              SmartFoxServer.getInstance().getExtensionHandler().createExtension(xtName, xtScript, type, u.getZone(), rm.getId());
            }

          }

          broadcastRoomAdded(z, rm);

          SmartFoxServer.log.info("Room [ " + roomName + " ] created!");

          if (!(z.isEventDisabled("onRoomAdded"))) {
            this.eventWriter.handleEvent(new ServerEvent(getRoomUpdateXML(rm), ch, z.getAllUsersInZone()));
          }

          if (!(game)) { return;
          }

          StringBuffer joinReq = new StringBuffer("<zz><room id='").append(rm.getId());

          joinReq.append("' pwd='").append(password);
          joinReq.append("' leave='").append(exit);
          joinReq.append("' old='").append(roomId).append("' /></zz>");
          try
          {
            IXMLParser xmlParser = XMLParserFactory.createDefaultXMLParser();

            IXMLReader reader = StdXMLReader.stringReader(joinReq.toString());
            xmlParser.setReader(reader);

            IXMLElement xmlMsg = (IXMLElement)xmlParser.parse();

            handleJoinRequest(xmlMsg, roomId, ch);
          }
          catch (Exception e)
          {
            SmartFoxServer.log.warning("Problems creating a room: " + e); } return;
        }

        sendRoomCreationError(errorMsg, z, ch);
      }
      catch (CreateRoomException cre)
      {
        sendRoomCreationError(cre.getMessage(), z, ch);
      }
    }
  }

  public String getRoomUpdateXML(Room rm)
  {
    StringBuffer sb = new StringBuffer();

    sb = makeHeader("sys", 0, "roomAdd");

    sb.append("<rm id='").append(rm.getId());
    sb.append("' priv='").append((rm.isPrivate()) ? "1" : "0");
    sb.append("' temp='").append((rm.isTemp()) ? "1" : "0");
    sb.append("' game='").append((rm.isGame()) ? "1" : "0");
    sb.append("' max='").append(rm.getMaxUsers());
    sb.append("' spec='").append(rm.getMaxSpectator());
    sb.append("' limbo='").append((rm.isLimbo()) ? "1" : "0");
    sb.append("'>");

    sb.append("<name><![CDATA[").append(rm.getName()).append("]]></name>");

    sb.append(rm.getXmlVarList());

    sb.append("</rm>");
    sb.append(closeHeader());

    return sb.toString();
  }

  private void sendRoomCreationError(String errorMsg, Zone z, SocketChannel ch)
  {
    if (!(z.isEventDisabled("onRoomAdded")))
    {
      String uName;
      User u = this.sfs.getUserByChannel(ch);

      if (u != null)
        uName = u.getName();
      else
        uName = "";

      SmartFoxServer.log.warning("Error during Room creation: [Zone = " + z.getName() + 
        "] " + errorMsg + " -- User: " + uName + ", IP: " + 
        ch.socket().getInetAddress());

      StringBuffer response = makeHeader("sys", 0, "createRmKO");
      response.append("<room e='").append(errorMsg).append("'/>");
      response.append(closeHeader());

      LinkedList ll = new LinkedList();
      ll.add(ch);

      this.eventWriter.handleEvent(new ServerEvent(response.toString(), ch, ll));
    }
  }

  private void handleLeaveRoom(IXMLElement xml, int fromRoom, SocketChannel ch)
  {
    xml = xml.getFirstChildNamed("rm");

    String rid = xml.getAttribute("id", "");
    try
    {
      toLeave = Integer.parseInt(rid);

      User u = this.sfs.getUserByChannel(ch);

      leaveRoom(u, toLeave, fromRoom, true);
    }
    catch (NumberFormatException toLeave)
    {
    }
  }

  public void leaveRoom(User u, int roomId, int fromRoom, boolean broadcastEvent)
  {
    if ((u != null) && 
      (u.getRoomsConnectedCount() > 1))
    {
      Zone z = this.sfs.getZone(u.getZone());

      Room r = z.getRoom(roomId);

      if (r != null)
      {
        int oldRoomPlayerIndex = u.getPlayerIndex(r);

        r.removeUser(u, true, true);

        broadcastUserRemoved(u, r, oldRoomPlayerIndex);

        checkEmptyRoom(r, u);

        if ((!(z.isEventDisabled("onRoomLeft"))) && (broadcastEvent))
        {
          StringBuffer response = makeHeader("sys", fromRoom, "leaveRoom");
          response.append("<rm id='").append(String.valueOf(roomId)).append("' />");
          response.append(closeHeader());

          LinkedList ll = new LinkedList();
          ll.add(u.getChannel());

          this.eventWriter.handleEvent(new ServerEvent(response.toString(), u.getChannel(), ll));
        }
      }
    }
  }

  public void leaveRoom(User u, int roomId, boolean broadcastEvent)
  {
    leaveRoom(u, roomId, -1, broadcastEvent);
  }

  public void kickUser(SocketChannel ch)
  {
    SmartFoxServer.getInstance().lostConnection(ch);
  }

  private void handleRoundTripRequest(int roomId, SocketChannel ch)
  {
    StringBuffer response = makeHeader("sys", roomId, "roundTripRes").append(closeHeader());

    LinkedList ll = new LinkedList();
    ll.add(ch);

    this.eventWriter.handleEvent(new ServerEvent(response.toString(), ch, ll));
  }

  private void handleSpectatorSwitch(int roomId, SocketChannel ch)
  {
    User u = SmartFoxServer.getInstance().getUserByChannel(ch);
    switchSpectator(u, roomId, true);
  }

  public void switchSpectator(User u, int roomId, boolean broadcastEvent)
  {
    if (u != null)
    {
      SocketChannel userChannel = u.getChannel();

      StringBuffer responseForRequester = makeHeader("sys", roomId, "swSpec");
      StringBuffer responseForRoomMates = makeHeader("sys", roomId, "swSpec");

      Zone z = this.sfs.getZone(u.getZone());

      Room r = z.getRoom(roomId);

      if ((r != null) && (u.isSpectator()))
      {
        int res = r.switchSpectator(u);

        responseForRequester.append("<pid id='").append(res).append("' />").append(closeHeader());
        responseForRoomMates.append("<pid id='").append(res).append("' u='").append(u.getUserId()).append("'/>").append(closeHeader());

        LinkedList userResponse = new LinkedList();
        userResponse.add(userChannel);

        LinkedList otherChannels = r.getChannellList();
        otherChannels.remove(userChannel);

        if ((!(z.isEventDisabled("onSpectatorSwitched"))) && (broadcastEvent))
        {
          this.eventWriter.handleEvent(new ServerEvent(responseForRequester.toString(), u.getChannel(), userResponse));

          if (res > 0)
          {
            this.eventWriter.handleEvent(new ServerEvent(responseForRoomMates.toString(), u.getChannel(), otherChannels));
          }

        }

        broadcastSpectatorSwitched(u, res, r);
      }
    }
  }

  private void handleLoadBuddyList(SocketChannel ch)
  {
    StringBuffer response = makeHeader("sys", -1, "bList");

    User u = SmartFoxServer.getInstance().getUserByChannel(ch);

    if (u != null)
    {
      Zone z = SmartFoxServer.getInstance().getZone(u.getZone());

      if (z != null)
      {
        try
        {
          LoadableBuddyList bList = z.loadBuddyList(u.getName());
          LoadableBuddyItem b = null;

          response.append(getXmlBuddyVariables("mv", bList.variables));

          response.append("<bList>");

          synchronized (bList.buddies)
          {
            for (Iterator i = bList.buddies.iterator(); i.hasNext(); )
            {
              b = (LoadableBuddyItem)i.next();

              response.append("<b s='").append((b.isActive()) ? "1" : "0");
              response.append("' i='").append(b.getUserId());
              response.append("' x='").append((b.isIgnored()) ? "1" : "0");
              response.append("'><n><![CDATA[").append(b.getName()).append("]]></n>");

              User user = SmartFoxServer.getInstance().getUserById(Integer.valueOf(b.getUserId()));

              if (user != null)
              {
                if (user.buddyVariables.size() > 0) {
                  response.append(getBuddyVariables(user));
                }

              }
              else
              {
                Map vars = z.getBuddyManager().getOfflineBuddyVariables(b.getName());

                if (vars != null)
                  response.append(getXmlBuddyVariables("vs", vars));
              }

              response.append("</b>");
            }
          }

          response.append("</bList>");
        }
        catch (BuddyListException ble)
        {
          SmartFoxServer.log.warning(ble.toString());
          response.append("<err>Unable to load the Buddy List!</err>");
        }

        response.append(closeHeader());

        LinkedList ll = new LinkedList();
        ll.add(ch);

        this.eventWriter.handleEvent(new ServerEvent(response.toString(), ch, ll));
      }
    }
  }

  private void handleAddBuddy(IXMLElement node, SocketChannel ch)
  {
    IXMLElement buddyNode = node.getFirstChildNamed("n");
    String buddyName = buddyNode.getContent();

    User sender = SmartFoxServer.getInstance().getUserByChannel(ch);

    if ((sender != null) && (buddyName != null) && (!(buddyName.equals(""))))
    {
      doAddBuddy(buddyName, sender);
    }
  }

  private void handleRemoveBuddy(IXMLElement node, SocketChannel ch)
  {
    IXMLElement buddyNode = node.getFirstChildNamed("n");
    String buddyName = buddyNode.getContent();

    User sender = SmartFoxServer.getInstance().getUserByChannel(ch);

    if ((sender != null) && (buddyName != null) && (!(buddyName.equals(""))))
    {
      doRemoveBuddy(buddyName, sender);
    }
  }

  private void handleClearBuddyList(SocketChannel ch)
  {
    User sender = SmartFoxServer.getInstance().getUserByChannel(ch);

    if (sender != null)
    {
      doClearBuddyList(sender);
    }
  }

  private void handleBuddyPermission(IXMLElement node, SocketChannel ch)
  {
    User sender = SmartFoxServer.getInstance().getUserByChannel(ch);
    IXMLElement textMsg = node.getFirstChildNamed("txt");
    node = node.getFirstChildNamed("n");

    if ((node != null) && (sender != null))
    {
      String response = node.getAttribute("res", "r");

      String target = node.getContent();

      String message = null;

      if (textMsg != null)
        message = textMsg.getContent();

      Zone zone = SmartFoxServer.getInstance().getZone(sender.getZone());
      PermissionStatus status = (response.equals("g")) ? PermissionStatus.GRANTED : PermissionStatus.REJECTED;

      boolean ok = zone.getBuddyManager().setPermissionStatus(target, sender.getName(), status);

      if (ok)
      {
        doAddBuddy(sender.getName(), zone.getUserByName(target), true);

        if (zone.getBuddyManager().isMutualAddBuddy()) {
          doAddBuddy(target, sender, true);
        }

      }
      else
      {
        SmartFoxServer.log.info("AddBuddy permission request rejected. Sender: " + sender.getName() + ", Target: " + target + ", status = " + response.toUpperCase());
      }
    }
  }

  private void doAddBuddy(String buddyName, User sender, boolean bypassSecurityCheck)
  {
    StringBuffer response;
    LinkedList ll;
    Zone z = SmartFoxServer.getInstance().getZone(sender.getZone());

    boolean isAdvanced = z.getBuddyManager().isAdvancedMode();
    boolean usePermission = z.getBuddyManager().isAddBuddyPermission();

    if ((!(bypassSecurityCheck)) && (isAdvanced) && (usePermission))
    {
      try
      {
        z.getBuddyManager().checkValidAddRequest(sender, buddyName);

        askAddBuddyPermission(sender, z.getUserByName(buddyName), z.getBuddyManager());
      }
      catch (BuddyListException error)
      {
        SmartFoxServer.log.info("Add buddy request rejected. Sender: " + 
          sender.getName() + 
          ", Target: " + buddyName + 
          ", Reason: " + error.getMessage());
      }

      return;
    }

    try
    {
      z.addBuddy(sender.getName(), buddyName);

      User b = SmartFoxServer.getInstance().getUserByChannel(z.getChannelFromName(buddyName));

      if (!(sender.buddyListChanged()))
        sender.setBuddyListChanged();

      response = makeHeader("sys", -1, "bAdd");
      response.append("<b s='").append((b != null) ? "1" : "0");
      response.append("' i='").append((b != null) ? b.getUserId() : -1);
      response.append("'><n><![CDATA[").append(buddyName).append("]]></n>");

      if ((b != null) && (b.buddyVariables.size() > 0)) {
        response.append(getBuddyVariables(b));
      }
      else
      {
        Map vars = z.getBuddyManager().getOfflineBuddyVariables(buddyName);

        if (vars != null)
          response.append(getXmlBuddyVariables("vs", vars));
      }

      response.append("</b>");
      response.append(closeHeader());

      ll = new LinkedList();
      ll.add(sender.getChannel());

      this.eventWriter.handleEvent(new ServerEvent(response.toString(), sender.getChannel(), ll));
    }
    catch (BuddyListException ble)
    {
      response = makeHeader("sys", -1, "bUpd");
      response.append("<err>Your buddy list is full!</err>");
      response.append(closeHeader());

      ll = new LinkedList();
      ll.add(sender.getChannel());

      this.eventWriter.handleEvent(new ServerEvent(response.toString(), sender.getChannel(), ll));
    }
  }

  public void doAddBuddy(String buddyName, User sender)
  {
    doAddBuddy(buddyName, sender, false);
  }

  public void doRemoveBuddy(String buddyName, User sender)
  {
    Zone z = SmartFoxServer.getInstance().getZone(sender.getZone());
    z.removeBuddy(sender.getName(), buddyName);

    if (!(sender.buddyListChanged())) {
      sender.setBuddyListChanged();
    }

    if ((z.getBuddyManager().isMutualRemoveBuddy()) && (z.getBuddyManager().isAddBuddyPermission()))
    {
      User removedBuddyUser = z.getUserByName(buddyName);

      if (removedBuddyUser != null)
      {
        if (!(removedBuddyUser.buddyListChanged()))
          removedBuddyUser.setBuddyListChanged();

        LinkedList ll = new LinkedList();
        ll.add(removedBuddyUser.getChannel());

        sendBuddyRemovedUpdate(ll, sender.getName());
      }
    }
  }

  public void doClearBuddyList(User sender)
  {
    Zone z = SmartFoxServer.getInstance().getZone(sender.getZone());
    z.clearBuddyList(sender.getName());

    if (!(sender.buddyListChanged()))
      sender.setBuddyListChanged();
  }

  private void askAddBuddyPermission(User requester, User target, BuddyListManager bManager)
  {
    askAddBuddyPermission(requester, target, null, bManager);
  }

  public void askAddBuddyPermission(User requester, User target, String extraMessage, BuddyListManager bManager)
  {
    bManager.addPermissionRequest(requester.getName(), target.getName());

    StringBuffer response = makeHeader("sys", -1, "bPrm");
    response.append("<n><![CDATA[").append(requester.getName()).append("]]></n>");

    if (extraMessage != null)
      response.append("<txt><![CDATA[").append(extraMessage).append("]]></txt>");

    response.append(closeHeader());

    LinkedList ll = new LinkedList();
    ll.add(target.getChannel());

    SmartFoxServer.log.info("Asking buddy permission. Sender: " + requester.getName() + ", Target:" + target);

    this.eventWriter.handleEvent(new ServerEvent(response.toString(), requester.getChannel(), ll));
  }

  private void handleGetBuddyRoom(IXMLElement node, SocketChannel ch)
  {
    node = node.getFirstChildNamed("b");
    String buddyId = node.getAttribute("id", "");

    if ((buddyId != null) && 
      (!(buddyId.equals(""))))
    {
      User u = SmartFoxServer.getInstance().getUserByChannel(ch);
      User b = SmartFoxServer.getInstance().getUserById(
        new Integer(buddyId));

      if ((u != null) && (b != null))
      {
        int[] roomIds = b.getRoomsConnected();

        StringBuffer response = makeHeader("sys", -1, "roomB");
        response.append("<br r='");

        for (int i = 0; i < roomIds.length; ++i)
          response.append(roomIds[i]).append(",");

        response = response.deleteCharAt(response.length() - 1);
        response.append("' />");
        response.append(closeHeader());

        LinkedList ll = new LinkedList();
        ll.add(ch);

        this.eventWriter.handleEvent(new ServerEvent(response.toString(), ch, ll));
      }
    }
  }

  private void handleRandomKeyRequest(SocketChannel senderChannel)
  {
    SelectionKey sk = senderChannel.keyFor(SmartFoxServer.getInstance().getReadSelector());
    Attachment att = (Attachment)sk.attachment();

    if (att.getSecretKey() == null)
    {
      String key = SecretKeyGenerator.instance().getKey();
      att.setSecretKey(key);

      StringBuffer response = makeHeader("sys", -1, "rndK");
      response.append("<k>").append(key).append("</k>").append(closeHeader());

      LinkedList ll = new LinkedList();
      ll.add(senderChannel);

      this.eventWriter.handleEvent(new ServerEvent(response.toString(), senderChannel, ll));
    }
  }

  private void handleZoneInfo(IXMLElement node, SocketChannel ch)
  {
    if (!(ConfigData.ALLOW_ZONE_INFO)) {
      return;
    }

    String request = node.getContent();

    StringBuffer response = new StringBuffer("<zInfo>");

    if ((request != null) && (request.length() > 0))
    {
      String[] requestedZones = request.split(",");
      Zone z = null;

      if (requestedZones.length > 0)
      {
        for (int i = 0; i < requestedZones.length; ++i)
        {
          z = this.sfs.getZone(requestedZones[i]);

          if (z != null)
            response.append(z.getUserCount()).append(",");

        }

        if (response.length() > 1)
        {
          response.deleteCharAt(response.length() - 1);
        }

        response.append("</zInfo>");

        LinkedList ll = new LinkedList();
        ll.add(ch);

        this.eventWriter.handleEvent(new ServerEvent(response.toString(), ch, ll));
      }
    }
  }

  public void sendBuddyListUpdate(LinkedList channelList, String buddyName, int uid, boolean isOnline, boolean sendVars)
  {
    StringBuffer response = prepareBuddyListUpdateHeader(buddyName, uid, isOnline);

    if (sendVars)
    {
      User u = SmartFoxServer.getInstance().getUserById(Integer.valueOf(uid));

      if ((u != null) && (u.buddyVariables.size() > 0))
        response.append(getBuddyVariables(u));
    }

    response.append("</b>");
    response.append(closeHeader());

    this.eventWriter.handleEvent(new ServerEvent(response.toString(), null, channelList));
  }

  public void sendBuddyListUpdate(LinkedList channelList, String buddyName, int uid, boolean isOnline, List<String> changedVars)
  {
    StringBuffer response = prepareBuddyListUpdateHeader(buddyName, uid, isOnline);

    User u = SmartFoxServer.getInstance().getUserById(Integer.valueOf(uid));

    if ((u != null) && (u.buddyVariables.size() > 0))
      response.append(getBuddyVariables(u, changedVars));

    response.append("</b>");
    response.append(closeHeader());

    this.eventWriter.handleEvent(new ServerEvent(response.toString(), null, channelList));
  }

  public void sendBuddyRemovedUpdate(LinkedList channelList, String buddyName)
  {
    StringBuffer response = makeHeader("sys", -1, "remB");
    response.append("<n><![CDATA[").append(buddyName).append("]]></n>");
    response.append(closeHeader());

    this.eventWriter.handleEvent(new ServerEvent(response.toString(), null, channelList));
  }

  private StringBuffer prepareBuddyListUpdateHeader(String buddyName, int uid, boolean isOnline)
  {
    StringBuffer response = makeHeader("sys", -1, "bUpd");

    response.append("<b s='").append((isOnline) ? "1" : "0");
    response.append("' i='").append(uid);
    response.append("'><n><![CDATA[").append(buddyName).append("]]></n>");

    return response;
  }

  public void sendBuddyRemovedUpdate(String buddyName, User recipient)
  {
    StringBuffer response = makeHeader("sys", -1, "bRem");
    response.append("<n><![CDATA[").append(buddyName).append("]]></n>");
    response.append(closeHeader());

    LinkedList recipients = new LinkedList();
    recipients.add(recipient);

    this.eventWriter.handleEvent(new ServerEvent(response.toString(), null, recipients));
  }

  public void sendBuddyListCleared(User recipient)
  {
    StringBuffer response = makeHeader("sys", -1, "bClr").append(closeHeader());

    LinkedList recipients = new LinkedList();
    recipients.add(recipient);

    this.eventWriter.handleEvent(new ServerEvent(response.toString(), null, recipients));
  }

  private String getBuddyVariables(User user)
  {
    return getBuddyVariables(user, null);
  }

  private String getBuddyVariables(User user, List<String> changedVars)
  {
    return getBuddyVariables(user.buddyVariables, changedVars);
  }

  public String getBuddyVariables(Map<String, String> variables, List<String> changedVars)
  {
    StringBuilder xml = new StringBuilder("<vs>");

    Set vars = variables.entrySet();

    boolean useChangedVars = (changedVars != null) && (changedVars.size() > 0);
    boolean isFound = false;

    for (Iterator i = vars.iterator(); i.hasNext(); )
    {
      Map.Entry var = (Map.Entry)i.next();
      String key = (String)var.getKey();

      if (useChangedVars) {
        isFound = changedVars.contains(key);
      }

      if ((useChangedVars) && (((!(useChangedVars)) || (!(isFound)))))
        break label156;
      xml.append("<v n='").append((String)var.getKey());
      label156: xml.append("'><![CDATA[").append((String)var.getValue()).append("]]></v>");
    }

    xml.append("</vs>");

    return xml.toString();
  }

  private String getXmlBuddyVariables(String enclosingTagName, Map<String, String> variables)
  {
    if ((variables == null) || (variables.size() == 0))
      return "";

    StringBuilder xml = new StringBuilder("<").append(enclosingTagName).append(">");

    for (Iterator iter = variables.entrySet().iterator(); iter.hasNext(); )
    {
      Map.Entry entry = (Map.Entry)iter.next();
      xml.append("<v n='").append((String)entry.getKey()).append("'>");
      xml.append("<![CDATA[").append((String)entry.getValue()).append("]]></v>");
    }

    xml.append("</").append(enclosingTagName).append(">");

    return xml.toString();
  }

  public StringBuffer makeHeader(String type, int room, String action)
  {
    StringBuffer sb = new StringBuffer();
    sb.append("<msg t='").append(type);
    sb.append("'><body action='").append(action);
    sb.append("' r='").append(room).append("'>");

    return sb;
  }

  public String closeHeader()
  {
    return "</body></msg>";
  }

  private void broadcastUserRemoved(User u, Room r, int oldPlayerIndex)
  {
    InternalEventObject ioe = new InternalEventObject("userExit");
    ioe.addParam("zone", u.getZone());
    ioe.addParam("uid", String.valueOf(u.getUserId()));
    ioe.addParam("oldPlayerIndex", String.valueOf(oldPlayerIndex));
    ioe.addObject("room", r);

    ioe.addObject("user", u);

    ioe.addParam("roomId", String.valueOf(r.getId()));
    ioe.addParam("roomName", r.getName());
    ioe.addParam("roomIsGame", String.valueOf(r.isGame()));

    dispatchEvent(ioe, u.getZone(), r.getName());
  }

  public void broadcastUserLost(User u, int[] roomIds, int[] playerIndexes)
  {
    InternalEventObject ioe = new InternalEventObject("userLost");
    ioe.addParam("zone", u.getZone());
    ioe.addParam("uid", String.valueOf(u.getUserId()));
    ioe.addObject("roomIds", roomIds);
    ioe.addObject("playerIndexes", playerIndexes);

    ioe.addObject("user", u);

    ioe.addParam("isPlaying", String.valueOf(u.isBusyForPlay()));

    dispatchEvent(ioe, u.getZone(), null, roomIds);
  }

  public void broadCastUserLogout(User u, String zone, int[] roomIds, int[] playerIndexes)
  {
    InternalEventObject event = new InternalEventObject("logOut");
    event.addParam("zone", zone);
    event.addParam("uid", String.valueOf(u.getUserId()));
    event.addObject("roomIds", roomIds);
    event.addObject("playerIndexes", playerIndexes);
    event.addObject("chan", u.getChannel());

    event.addObject("user", u);

    dispatchEvent(event, zone, null, roomIds);
  }

  public void broadcastRoomDestroyed(int roomId, String zone)
  {
    InternalEventObject ioe = new InternalEventObject("roomLost");
    ioe.addParam("zone", zone);
    ioe.addParam("roomId", String.valueOf(roomId));

    dispatchEvent(ioe, zone, null);
  }

  private void broadcastSpectatorSwitched(User u, int res, Room r)
  {
    InternalEventObject ioe = new InternalEventObject("spectatorSwitched");
    ioe.addObject("user", u);
    ioe.addParam("playerIndex", String.valueOf(res));

    dispatchEvent(ioe, u.getZone(), r.getName());
  }

  public void broadcastRoomAdded(Zone z, Room r)
  {
    InternalEventObject ioe = new InternalEventObject("newRoom");
    ioe.addParam("name", r.getName());
    ioe.addParam("zone", z.getName());
    ioe.addParam("roomName", r.getName());
    ioe.addParam("roomId", String.valueOf(r.getId()));
    ioe.addParam("game", String.valueOf(r.isGame()));
    ioe.addObject("room", r);

    dispatchEvent(ioe, z.getName(), null);
  }

  private void broadcastPublicMessage(Zone z, Room r, User u, String msg)
  {
    InternalEventObject ieo = new InternalEventObject("pubMsg");
    ieo.addParam("msg", msg);
    ieo.addObject("user", u);
    ieo.addObject("room", r);

    dispatchEvent(ieo, z.getName(), null);
  }

  private void broadcastPrivateMessage(Zone z, Room r, User sender, User recipient, String msg)
  {
    InternalEventObject ieo = new InternalEventObject("privMsg");
    ieo.addParam("msg", msg);
    ieo.addObject("room", r);
    ieo.addObject("sender", sender);
    ieo.addObject("recipient", recipient);

    dispatchEvent(ieo, z.getName(), r.getName());
  }

  public void broadcastFileUploaded(String uid, HashMap files)
  {
    User u = SmartFoxServer.getInstance().getUserById(new Integer(uid));

    if (u != null)
    {
      InternalEventObject ieo = new InternalEventObject("fileUpload");
      ieo.addObject("user", u);
      ieo.addObject("files", files);

      dispatchEvent(ieo, u.getZone(), null);
    }
  }

  private void applyFloodFilter(String txtmsg, User u)
  {
    if (!(u.isBeingKicked))
    {
      if (System.currentTimeMillis() <= u.getLastMessageTime() + ConfigData.ANTIFLOOD_MIN_MSG_TIME)
      {
        u.floodCounter += 1;

        if (u.floodCounter >= ConfigData.ANTIFLOOD_TOLERANCE)
        {
          u.floodWarningsCounter += 1;
          u.floodCounter = 0;

          sendAdminMessage(ConfigData.ANTIFLOOD_WARNING_MSG, u, "dmn");
        }
      }
      else {
        u.floodCounter = 0;
      }
      if (txtmsg.equals(u.lastMessage))
      {
        u.repeatedMsgCounter += 1;

        if (u.repeatedMsgCounter >= ConfigData.ANTIFLOOD_MAX_REPEATED)
        {
          u.floodWarningsCounter += 1;
          u.repeatedMsgCounter = 0;

          sendAdminMessage(ConfigData.ANTIFLOOD_WARNING_MSG, u, "dmn");
        }
      }
      else
      {
        u.repeatedMsgCounter = 0;
        u.lastMessage = txtmsg;
      }

      if (u.floodWarningsCounter >= ConfigData.ANTIFLOOD_WARNINGS)
      {
        SmartFoxServer.log.warning("User: " + u.getName() + " (" + 
          u.getIpAddress() + 
          ") is being kicked. Reason: flooding.");

        u.isBeingKicked = true;

        sendAdminMessage(ConfigData.ANTIFLOOD_KICK_MSG, u, "dmn");

        SmartFoxServer.getInstance().addKickedUser(u, 
          1);

        SmartFoxServer.getInstance().getExtensionHandler().getAdminExtension().addAdminTask(
          new KickUser(u.getChannel(), 
          System.currentTimeMillis() + 6000L, 
          false));
      }
    }

    u.updateMessageTime();
  }

  private String applyBadWordFilter(String inStr, User u)
  {
    String outStr = null;
    boolean userExist = u != null;

    if (ConfigData.BADWORDS_FILTER_MODE == 0)
    {
      outStr = BadWordsFilter.getInstance().filterWords(inStr);

      if ((ConfigData.BADWORDS_WARNINGS) && (userExist) && (!(outStr.equals(inStr))))
      {
        u.badwordsWarningCounter += 1;
        sendAdminMessage(ConfigData.BADWORDS_WARNING_MSG, u, "dmn");
      }

    }
    else if (ConfigData.BADWORDS_FILTER_MODE == 1)
    {
      if (BadWordsFilter.getInstance().containsBadWords(inStr))
      {
        outStr = "";

        if ((ConfigData.BADWORDS_WARNINGS) && (userExist))
        {
          u.badwordsWarningCounter += 1;
          sendAdminMessage(ConfigData.BADWORDS_WARNING_MSG, u, "dmn");
        }
      }
      else
      {
        outStr = inStr;
      }

    }

    if ((userExist) && 
      (!(u.isBeingKicked)) && 
      (u.badwordsWarningCounter >= ConfigData.BADWORDS_WARNINGS_BEFORE_KICK))
    {
      SmartFoxServer.log.warning("User: " + u.getName() + " (" + 
        u.getIpAddress() + 
        ") is being kicked. Reason: swearing.");

      u.isBeingKicked = true;

      sendAdminMessage(ConfigData.BADWORDS_KICK_MSG, u, "dmn");

      SmartFoxServer.getInstance().addKickedUser(u, 
        2);

      AdminExtension dmn = SmartFoxServer.getInstance()
        .getExtensionHandler().getAdminExtension();
      dmn.addAdminTask(
        new KickUser(u.getChannel(), 
        System.currentTimeMillis() + 
        6000L, false));
    }

    return outStr;
  }

  private String getUserIp(SocketChannel sc)
  {
    if ((sc != null) && (sc.isOpen()))
    {
      return sc.socket().getInetAddress().getHostAddress();
    }

    return "- not available -";
  }
}