package it.gotoandplay.smartfoxserver.data;

import it.gotoandplay.smartfoxserver.EventWriter;
import it.gotoandplay.smartfoxserver.SmartFoxServer;
import it.gotoandplay.smartfoxserver.config.ConfigData;
import it.gotoandplay.smartfoxserver.controllers.SystemHandler;
import it.gotoandplay.smartfoxserver.exceptions.JoinRoomException;
import it.gotoandplay.smartfoxserver.extensions.ExtensionManager;
import it.gotoandplay.smartfoxserver.lib.ServerEvent;
import java.io.PrintStream;
import java.nio.channels.SocketChannel;
import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.Map;
import java.util.Set;
import java.util.logging.Logger;

public class Room
{
  private static int autoId = 0;
  private int id;
  private String name;
  private int maxUsers;
  private int maxSpectators;
  private int childRoom;
  private String password;
  private HashMap roomVariables;
  private LinkedList userList;
  private LinkedList userNames;
  private boolean isPrivate;
  private boolean isTemporary;
  private boolean isGame;
  private boolean uCountUpdates;
  private boolean isLimbo;
  private int userCount;
  private int spectatorCount;
  private SocketChannel creator;
  private String zone;
  private boolean[] userSlots;
  private ExtensionManager extManager;
  public Map properties;

  public Room(String name, int maxUsers, boolean isTemp, String zone)
  {
    setRoomID();
    this.name = name;
    this.maxUsers = maxUsers;
    this.isTemporary = isTemp;
    this.isPrivate = false;
    this.isGame = false;
    this.uCountUpdates = true;
    this.creator = null;
    this.zone = zone;
    this.isLimbo = false;

    this.userList = new LinkedList();
    this.userNames = new LinkedList();
    this.roomVariables = new HashMap();

    this.userCount = 0;
    this.spectatorCount = 0;

    if (this.isGame) {
      initUserSlots();
    }

    try
    {
      this.properties = ((Map)Class.forName(ConfigData.dynamicPropertiesClass).newInstance());
    }
    catch (Exception e)
    {
      this.properties = new HashMap();
    }

    this.extManager = new ExtensionManager();
  }

  public Room(String name, int maxUsers, boolean isTemp, boolean isGame, boolean isPrivate, String password, String zone)
  {
    setRoomID();
    this.name = name;
    this.maxUsers = maxUsers;
    this.isTemporary = isTemp;
    this.isPrivate = isPrivate;
    this.isGame = isGame;
    this.isLimbo = false;
    this.password = password;
    this.zone = zone;

    this.uCountUpdates = (!(this.isGame));

    this.userList = new LinkedList();
    this.userNames = new LinkedList();

    this.roomVariables = new HashMap();

    this.extManager = new ExtensionManager();

    this.properties = new HashMap();

    if (isGame)
      initUserSlots();
  }

  public void destroyExtensions()
  {
    if (this.extManager != null)
    {
      this.extManager.shutDown();
      this.extManager = null;
    }
  }

  public synchronized int switchSpectator(User u)
  {
    int result = -1;

    if ((this.isGame) && 
      (getUserCount() < getMaxUsers()))
    {
      result = getPlayerSlot();

      this.userCount += 1;
      this.spectatorCount -= 1;

      u.setPlayerIndex(this, result);

      updateUserCount();
    }

    return result;
  }

  public void setMaxSpectators(int s)
  {
    this.maxSpectators = s;
  }

  public int getMaxSpectator()
  {
    return this.maxSpectators;
  }

  public void setUcountUpdates(boolean b)
  {
    this.uCountUpdates = b;
  }

  public boolean getUcountUpdates()
  {
    return this.uCountUpdates;
  }

  private void initUserSlots()
  {
    this.userSlots = new boolean[this.maxUsers + 1];
  }

  private void removeUserSlot(int i)
  {
    this.userSlots[i] = false;
  }

  public ExtensionManager getExtManager()
  {
    return this.extManager;
  }

  public void join(User u, String pwd, boolean isSpectator)
    throws JoinRoomException
  {
    if (!(contains(u.getName())))
    {
      boolean normalJoinOK = (!(this.isGame)) && (!(isSpectator)) && (this.userCount < this.maxUsers);
      boolean normalGameOK = (this.isGame) && (!(isSpectator)) && (this.userCount < this.maxUsers);
      boolean spectatorGameOK = (this.isGame) && (isSpectator) && (this.spectatorCount < this.maxSpectators);

      if ((normalJoinOK) || (normalGameOK) || (spectatorGameOK))
      {
        if (this.isPrivate)
        {
          if (this.password.equals(pwd))
          {
            addUser(u, isSpectator); break label220:
          }

          throw new JoinRoomException("Password is wrong");
        }

        addUser(u, isSpectator);
      }
      else
      {
        String s = "Problems joining this room";

        if (!(this.isGame))
          s = "This room is currently full";
        else if ((this.isGame) && (!(isSpectator)))
          s = "All player slots are occupied";
        else if ((this.isGame) && (isSpectator))
          s = "All spectator slots are occupied";

        throw new JoinRoomException(s);
      }

      if (isGame())
      {
        if (isSpectator)
          label220: u.setPlayerIndex(this, -1);
        else
          u.setPlayerIndex(this, getPlayerSlot());
      }
    }
    else {
      throw new JoinRoomException("User is already in this room!");
    }
  }

  private void addUser(User u, boolean isSpectator)
  {
    synchronized (this.userList)
    {
      this.userList.add(new Integer(u.getUserId()));
    }

    synchronized (this.userNames)
    {
      this.userNames.add(u.getName());
    }

    if (isSpectator)
      this.spectatorCount += 1;
    else {
      this.userCount += 1;
    }

    u.addRoom(this);
  }

  private int getPlayerSlot()
  {
    for (int i = 1; i <= this.maxUsers; ++i)
    {
      if (this.userSlots[i] == 0)
      {
        this.userSlots[i] = true;
        return i;
      }
    }
    return -1;
  }

  public String getXmlUserList()
  {
    StringBuffer xml = new StringBuffer();

    synchronized (this.userList)
    {
      Iterator it = this.userList.iterator();

      while (it.hasNext())
      {
        Integer id = (Integer)it.next();

        User u = SmartFoxServer.getInstance().getUserById(id);

        if (u != null)
        {
          String usrName = SmartFoxServer.getInstance().getUserById(id).getName();

          xml.append("<u i='").append(id.toString()).append("' m='").append((u.isModerator()) ? "1" : "0");

          if (isGame())
          {
            xml.append("' s='").append((u.isSpectator()) ? "1" : "0");
            xml.append("' p='").append(u.getPlayerIndex());
          }

          xml.append("'><n><![CDATA[").append(usrName).append("]]></n>");
          xml.append(u.getXmlVarList());
          xml.append("</u>");
        }
      }
    }

    return xml.toString();
  }

  public Object[] getUserList()
  {
    Object[] o = (Object[])null;

    synchronized (this.userList)
    {
      o = this.userList.toArray();
    }

    return o;
  }

  public Object[] getSpectatorList()
  {
    Object[] o = new Object[this.maxSpectators];
    int c = 0;
    User u = null;

    synchronized (this.userList)
    {
      for (Iterator it = this.userList.iterator(); it.hasNext(); )
      {
        u = SmartFoxServer.getInstance().getUserById((Integer)it.next());
        if (u.getPlayerIndex(this) != -1) break label65;
        label65: o[(c++)] = u;
      }
    }

    return o;
  }

  public User[] getAllUsers()
  {
    User[] allUsers = (User[])null;

    synchronized (this.userList)
    {
      allUsers = new User[this.userList.size()];
      int c = 0;

      for (Iterator i = this.userList.iterator(); i.hasNext(); )
      {
        allUsers[(c++)] = SmartFoxServer.getInstance().getUserById((Integer)i.next());
      }
    }

    return allUsers;
  }

  public User[] getAllUsersButOne(User excludedUser)
  {
    Integer id = null;

    if (excludedUser != null)
      id = Integer.valueOf(excludedUser.getUserId());

    return getAllUsersButOne(id);
  }

  public User[] getAllUsersButOne(Integer excludedUserId)
  {
    if (excludedUserId == null)
      return getAllUsers();

    User[] allUsers = (User[])null;

    boolean exist = false;

    synchronized (this.userList)
    {
      exist = this.userList.contains(excludedUserId);
    }

    if (!(exist)) {
      return getAllUsers();
    }

    synchronized (this.userList)
    {
      allUsers = new User[this.userList.size() - 1];
      int c = 0;

      for (Iterator i = this.userList.iterator(); i.hasNext(); )
      {
        Integer uid = (Integer)i.next();
        User u = SmartFoxServer.getInstance().getUserById(uid);

        if (uid.equals(excludedUserId)) break label129;
        label129: allUsers[(c++)] = u;
      }
    }

    return allUsers;
  }

  public User[] getAllPlayers()
  {
    LinkedList tempList = new LinkedList();

    synchronized (this.userList)
    {
      for (Iterator i = this.userList.iterator(); i.hasNext(); )
      {
        User u = SmartFoxServer.getInstance().getUserById((Integer)i.next());
        if (u.getPlayerIndex(this) == -1) break label59;
        label59: tempList.add(u);
      }
    }

    User[] allPlayers = new User[tempList.size()];

    for (int i = 0; i < tempList.size(); ++i)
      allPlayers[i] = ((User)tempList.get(i));

    return allPlayers;
  }

  public User[] getAllSpectators()
  {
    LinkedList tempList = new LinkedList();

    synchronized (this.userList)
    {
      for (Iterator i = this.userList.iterator(); i.hasNext(); )
      {
        User u = SmartFoxServer.getInstance().getUserById((Integer)i.next());
        if (u.getPlayerIndex(this) != -1) break label59;
        label59: tempList.add(u);
      }
    }

    User[] allSpectators = new User[tempList.size()];

    for (int i = 0; i < tempList.size(); ++i)
      allSpectators[i] = ((User)tempList.get(i));

    return allSpectators;
  }

  public User getUserByPlayerIndex(int pid)
  {
    User u = null;
    boolean found = false;

    synchronized (this.userList)
    {
      for (Iterator it = this.userList.iterator(); it.hasNext(); )
      {
        u = SmartFoxServer.getInstance().getUserById((Integer)it.next());

        if (u.getPlayerIndex(this) != pid)
          break label55;
        found = true;
        label55: break;
      }

    }

    if (found)
      return u;

    return null;
  }

  public String getXmlVarList()
  {
    if (this.roomVariables.size() == 0) {
      return "<vars />";
    }

    StringBuffer xml = new StringBuffer("<vars>");

    synchronized (this.roomVariables)
    {
      Set varList = this.roomVariables.keySet();

      Iterator it = varList.iterator();

      while (it.hasNext())
      {
        String varName = (String)it.next();
        RoomVariable rv = (RoomVariable)this.roomVariables.get(varName);

        xml.append("<var n='").append(varName);
        xml.append("' t='").append(rv.getType());
        xml.append("'><![CDATA[").append(rv.getValue()).append("]]></var>");
      }
    }

    xml.append("</vars>");

    return xml.toString();
  }

  public String getXmlVariable(String varName)
  {
    StringBuffer xmlVar = new StringBuffer();

    RoomVariable rv = (RoomVariable)this.roomVariables.get(varName);

    if (rv != null)
    {
      xmlVar.append("<var n='").append(varName);
      xmlVar.append("' t='").append(rv.getType());
      xmlVar.append("'><![CDATA[").append(rv.getValue()).append("]]></var>");
    }

    return xmlVar.toString();
  }

  public LinkedList getVariableNames()
  {
    LinkedList ll;
    synchronized (this.roomVariables)
    {
      ll = new LinkedList((Collection)this.roomVariables.keySet());
    }

    return ll;
  }

  public HashMap getVariables()
  {
    return this.roomVariables;
  }

  public RoomVariable getVariable(String varName)
  {
    return ((RoomVariable)this.roomVariables.get(varName));
  }

  public int getId()
  {
    return this.id;
  }

  public String getZone()
  {
    return this.zone;
  }

  public String getPassword()
  {
    return this.password;
  }

  public LinkedList getChannellList()
  {
    LinkedList ll = new LinkedList();

    synchronized (this.userList)
    {
      Iterator it = this.userList.iterator();

      while (it.hasNext())
      {
        Integer id = (Integer)it.next();

        User u = SmartFoxServer.getInstance().getUserById(id);

        if (u != null)
        {
          ll.add(u.getChannel());
        }
      }
    }

    return ll;
  }

  public String getName()
  {
    return this.name;
  }

  public void setName(String name)
  {
    this.name = name;
  }

  public boolean isPrivate()
  {
    return this.isPrivate;
  }

  public boolean isTemp()
  {
    return this.isTemporary;
  }

  public boolean isGame()
  {
    return this.isGame;
  }

  public void setGame()
  {
    this.isTemporary = true;
    this.isGame = true;
  }

  public void setPassword(String pw)
  {
    this.password = pw;
  }

  public int getMaxUsers()
  {
    return this.maxUsers;
  }

  public int getUserCount()
  {
    return this.userCount;
  }

  public int getSpectatorCount()
  {
    return this.spectatorCount;
  }

  public boolean removeUser(User u, boolean updateUserRoomList, boolean destroyVars)
  {
    Integer uid = new Integer(u.getUserId());
    String userName = u.getName();

    boolean result = false;
    boolean exist = false;

    synchronized (this.userList)
    {
      exist = this.userList.contains(uid);
    }

    if (exist)
    {
      int playerIndex = u.getPlayerIndex(this);

      if (updateUserRoomList)
        u.removeRoom(this);

      if ((playerIndex > 0) && (isGame()))
      {
        removeUserSlot(playerIndex);
      }

      synchronized (this.userList)
      {
        result = this.userList.remove(uid);

        if (!(result))
        {
          System.out.println("Couldn't remove user: " + userName + ":" + uid + ", from room: " + this.name);
        }

      }

      if (result)
      {
        if (this.isGame)
        {
          if (playerIndex > 0)
            this.userCount -= 1;
          else
            this.spectatorCount -= 1;
        }
        else {
          this.userCount -= 1;
        }

      }

      if (ConfigData.debugInfo)
      {
        if ((this.userCount < 0) || (this.userCount > this.maxUsers))
          SmartFoxServer.log.fine("userCount out of range: " + this.userCount + ", Room: " + this.name + ", range: 0 - " + this.maxUsers);
        else if ((this.spectatorCount < 0) || (this.spectatorCount > this.maxSpectators)) {
          SmartFoxServer.log.fine("spectatorCount out of range: " + this.spectatorCount + ", Room: " + this.name + ", range: 0 - " + this.maxUsers);
        }

      }

      synchronized (this.userNames)
      {
        this.userNames.remove(userName);
      }

      Zone z = SmartFoxServer.getInstance().getZone(u.getZone());

      SystemHandler sh = SmartFoxServer.getInstance().getSysHandler();
      sh.notifyUserRemoval(u.getUserId(), this.id, u.getZone());

      updateUserCount();

      if (destroyVars)
      {
        int varCount = 0;

        StringBuffer varResponse = new StringBuffer();

        synchronized (this.roomVariables)
        {
          LinkedList varNames = new LinkedList((Collection)this.roomVariables.keySet());

          Iterator it = varNames.iterator();

          while (it.hasNext())
          {
            String vName = (String)it.next();

            RoomVariable rv = (RoomVariable)this.roomVariables.get(vName);

            User varOwner = rv.getOwner();

            if ((varOwner != null) && 
              (rv.getOwner() == u) && 
              (deleteVariable(vName, u)))
            {
              varResponse.append("<var n='").append(vName);
              varResponse.append("' t='x' />");

              ++varCount;

              SmartFoxServer.log.fine("Deleting var: " + vName + ", from room: " + getName());
            }

          }

        }

        if (varCount > 0)
        {
          StringBuffer response = new StringBuffer("<msg t='sys'><body action='rVarsUpdate' r='").append(this.id).append("'><vars>").append(varResponse).append("</vars></body></msg>");

          SmartFoxServer.getInstance().getEW().handleEvent(new ServerEvent(response.toString(), null, getChannellList()));
        }

      }

    }

    return result;
  }

  public boolean setVariable(String vn, String vt, String vv, boolean pr, boolean ps, User owner, boolean setOwnership)
  {
    boolean result = false;

    if (vn.equals("")) {
      return result;
    }

    if (vv == null)
      vv = "";

    RoomVariable var = null;

    synchronized (this.roomVariables)
    {
      var = (RoomVariable)this.roomVariables.get(vn);

      if (var != null)
      {
        if ((!(var.isPrivate())) || ((var.isPrivate()) && (var.getOwner() == owner)))
        {
          if (!(var.getValue().equals(vv)))
          {
            var.setValue(vv);
            result = true;
          }

          if (!(var.getType().equals(vt)))
          {
            var.setType(vt);
            result = true;
          }

          if (var.isPrivate() != pr)
          {
            var.setPrivate(pr);
            result = true;
          }

          if (var.isPersistent() != ps)
          {
            var.setPersistent(ps);
            result = true;
          }

          if (setOwnership) {
            var.setOwner(owner);
          }

        }

      }
      else if ((this.roomVariables.size() <= SmartFoxServer.getMAX_ROOM_VARS()) || (SmartFoxServer.getMAX_ROOM_VARS() == -1))
      {
        var = new RoomVariable(vv, vt, owner, ps, pr);
        this.roomVariables.put(vn, var);

        result = true;
      }
      else
      {
        SmartFoxServer.log.warning("Variables limit exceeded in room: " + this.name);
      }

    }

    if (result)
      SmartFoxServer.log.fine("Setting RoomVar [" + vn + "]: " + var.getValue() + ", priv: " + var.isPersistent() + ", persist: " + var.isPersistent() + ", in room: " + getName());

    return result;
  }

  public boolean deleteVariable(String vName, User owner)
  {
    boolean result = false;

    synchronized (this.roomVariables)
    {
      RoomVariable rv = (RoomVariable)this.roomVariables.get(vName);

      if (rv == null)
        break label82;

      if (!(rv.isPersistent())) break label41;
      return result;

      label41: if ((rv == null) || (
        (rv.isPrivate()) && (((!(rv.isPrivate())) || (owner != rv.getOwner())))))
        break label82;
      this.roomVariables.remove(vName);
      label82: result = true;
    }

    return result;
  }

  public void updateUserCount()
  {
    Zone z = SmartFoxServer.getInstance().getZone(getZone());

    if ((z.getCountUpdate()) && (!(z.isEventDisabled("onUserCountChange"))))
    {
      HashSet allUsers = new HashSet();
      LinkedList roomList = z.getRoomList();

      Iterator it = roomList.iterator();

      while (it.hasNext())
      {
        Room r = (Room)it.next();

        if ((r == this) && (isLimbo())) {
          continue;
        }

        if (r.getUcountUpdates()) {
          allUsers.addAll(r.getChannellList());
        }

      }

      StringBuffer response = new StringBuffer("<msg t='sys'><body action='uCount' r='").append(this.id);
      response.append("' u='").append(this.userCount);

      if (isGame())
        response.append("' s='").append(this.spectatorCount);

      response.append("'></body></msg>");

      SmartFoxServer.getInstance().getEW().handleEvent(new ServerEvent(response.toString(), null, new LinkedList(allUsers)));
    }
  }

  public int howManyUsers()
  {
    return this.userList.size();
  }

  public void setCreator(SocketChannel ch)
  {
    this.creator = ch;
  }

  public SocketChannel getCreator()
  {
    return this.creator;
  }

  public boolean contains(String name)
  {
    synchronized (this.userNames)
    {
      return this.userNames.contains(name);
    }
  }

  private void setRoomID()
  {
    this.id = (autoId++);

    if (autoId < 0)
      autoId = 0;
  }

  public static void resetRoomStaticData()
  {
    autoId = 0;
  }

  public boolean isLimbo()
  {
    return this.isLimbo;
  }

  public void setLimbo(boolean isLimbo)
  {
    this.isLimbo = isLimbo;
  }
}