package it.gotoandplay.smartfoxserver.data;

import it.gotoandplay.smartfoxserver.SmartFoxServer;
import it.gotoandplay.smartfoxserver.config.ConfigData;
import it.gotoandplay.smartfoxserver.util.Attachment;
import java.net.InetAddress;
import java.net.Socket;
import java.nio.channels.SelectionKey;
import java.nio.channels.SocketChannel;
import java.util.Collection;
import java.util.HashMap;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.Map;
import java.util.Set;
import java.util.concurrent.ConcurrentHashMap;
import java.util.logging.Logger;

public class User
{
  private static int autoId = 0;
  private int uid;
  private String zone;
  private int dbId;
  private String name;
  private SocketChannel channel;
  private long lastMessageTime;
  private long loginTime;
  private boolean isChallenging;
  private boolean isPlaying;
  private boolean admin;
  private boolean moderator;
  private boolean buddyListChanged;
  private int userPrivileges;
  private LinkedList roomsConnected;
  private LinkedList roomsCreated;
  private HashMap userVars;
  private HashMap playerIndexes;
  private String ipAddr;
  private long droppedPackets;
  private boolean markedForRemoval;
  private boolean blueBoxed;
  private String bbSessionId;
  public boolean isPlayer;
  public boolean isBeingKicked;
  public int playerIndex;
  public int floodCounter;
  public int repeatedMsgCounter;
  public int floodWarningsCounter;
  public int badwordsWarningCounter;
  public String lastMessage;
  public Map properties;
  public Map<String, String> buddyVariables;

  public User(SocketChannel channel, String name, String zone)
  {
    setUserID();
    initUser();

    this.channel = channel;
    this.zone = zone;

    if (name.length() < 1)
    {
      this.name = "Guest_" + this.uid;
    }
    else
    {
      this.name = name;
    }

    this.roomsConnected = new LinkedList();
    this.roomsCreated = new LinkedList();
    this.userVars = new HashMap();
    this.playerIndexes = new HashMap();
    try
    {
      this.properties = ((Map)Class.forName(ConfigData.dynamicPropertiesClass).newInstance());
    }
    catch (Exception e)
    {
      this.properties = new HashMap();
    }

    this.buddyVariables = new ConcurrentHashMap();

    this.admin = false;
    this.moderator = false;
    this.isBeingKicked = false;
    this.markedForRemoval = false;

    this.lastMessageTime = System.currentTimeMillis();
    this.floodCounter = 0;
    this.repeatedMsgCounter = 0;
    this.floodWarningsCounter = 0;
    this.badwordsWarningCounter = 0;
    this.droppedPackets = 0L;

    this.ipAddr = this.channel.socket().getInetAddress().getHostAddress();

    SelectionKey sk = channel.keyFor(SmartFoxServer.getInstance().getReadSelector());
    Attachment attachment = (Attachment)sk.attachment();

    this.blueBoxed = attachment.blueBoxed;
    this.bbSessionId = attachment.bbSessionId;
  }

  public void initUser()
  {
    long t = System.currentTimeMillis();

    this.floodCounter = 0;
    this.loginTime = t;
    this.lastMessageTime = t;
    this.buddyListChanged = false;
  }

  public long getDroppedPackets()
  {
    return this.droppedPackets;
  }

  public void increaseDroppedPackets()
  {
    this.droppedPackets += 1L;
  }

  public boolean buddyListChanged()
  {
    return this.buddyListChanged;
  }

  public void setBuddyListChanged()
  {
    this.buddyListChanged = true;
  }

  public void setBuddyListSaved()
  {
    this.buddyListChanged = false;
  }

  public boolean isSpectator()
  {
    return (getPlayerIndex() == -1);
  }

  public void setPlayerIndex(Room r, int id)
  {
    synchronized (this.playerIndexes)
    {
      this.playerIndexes.put(r, new Integer(id));
    }
  }

  public int getPlayerIndex()
  {
    Room r = null;

    if (this.roomsConnected.size() > 0)
      r = (Room)this.roomsConnected.getLast();

    return getPlayerIndex(r);
  }

  public int getPlayerIndex(Room r)
  {
    if ((r != null) && (r.isGame()))
    {
      int res = -1;

      Integer id = (Integer)this.playerIndexes.get(r);

      if (id != null)
        res = id.intValue();

      return res;
    }

    return 0;
  }

  public void exitAllRooms()
  {
    Iterator it;
    Room r;
    synchronized (this.roomsConnected)
    {
      it = this.roomsConnected.iterator();

      while (it.hasNext())
      {
        r = (Room)it.next();

        boolean removed = r.removeUser(this, false, false);

        if (!(removed))
        {
          SmartFoxServer.log.warning("Problems during user removal: " + getName());
        }
        else if ((r.howManyUsers() < 1) && (r.isTemp()))
        {
          if (!(r.isGame()))
          {
            if (SmartFoxServer.getInstance().getUserByChannel(r.getCreator()) == null)
            {
              SmartFoxServer.getInstance().getZone(getZone()).removeRoom(r, false);
              SmartFoxServer.log.info("Dynamic Room Deleted: " + r.getName());
            }

          }
          else
          {
            SmartFoxServer.getInstance().getZone(getZone()).removeRoom(r, false);
            SmartFoxServer.log.info("Game Room Deleted: " + r.getName());
          }

        }

      }

    }

    synchronized (this.roomsCreated)
    {
      it = this.roomsCreated.iterator();

      while (it.hasNext())
      {
        r = (Room)it.next();

        if ((r.howManyUsers() < 1) && (r.isTemp()) && 
          (!(r.isGame())))
        {
          SmartFoxServer.getInstance().getZone(getZone()).removeRoom(r, true);
          SmartFoxServer.log.info("Dynamic Room Deleted: " + r.getName());
        }
      }
    }
  }

  public void addRoom(Room r)
  {
    synchronized (this.roomsConnected)
    {
      this.roomsConnected.add(r);
    }
  }

  public int[] getRoomsConnected()
  {
    int[] roomIds = new int[this.roomsConnected.size()];
    int c = 0;

    synchronized (this.roomsConnected)
    {
      for (Iterator i = this.roomsConnected.iterator(); i.hasNext(); )
      {
        roomIds[(c++)] = ((Room)i.next()).getId();
      }
    }

    return roomIds;
  }

  public int getRoomsConnectedCount()
  {
    return this.roomsConnected.size();
  }

  public int getRoomsCreatedCount()
  {
    return this.roomsCreated.size();
  }

  public int getRoom()
  {
    int res = -1;
    try
    {
      r = (Room)this.roomsConnected.get(0);
      res = r.getId();
    }
    catch (java.lang.IndexOutOfBoundsException r)
    {
    }

    return res;
  }

  public void setPrivilege(int p)
  {
    this.userPrivileges = p;
  }

  public int getPrivilege()
  {
    return this.userPrivileges;
  }

  public boolean isModerator()
  {
    return this.moderator;
  }

  public void setAsModerator()
  {
    this.moderator = true;
  }

  public void setAsModerator(boolean b)
  {
    this.moderator = b;
  }

  public boolean isAdmin()
  {
    return this.admin;
  }

  public void setAsAdmin()
  {
    this.admin = true;
  }

  public void removeRoom(Room r)
  {
    synchronized (this.roomsConnected)
    {
      this.roomsConnected.remove(r);
    }

    synchronized (this.playerIndexes)
    {
      this.playerIndexes.remove(r);
    }
  }

  public void addCreatedRoom(Room r)
  {
    synchronized (this.roomsCreated)
    {
      this.roomsCreated.add(r);
    }
  }

  public void removeCreatedRoom(Room r)
  {
    synchronized (this.roomsCreated)
    {
      this.roomsCreated.remove(r);
    }
  }

  public int getUserId()
  {
    return this.uid;
  }

  public SocketChannel getChannel()
  {
    return this.channel;
  }

  public String getName()
  {
    return this.name;
  }

  public String getZone()
  {
    return this.zone;
  }

  public long getLastMessageTime()
  {
    return this.lastMessageTime;
  }

  public long getLoginTime()
  {
    return this.loginTime;
  }

  public void updateMessageTime()
  {
    this.lastMessageTime = System.currentTimeMillis();
  }

  private void setUserID()
  {
    this.uid = (autoId++);

    if (autoId >= 2147483647)
      autoId = 0;
  }

  public boolean setVariable(String vn, String vv, String vt)
  {
    boolean ok = false;

    if (vn.equals(""))
      return false;

    UserVariable uv = null;

    synchronized (this.userVars)
    {
      uv = (UserVariable)this.userVars.get(vn);

      if (uv != null)
      {
        uv.setType(vt);
        uv.setValue(vv);

        ok = true;
      }
      else if ((this.userVars.size() < SmartFoxServer.getMAX_USER_VARS()) || (SmartFoxServer.getMAX_USER_VARS() == -1))
      {
        this.userVars.put(vn, new UserVariable(vv, vt));
        SmartFoxServer.log.fine("User variable [" + vn + "] = " + vv + ", created for user: " + this.name);
        ok = true;
      }
      else
      {
        SmartFoxServer.log.warning("UserVariable limit exceeded for user: " + this.name);
      }

    }

    return ok;
  }

  public void deleteVariable(String name)
  {
    synchronized (this.userVars)
    {
      this.userVars.remove(name);
      SmartFoxServer.log.fine("User variable [" + name + "] deleted for user: " + getName());
    }
  }

  public boolean isBusyForChallenge()
  {
    return this.isChallenging;
  }

  public void setBusyForChallenge(boolean status)
  {
    this.isChallenging = status;
  }

  public boolean isBusyForPlay()
  {
    return this.isPlaying;
  }

  public void setIsBusyForPlay(boolean status)
  {
    this.isPlaying = status;
  }

  public String getXmlVarList()
  {
    StringBuffer xmlList = new StringBuffer("<vars>");

    synchronized (this.userVars)
    {
      Set s = this.userVars.keySet();

      Iterator it = s.iterator();

      while (it.hasNext())
      {
        String vName = (String)it.next();

        UserVariable uv = (UserVariable)this.userVars.get(vName);

        xmlList.append("<var n='").append(vName);
        xmlList.append("' t='").append(uv.getType());
        xmlList.append("'><![CDATA[").append(uv.getValue()).append("]]></var>");
      }
    }

    xmlList.append("</vars>");

    return xmlList.toString();
  }

  public UserVariable getVariable(String varName)
  {
    return ((UserVariable)this.userVars.get(varName));
  }

  public HashMap getVariables()
  {
    return this.userVars;
  }

  public LinkedList getVariableNames()
  {
    LinkedList ll;
    synchronized (this.userVars)
    {
      ll = new LinkedList((Collection)this.userVars.keySet());
    }

    return ll;
  }

  public String getIpAddress()
  {
    return this.ipAddr;
  }

  public static void resetUserStaticData()
  {
    autoId = 0;
  }

  public boolean isInRoom(int roomId)
  {
    boolean response = false;

    synchronized (this.roomsConnected)
    {
      Room r = null;

      for (Iterator it = this.roomsConnected.iterator(); it.hasNext(); )
      {
        r = (Room)it.next();

        if ((r == null) || (r.getId() != roomId))
          break label55;
        response = true;
        label55: break;
      }

    }

    return response;
  }

  public boolean isMarkedForRemoval()
  {
    return this.markedForRemoval;
  }

  public void setMarkedForRemoval(boolean b)
  {
    this.markedForRemoval = b;
  }

  public boolean isBlueBoxed()
  {
    return this.blueBoxed;
  }

  public String getBlueBoxSessionId()
  {
    return this.bbSessionId;
  }
}