package it.gotoandplay.smartfoxserver.data;

import it.gotoandplay.smartfoxserver.EventWriter;
import it.gotoandplay.smartfoxserver.SmartFoxServer;
import it.gotoandplay.smartfoxserver.controllers.SystemHandler;
import it.gotoandplay.smartfoxserver.data.buddylist.BuddyListManager;
import it.gotoandplay.smartfoxserver.data.buddylist.LoadableBuddyList;
import it.gotoandplay.smartfoxserver.data.buddylist.persistence.IBuddyListPersister;
import it.gotoandplay.smartfoxserver.db.DbManager;
import it.gotoandplay.smartfoxserver.exceptions.BuddyListException;
import it.gotoandplay.smartfoxserver.exceptions.CreateRoomException;
import it.gotoandplay.smartfoxserver.extensions.AbstractExtension;
import it.gotoandplay.smartfoxserver.extensions.ExtensionManager;
import it.gotoandplay.smartfoxserver.lib.IService;
import it.gotoandplay.smartfoxserver.lib.ServerEvent;
import java.nio.channels.SocketChannel;
import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.logging.Logger;

public class Zone
  implements IService
{
  public static final int STATUS_ACTIVE = 0;
  public static final int STATUS_INACTIVE = 1;
  public String userNameAvoidChars;
  public String roomNameAvoidChars;
  private String name;
  private HashMap roomList;
  private int autoJoinRoomId;
  private HashMap userNames;
  private boolean customLogin;
  private boolean isDeactivating;
  private boolean updateUserCount;
  private int maxUsers;
  private int maxRooms;
  private int maxRoomNameLen;
  private int maxUserNameLen;
  private boolean buddyList;
  private boolean emptyNames;
  private int maxRoomsPerUser;
  private boolean sendVarsInRoomList;
  private boolean roomUpdates;
  private boolean receivePubMsgInternalEvents;
  private boolean receivePrivMsgInternalEvents;
  private boolean autoReloadExtensions;
  private int status;
  public ModeratorManager modManager;
  private ExtensionManager extManager;
  private BuddyListManager buddyManager;
  private Map disabledEvents;
  private Map disabledSysActions;
  public DbManager dbManager;

  public Zone(String name, String customLogin)
  {
    this.name = name;
    this.roomList = new HashMap();
    this.userNames = new HashMap();
    this.autoJoinRoomId = -1;

    this.maxRooms = -1;
    this.maxUsers = -1;
    this.updateUserCount = true;
    this.buddyList = false;
    this.isDeactivating = false;
    this.emptyNames = true;
    this.status = 0;
    this.maxRoomsPerUser = 5;
    this.roomUpdates = true;
    this.maxRoomNameLen = 20;
    this.maxUserNameLen = 60;
    this.receivePubMsgInternalEvents = false;
    this.autoReloadExtensions = false;

    if ("true".equalsIgnoreCase(customLogin))
      this.customLogin = true;
    else
      this.customLogin = false;

    this.dbManager = null;
    this.modManager = new ModeratorManager();
    this.extManager = new ExtensionManager();

    this.disabledEvents = new HashMap();
    this.disabledSysActions = new HashMap();
  }

  public void init(Object param)
  {
  }

  public void destroy(Object param)
  {
    saveAllBuddyLists();
  }

  public void addDisabledEvent(String evt)
  {
    this.disabledEvents.put(evt.toLowerCase(), new Object());
  }

  public boolean isEventDisabled(String evt)
  {
    return (this.disabledEvents.get(evt.toLowerCase()) != null);
  }

  public void clearDisabledEvents()
  {
    this.disabledEvents.clear();
  }

  public void addDisabledSysAction(String action)
  {
    this.disabledSysActions.put(action.toLowerCase(), new Object());
  }

  public boolean isSysActionDisabled(String action)
  {
    return (this.disabledSysActions.get(action.toLowerCase()) != null);
  }

  public void clearDisabledSysActions()
  {
    this.disabledSysActions.clear();
  }

  public boolean isPubMsgInternalEventEnabled()
  {
    return this.receivePubMsgInternalEvents;
  }

  public void setPubMsgInternalEvent(boolean b)
  {
    this.receivePubMsgInternalEvents = b;
  }

  public boolean isPrivMsgInternalEventEnabled()
  {
    return this.receivePrivMsgInternalEvents;
  }

  public void setPrivMsgInternalEvent(boolean b)
  {
    this.receivePrivMsgInternalEvents = b;
  }

  public String getName()
  {
    return this.name;
  }

  public boolean hasCustomLogin()
  {
    return this.customLogin;
  }

  public int getStatus()
  {
    return this.status;
  }

  public boolean getVarsOnRoomList()
  {
    return this.sendVarsInRoomList;
  }

  public void setVarsOnRoomList(boolean b)
  {
    this.sendVarsInRoomList = b;
  }

  public boolean getRoomUpdates()
  {
    return this.roomUpdates;
  }

  public void setRoomUpdates(boolean b)
  {
    this.roomUpdates = b;
  }

  public ExtensionManager getExtManager()
  {
    return this.extManager;
  }

  public BuddyListManager getBuddyManager()
  {
    return this.buddyManager;
  }

  public boolean isActive()
  {
    return (this.status == 0);
  }

  public int getMaxRoomsPerUser()
  {
    return this.maxRoomsPerUser;
  }

  public void setMaxRoomsPerUser(int m)
  {
    this.maxRoomsPerUser = m;
  }

  public boolean validateUserName(String name)
  {
    boolean res = false;

    synchronized (this.userNames)
    {
      res = this.userNames.containsKey(name);
    }

    return res;
  }

  public void addName(String name, Integer uid)
  {
    synchronized (this.userNames)
    {
      this.userNames.put(name, uid);
    }

    if (hasBuddyList())
    {
      try
      {
        loadBuddyList(name);

        SmartFoxServer.getInstance()
          .getSysHandler()
          .sendBuddyListUpdate(
          this.buddyManager.getClientsToUpdate(name), 
          name, 
          uid.intValue(), 
          true, 
          true);
      }
      catch (BuddyListException e)
      {
        SmartFoxServer.log.warning("Error while loading buddy list for user: " + name + ". Reason: " + e.getMessage());
        e.printStackTrace();
      }
    }
  }

  public void removeName(String name, User u)
  {
    if (hasBuddyList())
    {
      if (u.buddyListChanged())
        this.buddyManager.saveBuddyList(name);

      this.buddyManager.removeBuddyList(name);

      SmartFoxServer.getInstance().getSysHandler().sendBuddyListUpdate(this.buddyManager.getClientsToUpdate(name), name, -1, false, false);
    }

    synchronized (this.userNames)
    {
      this.userNames.remove(name);
    }
  }

  public void destroyVariables(User u)
  {
    synchronized (this.roomList)
    {
      Collection c = this.roomList.values();

      for (Iterator i = c.iterator(); i.hasNext(); )
      {
        Room r = (Room)i.next();

        int varCount = 0;

        StringBuffer varResponse = new StringBuffer();

        HashMap roomVars = r.getVariables();

        synchronized (roomVars)
        {
          LinkedList vNames = new LinkedList((Collection)roomVars.keySet());

          for (Iterator v = vNames.iterator(); v.hasNext(); )
          {
            String varName = (String)v.next();

            RoomVariable rv = (RoomVariable)roomVars.get(varName);

            if (rv.getOwner() != u)
              break label194;
            varResponse.append("<var n='").append(varName);
            varResponse.append("' t='x' />");

            ++varCount;

            SmartFoxServer.log.fine("Deleting var: " + varName + 
              ", from room: " + r.getName());

            label194: roomVars.remove(varName);
          }

        }

        if (varCount <= 0)
          break label282;
        StringBuffer response = new StringBuffer(
          "<msg t='sys'><body action='rVarsUpdate' r='")
          .append(r.getId()).append("'><vars>").append(
          varResponse).append("</vars></body></msg>");

        label282: SmartFoxServer.getInstance().getEW().handleEvent(
          new ServerEvent(response.toString(), null, 
          r.getChannellList()));
      }
    }
  }

  public void clearAllExtensions()
  {
    getExtManager().shutDown();

    synchronized (this.roomList)
    {
      for (Iterator i = this.roomList.values().iterator(); i.hasNext(); )
      {
        Room r = (Room)i.next();

        if (r == null) break label51;
        label51: r.getExtManager().shutDown();
      }
    }
  }

  public AbstractExtension getExtension(String name)
  {
    return ((AbstractExtension)this.extManager.get(name));
  }

  public SocketChannel getChannelFromName(String name)
  {
    User u = getUserByName(name);

    if (u != null)
      return u.getChannel();

    return null;
  }

  public User getUserByName(String name)
  {
    return SmartFoxServer.getInstance().getUserById(getUserIdByName(name));
  }

  public Integer getUserIdByName(String name)
  {
    return ((Integer)this.userNames.get(name));
  }

  public void addRoom(Room rm)
    throws CreateRoomException
  {
    int i = rm.getId();

    if (this.maxRooms == -1)
    {
      synchronized (this.roomList)
      {
        this.roomList.put(new Integer(i), rm);
      }

    }

    if (this.maxRooms > this.roomList.size())
    {
      synchronized (this.roomList)
      {
        this.roomList.put(new Integer(i), rm);
      }

    }

    throw new CreateRoomException("The max. amount of rooms for this zone has been reached!");
  }

  public void removeRoom(Room rm, boolean exitingAllRooms)
  {
    int i = rm.getId();

    rm.destroyExtensions();

    SocketChannel ch = rm.getCreator();
    User u = SmartFoxServer.getInstance().getUserByChannel(ch);

    if ((u != null) && (!(exitingAllRooms)))
    {
      u.removeCreatedRoom(rm);
    }

    synchronized (this.roomList)
    {
      this.roomList.remove(new Integer(i));
    }

    SmartFoxServer.getInstance().getSysHandler().notifyRoomDeleted(this, i);

    SmartFoxServer.getInstance().getSysHandler().broadcastRoomDestroyed(i, this.name);
  }

  public void deactivate()
  {
    if ((!(this.isDeactivating)) && (this.status == 0))
    {
      this.status = 1;

      this.isDeactivating = true;

      LinkedList ll = getRoomList();

      for (Iterator i = ll.iterator(); i.hasNext(); )
      {
        Room r = (Room)i.next();

        Object[] userList = r.getUserList();

        for (int id = 0; id < userList.length; ++id) {
          SmartFoxServer.getInstance().lostConnection((Integer)userList[id]);
        }

        ExtensionManager em = r.getExtManager();

        if (em == null) break label98;
        label98: em.deactivateAll();
      }

      getExtManager().deactivateAll();

      this.isDeactivating = false;
    }
  }

  public void activate()
  {
    if ((!(this.isDeactivating)) && (this.status == 1))
    {
      getExtManager().activateAll();

      synchronized (this.roomList)
      {
        for (Iterator i = this.roomList.values().iterator(); i.hasNext(); )
        {
          Room r = (Room)i.next();

          r.getExtManager().activateAll();
        }
      }

      this.status = 0;
    }
  }

  public void setAutoJoinRoom(int roomId)
  {
    this.autoJoinRoomId = roomId;
  }

  public int getAutoJoinRoom()
  {
    return this.autoJoinRoomId;
  }

  public Room getRoom(int roomId)
  {
    return ((Room)this.roomList.get(new Integer(roomId)));
  }

  public Room getRoomByName(String name)
  {
    Room r = null;
    boolean found = false;

    synchronized (this.roomList)
    {
      for (Iterator i = this.roomList.values().iterator(); i.hasNext(); )
      {
        r = (Room)i.next();

        if (!(name.equals(r.getName())))
          break label56;
        found = true;
        label56: break;
      }

    }

    if (found)
      return r;

    return null;
  }

  public LinkedList getRoomList()
  {
    LinkedList list;
    synchronized (this.roomList)
    {
      list = new LinkedList(this.roomList.values());
    }

    return list;
  }

  public Object[] getRooms()
  {
    Object[] rList = new Room[this.roomList.size()];

    synchronized (this.roomList)
    {
      rList = this.roomList.values().toArray();
    }

    return rList;
  }

  public boolean roomNameAlreadyExist(String roomName)
  {
    LinkedList rmList = getRoomList();

    Iterator it = rmList.iterator();

    while (it.hasNext())
    {
      if (((Room)it.next()).getName().equals(roomName))
        return true;
    }

    return false;
  }

  public int getRoomCount()
  {
    return this.roomList.size();
  }

  public int getUserCount()
  {
    return this.userNames.size();
  }

  public boolean isThereRoom()
  {
    boolean response = true;

    if ((this.maxUsers > -1) && 
      (this.maxUsers <= this.userNames.size())) {
      response = false;
    }

    return response;
  }

  public LinkedList getAllUsersInZone()
  {
    LinkedList allRooms = getRoomList();

    HashSet allChannells = new HashSet();

    Iterator it = allRooms.iterator();

    while (it.hasNext())
    {
      Room tempRoom = (Room)it.next();

      if (!(tempRoom.isGame()))
      {
        allChannells.addAll(tempRoom.getChannellList());
      }

    }

    return new LinkedList(allChannells);
  }

  public LinkedList getChannelList()
  {
    LinkedList ll = new LinkedList();

    synchronized (this.userNames)
    {
      Collection c = this.userNames.values();

      for (Iterator i = c.iterator(); i.hasNext(); )
      {
        Integer id = (Integer)i.next();
        SocketChannel sc = SmartFoxServer.getInstance().getUserById(id).getChannel();

        ll.add(sc);
      }
    }

    return ll;
  }

  public List getUserList()
  {
    List list = new LinkedList();

    synchronized (this.userNames)
    {
      Collection c = this.userNames.values();

      for (Iterator i = c.iterator(); i.hasNext(); )
      {
        Integer id = (Integer)i.next();
        User user = SmartFoxServer.getInstance().getUserById(id);

        list.add(user);
      }
    }

    return list;
  }

  public boolean hasBuddyList()
  {
    return this.buddyList;
  }

  public void initBuddyList(int listLength, IBuddyListPersister persister)
  {
    if (listLength > 0)
    {
      this.buddyManager = new BuddyListManager(getName(), listLength, persister);
      this.buddyList = true;
    }
    else
    {
      SmartFoxServer.log.warning("Can't initialize BuddyListManager with passed list-size of: " + listLength);
    }
  }

  public void addBuddy(String owner, String buddyName)
    throws BuddyListException
  {
    if (hasBuddyList())
      this.buddyManager.addBuddy(owner, buddyName);
  }

  public boolean removeBuddy(String owner, String buddyName)
  {
    if (hasBuddyList())
      return this.buddyManager.removeBuddy(owner, buddyName);

    return false;
  }

  public LoadableBuddyList loadBuddyList(String owner)
    throws BuddyListException
  {
    if (hasBuddyList())
      return this.buddyManager.loadBuddyList(owner);

    return null;
  }

  public void saveBuddyList(String name)
  {
    if (hasBuddyList())
    {
      this.buddyManager.saveBuddyList(name);
    }
  }

  public void saveAllBuddyLists()
  {
    if (hasBuddyList())
      this.buddyManager.destroy(null);
  }

  /**
   * @deprecated
   */
  public void clearBuddyList(String owner)
  {
    if (hasBuddyList())
    {
      this.buddyManager.clearBuddyList(owner);
    }
  }

  public boolean checkBuddy(String name)
  {
    return validateUserName(name);
  }

  public String getXmlBuddy(String name)
  {
    StringBuffer xmlBuddy = new StringBuffer();

    String status = (checkBuddy(name)) ? "1" : "0";

    xmlBuddy.append("<b s='").append(status);
    xmlBuddy.append("'><![CDATA[").append(name).append("]]></b>");

    return xmlBuddy.toString();
  }

  public boolean getCountUpdate()
  {
    return this.updateUserCount;
  }

  public void setCountUpdate(boolean b)
  {
    this.updateUserCount = b;
  }

  public int getMaxUsers()
  {
    return this.maxUsers;
  }

  public void setMaxUsers(int i)
  {
    if (i == 0)
      i = 1;

    this.maxUsers = i;
  }

  public int getMaxRooms()
  {
    return this.maxRooms;
  }

  public void setMaxRooms(int i)
  {
    this.maxRooms = i;
  }

  public void setEmptyNames(boolean b)
  {
    this.emptyNames = b;
  }

  public boolean getEmptyNames()
  {
    return this.emptyNames;
  }

  public int getMaxRoomNameLen()
  {
    return this.maxRoomNameLen;
  }

  public void setMaxRoomNameLen(int maxRoomNameLen)
  {
    this.maxRoomNameLen = maxRoomNameLen;
  }

  public int getMaxUserNameLen()
  {
    return this.maxUserNameLen;
  }

  public void setMaxUserNameLen(int maxUserNameLen)
  {
    this.maxUserNameLen = maxUserNameLen;
  }

  public boolean isAutoReloadExtensions()
  {
    return this.autoReloadExtensions;
  }

  public void setAutoReloadExtensions(boolean autoReloadExtensions)
  {
    this.autoReloadExtensions = autoReloadExtensions;
  }
}