package it.gotoandplay.smartfoxserver.data.buddylist.persistence;

import it.gotoandplay.smartfoxserver.SmartFoxServer;
import it.gotoandplay.smartfoxserver.data.buddylist.BuddyListManager;
import it.gotoandplay.smartfoxserver.data.buddylist.StorableBuddyItem;
import it.gotoandplay.smartfoxserver.data.buddylist.StorableBuddyList;
import it.gotoandplay.smartfoxserver.db.DataRow;
import it.gotoandplay.smartfoxserver.db.DbManager;
import it.gotoandplay.smartfoxserver.lib.SmartFoxLib;
import java.io.PrintStream;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.concurrent.ConcurrentHashMap;
import java.util.logging.Logger;

public class AdvancedBuddyListPersister extends AbstractBuddyPersister
{
  private static final String SQL_CHECK_TABLE = "SELECT * FROM INFORMATION_SCHEMA.TABLES WHERE TABLE_NAME=";
  private static final String TABLE_PREFIX = "BLIST_";
  private static final byte[] TOKEN = { 1 };
  private static final String DATA_SEPARATOR = new String(TOKEN);
  private Logger log;
  private DbManager h2db;
  private String hexTableName;

  public void init(Object param)
  {
    SmartFoxServer.log.fine("AdvancedBuddyListPersister <INIT>");
    this.h2db = SmartFoxServer.getInstance().getCoreDbManager();
    this.log = SmartFoxServer.log;

    this.hexTableName = "BLIST_" + SmartFoxLib.getHexFileName(this.zoneName);
    this.hexTableName = this.hexTableName.toUpperCase();

    checkIfTableExists();
  }

  public void destroy(Object param)
  {
    this.log.fine("AdvancedBuddyListPersister <SHUT DOWN>");
  }

  public StorableBuddyList loadList(String userName)
  {
    StorableBuddyList buddyList = null;

    StringBuilder sql = new StringBuilder("SELECT * FROM ").append(this.hexTableName);
    sql.append(" WHERE OWNER='" + SmartFoxLib.escapeQuotes(userName) + "'");
    try
    {
      ArrayList resultSet = this.h2db.executeQuery(sql.toString());

      if ((resultSet == null) || (resultSet.size() <= 0))
        break label167;
      buddyList = new StorableBuddyList(userName);
      populateBuddyList(userName, buddyList, (DataRow)resultSet.get(0));
      this.log.fine("Buddy List loaded --> " + userName);
    }
    catch (Exception e)
    {
      this.log.warning("Error loading buddy list for user: " + userName + ". Reason: " + e.getMessage());
      e.printStackTrace();
    }

    label167: return buddyList;
  }

  public boolean saveList(String userName, StorableBuddyList buddyList)
  {
    int i;
    boolean success = false;

    StringBuilder namesList = new StringBuilder();
    StringBuilder blockList = new StringBuilder();
    StringBuilder varKeys = new StringBuilder();
    StringBuilder varValues = new StringBuilder();

    synchronized (buddyList.buddies)
    {
      for (i = 0; i < buddyList.buddies.size(); ++i)
      {
        namesList.append(((StorableBuddyItem)buddyList.buddies.get(i)).name).append(DATA_SEPARATOR);
        blockList.append((((StorableBuddyItem)buddyList.buddies.get(i)).isBlocked) ? "1" : "0").append(DATA_SEPARATOR);
      }
    }

    Map vars = getBuddyManager().getBuddyVariables(userName);

    for (Map.Entry entry : vars.entrySet())
    {
      varKeys.append((String)entry.getKey()).append(DATA_SEPARATOR);
      varValues.append((String)entry.getValue()).append(DATA_SEPARATOR);
    }

    try
    {
      StringBuilder sql = new StringBuilder();

      String names_list = (namesList.length() > 1) ? SmartFoxLib.escapeQuotes(namesList.substring(0, namesList.length() - 1)) : "";
      String block_list = (blockList.length() > 1) ? SmartFoxLib.escapeQuotes(blockList.substring(0, blockList.length() - 1)) : "";
      String var_keys = (varKeys.length() > 1) ? SmartFoxLib.escapeQuotes(varKeys.substring(0, varKeys.length() - 1)) : "";
      String var_values = (varValues.length() > 1) ? SmartFoxLib.escapeQuotes(varValues.substring(0, varValues.length() - 1)) : "";

      if (listAlreadyExist(userName))
      {
        sql.append("UPDATE ").append(this.hexTableName).append(" SET ");
        sql.append("NAMES_LIST='").append(names_list).append("',");
        sql.append("BLOCK_LIST='").append(block_list).append("',");
        sql.append("VAR_KEYS='").append(var_keys).append("',");
        sql.append("VAR_VALUES='").append(var_values);

        sql.append("' WHERE OWNER='").append(userName).append("'");
      }
      else
      {
        sql.append("INSERT INTO ")
          .append(this.hexTableName)
          .append(" VALUES('")
          .append(SmartFoxLib.escapeQuotes(userName)).append("','")
          .append(names_list).append("','")
          .append(block_list).append("','")
          .append(var_keys).append("','")
          .append(var_values).append("')");
      }

      success = this.h2db.executeCommand(sql.toString());
      this.log.finer("Buddy list stored: " + sql);
    }
    catch (Exception error)
    {
      error.printStackTrace();
      this.log.warning("Error while saving buddy list for user: " + userName + ". Reason: " + error.getMessage());
    }

    return success;
  }

  public Map<String, String> getOfflineVariables(String userName)
  {
    Map offlineVars = null;

    StringBuilder sql = new StringBuilder("SELECT VAR_KEYS, VAR_VALUES FROM ")
      .append(this.hexTableName)
      .append(" WHERE OWNER='")
      .append(SmartFoxLib.escapeQuotes(userName))
      .append("'");

    ArrayList result = this.h2db.executeQuery(sql.toString());

    if ((result != null) && (result.size() > 0))
    {
      DataRow row = (DataRow)result.get(0);

      if (row != null)
      {
        offlineVars = new HashMap();

        String[] allKeys = row.getItem("VAR_KEYS").split(DATA_SEPARATOR);
        String[] allValues = row.getItem("VAR_VALUES").split(DATA_SEPARATOR);

        for (int i = 0; i < allKeys.length; ++i)
        {
          if (allKeys[i].startsWith(BuddyListManager.OFFLINE_VARS_PREFIX))
            offlineVars.put(allKeys[i], allValues[i]);
        }
      }
    }

    return offlineVars;
  }

  public boolean removeOfflineBuddy(String owner, String buddyName)
  {
    boolean success = false;

    StringBuilder sql = new StringBuilder("SELECT NAMES_LIST FROM ")
      .append(this.hexTableName)
      .append(" WHERE OWNER='")
      .append(SmartFoxLib.escapeQuotes(owner))
      .append("'");

    ArrayList result = this.h2db.executeQuery(sql.toString());

    if ((result != null) && (result.size() > 0))
    {
      DataRow row = (DataRow)result.get(0);
      String names = row.getItem("NAMES_LIST");
      int buddyNamePos = names.indexOf(buddyName);

      if (buddyNamePos > -1)
      {
        StringBuilder newNameList = new StringBuilder(names.substring(0, buddyNamePos));

        int endPoint = buddyNamePos + buddyName.length();
        if (endPoint < names.length()) {
          ++endPoint;
        }

        newNameList.append(names.substring(endPoint, names.length()));

        if ((newNameList.length() > 0) && (newNameList.charAt(newNameList.length() - 1) == DATA_SEPARATOR.charAt(0)))
        {
          newNameList.deleteCharAt(newNameList.length() - 1);
        }

        sql.delete(0, sql.length());

        sql.append("UPDATE ")
          .append(this.hexTableName)
          .append(" SET NAMES_LIST='")
          .append(SmartFoxLib.escapeQuotes(newNameList.toString()))
          .append("' WHERE OWNER='")
          .append(SmartFoxLib.escapeQuotes(owner))
          .append("'");

        success = this.h2db.executeCommand(sql.toString());

        this.log.fine("Removing off-line mutual buddy: " + sql.toString());
      }

    }

    return success;
  }

  private void checkIfTableExists()
  {
    String sql = "SELECT * FROM INFORMATION_SCHEMA.TABLES WHERE TABLE_NAME='" + this.hexTableName + "'";

    ArrayList res = this.h2db.executeQuery(sql);

    if (res != null)
    {
      if (res.size() == 0)
      {
        try
        {
          createBuddyListTable(this.hexTableName);
          this.log.finer("Table " + this.hexTableName + " created!");
        }
        catch (IllegalStateException error)
        {
          this.log.warning(error.getMessage());
        }

      }

    }
    else
      this.log.warning("Could not connect to H2 Engine, buddy list persister may not work!");
  }

  private void createBuddyListTable(String name)
    throws IllegalStateException
  {
    String sql = "CREATE TABLE " + name;
    sql = sql + "( OWNER VARCHAR(255) NOT NULL, NAMES_LIST LONGVARCHAR, BLOCK_LIST LONGVARCHAR, VAR_KEYS LONGVARCHAR, VAR_VALUES LONGVARCHAR )";

    boolean success = this.h2db.executeCommand(sql);

    if (!(success))
      throw new IllegalStateException("Can't create db table [" + name + "] for buddy list!");
  }

  private void populateBuddyList(String ownerName, StorableBuddyList buddyList, DataRow row)
    throws Exception
  {
    String[] blockList;
    String[] varKeys;
    String[] varValues;
    int i;
    String[] namesList = blockList = varKeys = varValues = (String[])null;

    if (row.getItem("NAMES_LIST").length() > 1)
    {
      namesList = row.getItem("NAMES_LIST").split(DATA_SEPARATOR);
      blockList = row.getItem("BLOCK_LIST").split(DATA_SEPARATOR);
    }

    if (row.getItem("VAR_KEYS").length() > 1)
    {
      varKeys = row.getItem("VAR_KEYS").split(DATA_SEPARATOR);
      varValues = row.getItem("VAR_VALUES").split(DATA_SEPARATOR);
    }

    ConcurrentHashMap variables = null;

    if ((namesList != null) && (namesList.length > 0))
    {
      for (i = 0; i < namesList.length; ++i)
        buddyList.buddies.add(new StorableBuddyItem(namesList[i], "1".equals(blockList[i])));

    }

    if ((varKeys != null) && (varKeys.length > 0))
    {
      variables = new ConcurrentHashMap();

      for (i = 0; i < varKeys.length; ++i) {
        variables.put(varKeys[i], varValues[i]);
      }

      getBuddyManager().setBuddyVariables(ownerName, variables);

      buddyList.variables = variables;
    }
  }

  private boolean listAlreadyExist(String userName)
    throws Exception
  {
    boolean exist = false;

    StringBuilder sql = new StringBuilder("SELECT OWNER FROM ")
      .append(this.hexTableName)
      .append(" WHERE OWNER='")
      .append(SmartFoxLib.escapeQuotes(userName))
      .append("'");

    ArrayList resultSet = this.h2db.executeQuery(sql.toString());

    if ((resultSet != null) && (resultSet.size() > 0))
      exist = true;

    return exist;
  }

  private void debugBuddyList(String owner, StorableBuddyList list, Map<String, String> vars)
  {
    System.out.println("\nBuddy list loaded --> " + owner);
    System.out.println("----------------------------------------------------");

    for (StorableBuddyItem item : list.buddies)
      System.out.println("\t" + item.name + ", block = " + item.isBlocked);

    System.out.println("\nVariables: " + ((vars != null) ? Integer.valueOf(vars.size()) : "n/a"));
    System.out.println("----------------------------------------------------");

    if (vars != null)
    {
      for (Map.Entry entry : vars.entrySet())
      {
        System.out.println("\t" + ((String)entry.getKey()) + " -> " + ((String)entry.getValue()));
      }
    }
  }
}