package it.gotoandplay.smartfoxserver.data.buddylist.persistence;

import it.gotoandplay.smartfoxserver.SmartFoxServer;
import it.gotoandplay.smartfoxserver.data.buddylist.StorableBuddyItem;
import it.gotoandplay.smartfoxserver.data.buddylist.StorableBuddyList;
import it.gotoandplay.smartfoxserver.exceptions.BuddyListException;
import it.gotoandplay.smartfoxserver.lib.SmartFoxLib;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.logging.Logger;

public class DefaultBuddyListPersister extends AbstractBuddyPersister
{
  public static final String MAIN_FOLDER = "./buddyLists";
  private static final String EXTENSION = ".list";
  private String buddyListPath;

  public void init(Object param)
  {
    this.buddyListPath = "./buddyLists/" + SmartFoxLib.getHexFileName(this.zoneName);
    try
    {
      SmartFoxLib.checkBuddyListFolders(this.buddyListPath);
      SmartFoxServer.log.info("Default Buddy List Persister <INITED>");
    }
    catch (BuddyListException exception)
    {
      SmartFoxServer.log
        .warning("Could not initialize DefaultBuddyPersister for zone: " + this.zoneName + ". " + this.zoneName.toString());
    }
  }

  public void destroy(Object param)
  {
    SmartFoxServer.log.info("DefaultBuddyList Persister <SHUT DOWN>");
  }

  public StorableBuddyList loadList(String userName)
  {
    String hexName = SmartFoxLib.getHexFileName(userName);
    LinkedList fileBasedBuddyList = null;
    try
    {
      fis = new FileInputStream(this.buddyListPath + "/" + hexName + ".list");
      ObjectInputStream ois = new ObjectInputStream(fis);

      fileBasedBuddyList = (LinkedList)ois.readObject();
      ois.close();
    }
    catch (java.io.FileNotFoundException fis)
    {
    }
    catch (IOException ioe)
    {
      SmartFoxServer.log.warning("I/O Error. Could not load buddy list for user: " + userName + " -> " + ioe);
    }
    catch (ClassNotFoundException cnfe)
    {
      SmartFoxServer.log.warning("Problems loading buddy list for user: " + userName + " -> " + cnfe);
    }

    return populateBuddyList(userName, fileBasedBuddyList);
  }

  public boolean saveList(String userName, StorableBuddyList buddyList)
  {
    String hexName = SmartFoxLib.getHexFileName(userName);
    boolean result = false;

    if (buddyList != null)
    {
      try
      {
        FileOutputStream fos = new FileOutputStream(this.buddyListPath + "/" + hexName + ".list");
        ObjectOutputStream oos = new ObjectOutputStream(fos);

        synchronized (buddyList.buddies)
        {
          oos.writeObject(getSerializableList(buddyList.buddies));
          oos.flush();
          oos.close();
        }

        result = true;

        SmartFoxServer.log.info("Written buddy list file: " + this.buddyListPath + "/" + hexName + ".list" + ", for user " + userName);
      }
      catch (IOException ioe)
      {
        SmartFoxServer.log.warning("IO exception while saving buddy list for user: " + userName + ", " + ioe);
      }

    }

    return result;
  }

  private StorableBuddyList populateBuddyList(String userName, List<String> fileBasedBuddyList)
  {
    if (fileBasedBuddyList == null)
      return null;

    StorableBuddyList buddyList = new StorableBuddyList();

    for (Iterator iterator = fileBasedBuddyList.iterator(); iterator.hasNext(); )
      buddyList.buddies.add(new StorableBuddyItem((String)iterator.next()));

    return buddyList;
  }

  private List<String> getSerializableList(List<StorableBuddyItem> listOfBuddies)
  {
    List namesList = new LinkedList();

    for (Iterator iterator = listOfBuddies.iterator(); iterator.hasNext(); )
      namesList.add(((StorableBuddyItem)iterator.next()).name);

    return namesList;
  }
}