package it.gotoandplay.smartfoxserver.extensions;

import it.gotoandplay.smartfoxserver.EventWriter;
import it.gotoandplay.smartfoxserver.SmartFoxServer;
import it.gotoandplay.smartfoxserver.config.ConfigData;
import it.gotoandplay.smartfoxserver.controllers.ExtensionHandler;
import it.gotoandplay.smartfoxserver.controllers.SystemHandler;
import it.gotoandplay.smartfoxserver.data.User;
import it.gotoandplay.smartfoxserver.events.IEventListener;
import it.gotoandplay.smartfoxserver.lib.ASObjectSerializer;
import it.gotoandplay.smartfoxserver.lib.ActionscriptObject;
import it.gotoandplay.smartfoxserver.lib.ConfigReader;
import it.gotoandplay.smartfoxserver.lib.ServerEvent;
import java.io.FileNotFoundException;
import java.io.PrintStream;
import java.util.LinkedList;
import java.util.logging.Logger;
import net.n3.nanoxml.IXMLElement;

public abstract class AbstractExtension
  implements ISmartFoxExtension, IEventListener
{
  protected String __zoneName;
  protected String __roomName;
  protected boolean __isActive;
  public AdminExtension adminExtension;

  public AbstractExtension()
  {
    this.__isActive = true;
    this.adminExtension = SmartFoxServer.getInstance().getExtensionHandler().getAdminExtension();
  }

  public void setOwner(String zoneName, String roomName)
  {
    this.__roomName = roomName;
    this.__zoneName = zoneName;
  }

  public String getOwnerZone()
  {
    return this.__zoneName;
  }

  public String getOwnerRoom()
  {
    return this.__roomName;
  }

  public void init()
  {
  }

  public void destroy()
  {
  }

  public boolean isActive()
  {
    return this.__isActive;
  }

  public void setActive(boolean b)
  {
    this.__isActive = b;
  }

  public void registerForEvents(String zoneName, String roomName)
  {
    setOwner(zoneName, roomName);
    SmartFoxServer.getInstance().getSysHandler().addListener(this);
  }

  public void unRegister()
  {
    SmartFoxServer.getInstance().getSysHandler().removeListener(this);
  }

  public void trace(String msg)
  {
    StringBuffer sb = new StringBuffer("[ ");
    sb.append(super.getClass().getName()).append(" ]: ").append(msg);

    this.adminExtension.logRemoteTrace(getOwnerZone(), getOwnerRoom(), sb.toString());

    System.out.println(sb);
  }

  public IXMLElement loadConfig(String xmlFile)
  {
    try
    {
      return ConfigReader.loadConfigFile(xmlFile);
    }
    catch (FileNotFoundException fnfe)
    {
      SmartFoxServer.log.severe("Can't load extension config file: " + xmlFile);
    }
    catch (Exception e)
    {
      SmartFoxServer.log.severe("Unexpected error during while loading extension config file: " + xmlFile);
    }

    return null;
  }

  public void handleRequest(String cmd, org.json.JSONObject jso, User u, int fromRoom)
  {
  }

  public Object handleInternalRequest(Object params)
  {
    return null;
  }

  public void sendResponse(ActionscriptObject ao, int fromRoom, User sender, LinkedList recipients)
  {
    StringBuffer xmlMsg = new StringBuffer("<msg t='xt'><body action='xtRes' r='").append(fromRoom).append("'>");

    xmlMsg.append("<![CDATA[").append(ASObjectSerializer.serialize(ao)).append("]]>");
    xmlMsg.append("</body></msg>");

    SmartFoxServer.getInstance().getEW().handleEvent(new ServerEvent(xmlMsg.toString(), sender.getChannel(), recipients));
  }

  public void sendResponse(String[] params, int fromRoom, User sender, LinkedList recipients)
  {
    StringBuffer response = new StringBuffer(ConfigData.STR_DELIMITER)
      .append("xt").append(ConfigData.STR_DELIMITER);

    response.append(params[0]).append(ConfigData.STR_DELIMITER);
    response.append(fromRoom).append(ConfigData.STR_DELIMITER);

    for (int i = 1; i < params.length; ++i)
      response.append(params[i]).append(ConfigData.STR_DELIMITER);

    SmartFoxServer.getInstance().getEW().handleEvent(new ServerEvent(response.toString(), sender.getChannel(), recipients));
  }

  public void sendResponse(org.json.JSONObject jso, int fromRoom, User sender, LinkedList recipients)
  {
    org.json.JSONObject jsoBody;
    try
    {
      jsoBody = new org.json.JSONObject();
      jsoBody.put("r", fromRoom);
      jsoBody.put("o", jso);

      org.json.JSONObject jsoMsg = new org.json.JSONObject();
      jsoMsg.put("t", "xt");
      jsoMsg.put("b", jsoBody);

      SmartFoxServer.getInstance().getEW().handleEvent(new ServerEvent(jsoMsg.toString(), sender.getChannel(), recipients));
    }
    catch (org.json.JSONException jsoEx)
    {
      SmartFoxServer.log.warning("Error creating JSON response: " + jsoEx);
    }
  }

  public void sendResponse(net.sf.json.JSONObject jso, int fromRoom, User sender, LinkedList recipients)
  {
    net.sf.json.JSONObject jsoBody;
    try
    {
      jsoBody = new net.sf.json.JSONObject();
      jsoBody.put("r", Integer.valueOf(fromRoom));
      jsoBody.put("o", jso);

      net.sf.json.JSONObject jsoMsg = new net.sf.json.JSONObject();
      jsoMsg.put("t", "xt");
      jsoMsg.put("b", jsoBody);

      SmartFoxServer.getInstance().getEW().handleEvent(new ServerEvent(jsoMsg.toString(), sender.getChannel(), recipients));
    }
    catch (net.sf.json.JSONException jsoEx)
    {
      SmartFoxServer.log.warning("Error creating JSON response: " + jsoEx);
    }
  }
}