package it.gotoandplay.smartfoxserver.extensions;

import it.gotoandplay.smartfoxserver.SmartFoxServer;
import it.gotoandplay.smartfoxserver.admin.AbstractAdminTask;
import it.gotoandplay.smartfoxserver.admin.AdminTaskExecutor;
import it.gotoandplay.smartfoxserver.admin.BanUserTask;
import it.gotoandplay.smartfoxserver.admin.HaltTask;
import it.gotoandplay.smartfoxserver.admin.KickUser;
import it.gotoandplay.smartfoxserver.admin.KickedUserCleanerTask;
import it.gotoandplay.smartfoxserver.admin.RestartTask;
import it.gotoandplay.smartfoxserver.config.ConfigData;
import it.gotoandplay.smartfoxserver.controllers.ExtensionHandler;
import it.gotoandplay.smartfoxserver.controllers.SystemHandler;
import it.gotoandplay.smartfoxserver.crypto.MD5;
import it.gotoandplay.smartfoxserver.data.BannedUser;
import it.gotoandplay.smartfoxserver.data.Moderator;
import it.gotoandplay.smartfoxserver.data.ModeratorManager;
import it.gotoandplay.smartfoxserver.data.Room;
import it.gotoandplay.smartfoxserver.data.RoomVariable;
import it.gotoandplay.smartfoxserver.data.User;
import it.gotoandplay.smartfoxserver.data.UserVariable;
import it.gotoandplay.smartfoxserver.data.Zone;
import it.gotoandplay.smartfoxserver.events.InternalEventObject;
import it.gotoandplay.smartfoxserver.exceptions.LoginException;
import it.gotoandplay.smartfoxserver.lib.ActionscriptObject;
import it.gotoandplay.smartfoxserver.lib.ConfigReader;
import it.gotoandplay.smartfoxserver.lib.Password;
import it.gotoandplay.smartfoxserver.lib.SmartFoxLib;
import it.gotoandplay.smartfoxserver.util.Attachment;
import it.gotoandplay.smartfoxserver.webserver.WebServer;
import java.io.BufferedReader;
import java.io.File;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.FilenameFilter;
import java.io.IOException;
import java.net.InetAddress;
import java.net.Socket;
import java.nio.channels.SelectionKey;
import java.nio.channels.SocketChannel;
import java.text.DateFormat;
import java.text.SimpleDateFormat;
import java.util.Collection;
import java.util.Date;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Timer;
import java.util.logging.Logger;

public class AdminExtension extends AbstractExtension
{
  private SystemHandler sh;
  private Timer timer;
  private AdminTaskExecutor taskExecutor;
  public static final int ONE_DAY = 86400000;
  public static final int ONE_HOUR = 3600000;
  public static final int ONE_MINUTE = 60000;
  public static final int ONE_SECOND = 1000;
  private static final String KEY_TOKEN = "###";
  public static final long EXECUTOR_INTERVAL = 2000L;
  public static final long KICK_DELAY = 6000L;
  public static final long KICK_CLEAN_DELAY = 60000L;
  private LinkedList adminList;
  private Password password;
  private int currentRoom;
  private Integer currentAdminId = null;

  public AdminExtension()
  {
    this.currentRoom = -1;
    this.sh = SmartFoxServer.getInstance().getSysHandler();

    this.taskExecutor = new AdminTaskExecutor();

    this.timer = new Timer();
    this.timer.schedule(this.taskExecutor, 2000L, 2000L);

    this.adminList = new LinkedList();

    this.password = Password.getInstance();

    addAdminTask(new KickedUserCleanerTask(System.currentTimeMillis() + 60000L, true));
  }

  public void init()
  {
  }

  public void destroy()
  {
    this.timer.cancel();
  }

  public void handleInternalEvent(InternalEventObject ieo)
  {
    LinkedList recipients;
    ActionscriptObject res;
    String evt = ieo.getEventName();
    String evtZone = ieo.getParam("zone");

    if ((evt.equals("loginRequest")) && (evtZone.equals("$dmn")))
    {
      String nick = ieo.getParam("nick");
      String pass = ieo.getParam("pass");
      SocketChannel chan = (SocketChannel)ieo.getObject("chan");
      String ipAdr = chan.socket().getInetAddress().toString().substring(1);
      boolean forceDisconnect = false;

      if (nick.startsWith("$_$"))
      {
        nick = nick.substring(3);
        forceDisconnect = true;
      }

      SelectionKey sk = chan.keyFor(SmartFoxServer.getInstance().getReadSelector());
      Attachment at = (Attachment)sk.attachment();
      String key = at.getSecretKey();

      String md5Hash = MD5.instance().getHash(key + SmartFoxServer.getInstance().getAdminPassword());
      try
      {
        if (SmartFoxServer.getInstance().getAdminAddress(ipAdr))
        {
          if ((nick.equals(SmartFoxServer.getInstance().getAdminLogin())) && (pass.equals(md5Hash)))
          {
            if (forceDisconnect)
              disconnectCurrentAdmin();

            User u = SmartFoxServer.getInstance().canLogin(nick, SmartFoxServer.getInstance().getAdminPassword(), chan, "$dmn");
            u.setAsAdmin();

            this.currentAdminId = new Integer(u.getUserId());

            res = new ActionscriptObject();

            res.put("_cmd", "logOK");

            res.put("name", u.getName());
            res.putNumber("id", u.getUserId());
            res.put("ver", SmartFoxServer.getInstance().getVersion());
            res.putNumber("type", SmartFoxServer.getInstance().getType());

            recipients = new LinkedList();
            recipients.add(chan);

            sendResponse(res, -1, null, recipients); return;
          }

          throw new LoginException("Invalid name or password, try again!");
        }

        throw new LoginException("Your IP addr. is not allowed to connect as admin.");
      }
      catch (LoginException le)
      {
        res = new ActionscriptObject();

        res.put("_cmd", "logKO");
        res.put("err", le.getLocalizedMessage());

        recipients = new LinkedList();
        recipients.add(chan);

        sendResponse(res, -1, null, recipients);
      }
    }
    else
    {
      int rid;
      Integer uid;
      if (evt.equals("userJoin"))
      {
        rid = Integer.parseInt(ieo.getParam("roomId"));

        if (rid == this.currentRoom)
        {
          res = new ActionscriptObject();

          uid = new Integer(ieo.getParam("uid"));

          res.put("_cmd", "userJoin");
          res.put("uid", ieo.getParam("uid"));
          res.put("uName", SmartFoxServer.getInstance().getUserById(uid).getName());
          res.put("rid", ieo.getParam("roomName"));
          res.put("zone", ieo.getParam("zone"));

          recipients = getAllAdmins();

          sendResponse(res, -1, null, recipients);
        }

      }
      else if (evt.equals("userExit"))
      {
        rid = Integer.parseInt(ieo.getParam("roomId"));

        if (rid == this.currentRoom)
        {
          res = new ActionscriptObject();

          uid = new Integer(ieo.getParam("uid"));

          res.put("_cmd", "userExit");
          res.put("uid", ieo.getParam("uid"));
          res.put("rid", ieo.getParam("roomName"));
          res.put("zone", ieo.getParam("zone"));

          recipients = getAllAdmins();

          sendResponse(res, -1, null, recipients);
        }

      }
      else if ((evt.equals("userLost")) || (evt.equals("logOut")))
      {
        int[] rids = (int[])ieo.getObject("roomIds");

        for (int r = 0; r < rids.length; ++r)
        {
          if (rids[r] == this.currentRoom)
          {
            Zone zn = SmartFoxServer.getInstance().getZone(ieo.getParam("zone"));
            Room rm = zn.getRoom(rids[r]);

            if (rm == null) return;

            res = new ActionscriptObject();

            Integer uid = new Integer(ieo.getParam("uid"));

            res.put("_cmd", "userLost");
            res.put("uid", ieo.getParam("uid"));
            res.put("rid", rm.getName());
            res.put("zone", ieo.getParam("zone"));

            recipients = getAllAdmins();

            sendResponse(res, -1, null, recipients);

            return;
          }

        }

      }
      else if (evt.equals("newRoom"))
      {
        res = new ActionscriptObject();

        res.put("_cmd", "newRoom");
        res.put("z", ieo.getParam("zone"));
        res.put("id", ieo.getParam("roomId"));
        res.put("rn", ieo.getParam("roomName"));

        recipients = getAllAdmins();

        sendResponse(res, -1, null, recipients);
      }
      else if (evt.equals("roomLost"))
      {
        res = new ActionscriptObject();

        res.put("_cmd", "roomLost");
        res.put("z", ieo.getParam("zone"));
        res.put("id", ieo.getParam("roomId"));

        recipients = getAllAdmins();

        sendResponse(res, -1, null, recipients);
      }
    }
  }

  public void handleRequest(String cmd, String[] params, User user, int param)
  {
  }

  public void handleRequest(String cmd, ActionscriptObject ao, User user, int param)
  {
    if (user.isAdmin())
    {
      if (cmd.equals("zoneInfo"))
      {
        sendZoneInfo(user);
      }
      else if (cmd.equals("roomInfo"))
      {
        sendRoomInfo(ao, user);
      }
      else if (cmd.equals("userInfo"))
      {
        sendUserInfo(ao, user);
      }
      else if (cmd.equals("zInfo"))
      {
        sendSpecificZoneInfo(ao, user);
      }
      else if (cmd.equals("kick"))
      {
        kickUser(ao, user);
      }
      else if (cmd.equals("dmnMsg"))
      {
        sendAdminMessage(ao, user);
      }
      else if (cmd.equals("ban"))
      {
        banishUser(ao, user);
      }
      else if (cmd.equals("status"))
      {
        sendServerStatus(user);
      }
      else if (cmd.equals("getCfg"))
      {
        sendConfigFile(user);
      }
      else if (cmd.equals("putCfg"))
      {
        storeConfigFile(ao, user);
      }
      else if (cmd.equals("restart"))
      {
        sendRestartSignal(ao, user);
      }
      else if (cmd.equals("halt"))
      {
        sendHaltSignal(ao, user);
      }
      else if (cmd.equals("getBan"))
      {
        sendBannedLists(user);
      }
      else if (cmd.equals("remBan"))
      {
        removeBannedItem(ao, user);
      }
      else if (cmd.equals("getLog"))
      {
        sendLogsTail(ao, user);
      }
      else if (cmd.equals("zStop"))
      {
        stopZone(ao, user);
      }
      else if (cmd.equals("zReload"))
      {
        reloadZones(ao, user);
      }
      else if (cmd.equals("zToggle"))
      {
        toggleZoneStatus(ao, user);
      }
      else if (cmd.equals("reloadExt"))
      {
        reloadExtension(ao, user);
      }
      else if (cmd.equals("toggleExt"))
      {
        toggleExtensionStatus(ao, user);
      }
      else if (cmd.equals("getXt"))
      {
        sendExtensionsList(user);
      }
      else if (cmd.equals("loadXt"))
      {
        loadExtensionFile(ao, user);
      }
      else if (cmd.equals("saveXt"))
      {
        saveExtensionFile(ao, user);
      }
      else if (cmd.equals("roomDst"))
      {
        destroyRoom(ao, user);
      }
      else if (cmd.equals("bbStatus"))
      {
        sendBBStatus(user);
      }
      else if (cmd.equals("toggleHttpServer"))
      {
        toggleHttpServerStatus(ao, user);
      }
    }
    else if (user.isModerator())
    {
      if (cmd.equals("kick"))
      {
        kickUser(ao, user);
      }
      else if (cmd.equals("ban"))
      {
        banishUser(ao, user);
      }

    }
    else
    {
      SmartFoxServer.log.warning("A non-Admin user has sent an Admin request. Req: " + cmd + ", User: " + user.getName() + ", IP: " + user.getIpAddress());
    }
  }

  private void sendServerStatus(User who)
  {
    ActionscriptObject response = new ActionscriptObject();
    Runtime rt = Runtime.getRuntime();

    response.put("_cmd", "status");

    response.put("upTime", getUptime());
    response.put("dataIn", String.valueOf(ConfigData.dataIN));
    response.put("dataOut", String.valueOf(ConfigData.dataOUT));

    int load = SmartFoxServer.getInstance().getGlobalUserCount() * ConfigData.MAX_CHANNEL_QUEUE;
    int outGoingMessagesQSize = SmartFoxServer.getInstance().getGlobalQueueSize();
    load = outGoingMessagesQSize * 100 / load;

    response.put("load", String.valueOf(load));

    response.put("vmMax", String.valueOf(rt.maxMemory()));
    response.put("vmUsed", String.valueOf(rt.maxMemory() - rt.freeMemory()));

    response.put("omq", String.valueOf(outGoingMessagesQSize));
    response.put("shq", String.valueOf(this.sh.getQueueSize()));
    response.put("ehq", String.valueOf(SmartFoxServer.getInstance().getExtensionHandler().getQueueSize()));

    response.put("s_zones", String.valueOf(SmartFoxServer.getInstance().getZoneNumber()));
    response.put("s_rooms", String.valueOf(SmartFoxServer.getInstance().getRoomNumber()));
    response.put("s_users", String.valueOf(SmartFoxServer.getInstance().getGlobalUserCount()));
    response.put("s_banned", String.valueOf(SmartFoxServer.getInstance().getBanList().size()));
    response.put("s_max", String.valueOf(ConfigData.maxSimultanousConnections));
    response.put("s_dropIn", String.valueOf(ConfigData.inComingDroppedMessages));
    response.put("s_dropOut", String.valueOf(ConfigData.outGoingDroppedMessages));
    response.put("s_sockets", String.valueOf(SmartFoxServer.getInstance().getChannels().size()));
    response.put("s_threads", String.valueOf(Thread.activeCount()));
    response.put("s_restarts", String.valueOf(ConfigData.restartCount));

    LinkedList ll = new LinkedList();
    ll.add(who.getChannel());

    sendResponse(response, -1, who, ll);
  }

  private void sendBannedLists(User who)
  {
    LinkedList bannedUser = SmartFoxServer.getInstance().getBanList();

    ActionscriptObject response = new ActionscriptObject();
    ActionscriptObject names = new ActionscriptObject();
    ActionscriptObject addresses = new ActionscriptObject();

    SimpleDateFormat sdf = new SimpleDateFormat("MM/dd/yyyy, HH:mm:ss");

    synchronized (bannedUser)
    {
      for (Iterator i = bannedUser.iterator(); i.hasNext(); )
      {
        BannedUser bu = (BannedUser)i.next();
        String ip = bu.getIp();
        String name = bu.getName();

        if ((ip != null) && 
          (!(ip.equals(""))))
          addresses.put(ip, sdf.format(new Date(bu.getBanDate())));

        if ((name == null) || 
          (name.equals(""))) break label161;
        label161: names.put(name, sdf.format(new Date(bu.getBanDate())));
      }
    }

    response.put("_cmd", "getBan");
    response.put("ips", addresses);
    response.put("names", names);

    LinkedList ll = new LinkedList();
    ll.add(who.getChannel());

    sendResponse(response, -1, who, ll);
  }

  private void sendConfigFile(User who)
  {
    ActionscriptObject response = new ActionscriptObject();
    response.put("_cmd", "getCfg");
    try
    {
      byte[] fileData = SmartFoxLib.readFileInClassPath("config.xml");

      response.put("cfg", SmartFoxLib.fixLineFeeds(fileData));

      LinkedList ll = new LinkedList();
      ll.add(who.getChannel());

      sendResponse(response, -1, who, ll);
    }
    catch (IOException ioe)
    {
      SmartFoxServer.log.warning("Could not read config.xml file!");
    }
  }

  private void storeConfigFile(ActionscriptObject ao, User who)
  {
    ActionscriptObject response = new ActionscriptObject();
    response.put("_cmd", "putCfg");

    boolean backup = ao.getBool("backup");
    String data = ao.getString("cfg");
    boolean success = false;
    try
    {
      if (data == null)
        break label263;
      if (backup)
      {
        boolean dirExist = false;
        boolean renameOk = false;

        File backupDir = new File("./confBackup");

        if (!(backupDir.isDirectory()))
        {
          dirExist = backupDir.mkdir();
        }
        else
          dirExist = true;

        if (dirExist)
        {
          File source = new File("config.xml");

          Date d = new Date(System.currentTimeMillis());

          SimpleDateFormat sdf = new SimpleDateFormat("yyyyMMddHHmmss");
          String fname = "confBackup/backup_" + sdf.format(d) + ".xml";

          File target = new File(fname);

          renameOk = source.renameTo(target);

          if (renameOk)
          {
            writeConfigFile(data);
            success = true; break label263:
          }

          SmartFoxServer.log.warning("Could not rename old config file!"); break label263:
        }

        SmartFoxServer.log.warning("Could not find/create the config backup folder!"); break label263:
      }

      writeConfigFile(data);
      success = true;
    }
    catch (IOException e)
    {
      SmartFoxServer.log.warning("I/O Exception while saving config file: " + e);
      e.printStackTrace();
    }

    if (success)
    {
      label263: response.put("res", "Configuration saved succesfully");
    }
    else
    {
      response.put("res", "An error occurred while saving the config file.\nCheck your server logs!");
    }

    LinkedList ll = new LinkedList();
    ll.add(who.getChannel());

    sendResponse(response, -1, who, ll);
  }

  private void writeConfigFile(String data)
    throws IOException
  {
    FileWriter fw = new FileWriter("config.xml");
    fw.write(data.toCharArray());
    fw.close();
  }

  private void sendLogsTail(ActionscriptObject ao, User who)
  {
    int lines = (int)ao.getNumber("n");
    try
    {
      if (lines > 500)
        lines = 500;
      else if (lines < 0) {
        lines = 30;
      }

      ActionscriptObject response = new ActionscriptObject();

      LinkedList logLines = new LinkedList();

      File currDir = new File("logs/");
      FilenameFilter filter = new TxtFileFilter(this, "txt");

      File[] entries = currDir.listFiles(filter);

      long fileDate = 0L;
      String mostRecent = null;

      for (int f = 0; f < entries.length; ++f)
      {
        File logFile = entries[f];

        if (logFile.lastModified() > fileDate)
        {
          fileDate = logFile.lastModified();
          mostRecent = logFile.getName();
        }

      }

      if (mostRecent != null)
      {
        String s;
        FileReader fr = new FileReader(new File("logs/" + mostRecent));
        BufferedReader br = new BufferedReader(fr);
        while (true)
        {
          s = br.readLine();

          if (s == null) {
            break;
          }

          if (logLines.size() == lines)
          {
            logLines.removeFirst();
          }

          logLines.addLast(s);
        }

        ActionscriptObject txtData = new ActionscriptObject();

        int c = 0;

        for (Iterator i = logLines.iterator(); i.hasNext(); ) {
          txtData.put(c++, i.next());
        }

        response.put("logs", txtData);
      }

      response.put("_cmd", "getLog");

      LinkedList ll = new LinkedList();
      ll.add(who.getChannel());

      sendResponse(response, -1, who, ll);
    }
    catch (NumberFormatException nfe)
    {
      SmartFoxServer.log.warning("Bad number of log lines requested");
    }
    catch (Exception e)
    {
      SmartFoxServer.log.warning(e.toString());
    }
  }

  private void sendRestartSignal(ActionscriptObject ao, User who)
  {
    SmartFoxServer.log.info("Server Restart was requested by administrator.");

    String msg = ao.getString("msg");

    this.timer.cancel();

    if (msg == null) {
      msg = "Server is going to be restarted in 10 seconds...";
    }

    this.sh.sendAdminMessage(who, msg, "dmn");

    new RestartTask(who);
  }

  private void sendHaltSignal(ActionscriptObject ao, User who)
  {
    String msg = ao.getString("msg");

    this.timer.cancel();

    if (msg == null)
      msg = "Server is going to be halted in 10 seconds...";

    SmartFoxServer.log.info("Server Halt was requested by administrator.");

    this.sh.sendAdminMessage(who, msg, "dmn");

    new HaltTask(who);
  }

  private void removeBannedItem(ActionscriptObject ao, User who)
  {
    ActionscriptObject response = new ActionscriptObject();
    response.put("_cmd", "remBan");

    String itemId = ao.getString("id");
    String type = ao.getString("t");

    if ((itemId != null) && (type != null))
    {
      BannedUser bu = null;

      if (type.equals("a"))
        bu = new BannedUser("", itemId);
      else if (type.equals("n")) {
        bu = new BannedUser(itemId, "");
      }

      if (bu == null)
        return;

      SmartFoxServer.getInstance().removeBannedUser(bu);

      response.put("id", itemId);
      response.put("t", type);

      LinkedList ll = new LinkedList();
      ll.add(who.getChannel());

      sendResponse(response, -1, who, ll);
    }
  }

  private String getUptime()
  {
    StringBuffer result = new StringBuffer();

    long now = System.currentTimeMillis();
    long start = SmartFoxServer.getInstance().getServerStartTime();
    long temp = 0L;

    long elapsed = now - start;
    int days = (int)Math.floor(elapsed / 86400000L);

    temp = 86400000L * days;

    elapsed -= temp;
    int hours = (int)Math.floor(elapsed / 3600000L);

    temp = 3600000 * hours;
    elapsed -= temp;
    int minutes = (int)Math.floor(elapsed / 60000L);

    String s_days = String.valueOf(days);

    for (int i = 0; i < 4 - s_days.length(); ++i) {
      result.append("0");
    }

    result.append(s_days);
    result.append(":");

    if (hours < 10)
      result.append("0");

    result.append(hours);
    result.append(":");

    if (minutes < 10)
      result.append("0");

    result.append(minutes);

    return result.toString();
  }

  private String getDataSize(long size, int decimalPlaces)
  {
    int thousands = 0;
    double dsize = size;
    do
    {
      dsize /= 1000.0D;
      ++thousands;
    }
    while (dsize > 1000.0D);

    String magnitude = null;

    switch (thousands)
    {
    case 1:
      magnitude = "Kb";
      break;
    case 2:
      magnitude = "Mb";
      break;
    case 3:
      magnitude = "Gb";
      break;
    case 4:
      magnitude = "Tb";
      break;
    case 5:
      magnitude = "Pb";
    }

    String finalSize = String.valueOf(dsize);

    if ((int)dsize != dsize)
      finalSize = finalSize.substring(0, finalSize.indexOf(".") + 1 + decimalPlaces);

    return finalSize + " " + magnitude;
  }

  private String getDataSize(int size, int decimalPlaces)
  {
    return getDataSize(size, decimalPlaces);
  }

  private void _sendZoneInfo(User who)
  {
    LinkedList zoneList = SmartFoxServer.getInstance().getAllZones();

    ActionscriptObject response = new ActionscriptObject();
    ActionscriptObject zoneObj = new ActionscriptObject();

    response.put("_cmd", "zoneList");
    response.put("zones", zoneObj);

    Iterator i = zoneList.iterator(); while (i.hasNext())
    {
      Zone z = (Zone)i.next();

      ActionscriptObject rList = new ActionscriptObject();
      zoneObj.put(z.getName(), rList);

      LinkedList roomList = z.getRoomList();

      int roomCount = 0;

      for (Iterator ii = roomList.iterator(); ii.hasNext(); )
      {
        Room r = (Room)ii.next();

        rList.put(roomCount++, r.getName());
      }
    }

    LinkedList ll = new LinkedList();
    ll.add(who.getChannel());

    sendResponse(response, -1, who, ll);
  }

  private void sendUserInfo(ActionscriptObject ao, User who)
  {
    String id = ao.getString("id");

    if ((id != null) && (!(id.equals(""))))
    {
      User u = SmartFoxServer.getInstance().getUserById(new Integer(id));

      if (u != null)
      {
        ActionscriptObject response = new ActionscriptObject();

        response.put("_cmd", "userInfo");

        response.put("id", String.valueOf(u.getUserId()));
        response.put("name", u.getName());
        response.put("ip", u.getChannel().socket().getInetAddress().getHostAddress());

        Date d = new Date(u.getLoginTime());
        DateFormat df = DateFormat.getInstance();

        response.put("time", df.format(d));
        response.put("mod", String.valueOf(u.isModerator()));
        response.put("admin", String.valueOf(u.isAdmin()));

        response.put("pid", String.valueOf(u.playerIndex));

        ActionscriptObject vars = new ActionscriptObject();
        response.put("vars", vars);

        LinkedList varNameList = u.getVariableNames();

        for (Iterator i = varNameList.iterator(); i.hasNext(); )
        {
          String varName = (String)i.next();
          UserVariable uv = u.getVariable(varName);

          ActionscriptObject var = new ActionscriptObject();

          var.put("name", varName);
          var.put("value", uv.getValue());
          var.put("type", String.valueOf(uv.getType()));

          vars.put(varName, var);
        }

        LinkedList ll = new LinkedList();
        ll.add(who.getChannel());

        sendResponse(response, -1, who, ll);
      }
    }
  }

  private void sendRoomInfo(ActionscriptObject ao, User who)
  {
    int roomRequested = Integer.parseInt((String)ao.get("r"));
    String zoneRequested = ao.getString("z");

    if ((roomRequested > -1) && (zoneRequested != null) && (!(zoneRequested.equals(""))))
    {
      Zone z = SmartFoxServer.getInstance().getZone(zoneRequested);

      this.currentRoom = roomRequested;

      if (z != null)
      {
        Room r = z.getRoom(roomRequested);

        if (r != null)
        {
          User u;
          ActionscriptObject response = new ActionscriptObject();

          response.put("_cmd", "roomInfo");
          response.put("id", String.valueOf(r.getId()));
          response.put("name", r.getName());
          response.put("isGame", String.valueOf(r.isGame()));
          response.put("isTemp", String.valueOf(r.isTemp()));
          response.put("isPriv", String.valueOf(r.isPrivate()));
          response.put("pwd", r.getPassword());
          response.put("users", String.valueOf(r.howManyUsers()));
          response.put("maxusers", String.valueOf(r.getMaxUsers()));
          response.put("ucount", String.valueOf(r.getUcountUpdates()));
          response.put("isLimbo", String.valueOf(r.isLimbo()));
          response.put("isAj", String.valueOf(r.getId() == z.getAutoJoinRoom()));

          int spec = 0;
          int maxspec = 0;

          if (r.isGame())
          {
            spec = r.getSpectatorCount();
            maxspec = r.getMaxSpectator();
          }

          response.put("spec", String.valueOf(spec));
          response.put("maxspec", String.valueOf(maxspec));

          SocketChannel sc = r.getCreator();
          if (sc != null)
          {
            u = SmartFoxServer.getInstance().getUserByChannel(sc);

            if (u != null)
              response.put("owner", u.getName());

          }

          ActionscriptObject vars = new ActionscriptObject();
          response.put("vars", vars);

          LinkedList varNameList = r.getVariableNames();

          for (Iterator i = varNameList.iterator(); i.hasNext(); )
          {
            String ownerName;
            String varName = (String)i.next();
            RoomVariable rv = r.getVariable(varName);

            ActionscriptObject var = new ActionscriptObject();

            var.put("name", varName);
            var.put("value", rv.getValue());

            u = rv.getOwner();
            if (u == null)
              ownerName = "[ server ]";
            else
              ownerName = u.getName();

            var.put("owner", ownerName);
            var.put("priv", (rv.isPrivate()) ? "Y" : "N");
            var.put("pers", (rv.isPersistent()) ? "Y" : "N");

            var.put("type", String.valueOf(rv.getType()));

            vars.put(varName, var);
          }

          ActionscriptObject uList = new ActionscriptObject();
          response.put("uList", uList);

          Object[] userIds = r.getUserList();

          for (int iii = 0; iii < userIds.length; ++iii)
          {
            ActionscriptObject usr = new ActionscriptObject();
            u = SmartFoxServer.getInstance().getUserById((Integer)userIds[iii]);

            usr.put("id", String.valueOf(u.getUserId()));
            usr.put("n", u.getName());

            uList.put(iii, usr);
          }

          ActionscriptObject extList = new ActionscriptObject();
          r.getExtManager().getExtensionList(extList);

          if (extList.size() > 0)
            response.put("ext", extList);

          LinkedList ll = new LinkedList();
          ll.add(who.getChannel());

          sendResponse(response, -1, who, ll);
        }
      }
    }
  }

  private void sendSpecificZoneInfo(ActionscriptObject ao, User who)
  {
    String zoneRequested = ao.getString("z");

    if ((zoneRequested != null) && (!("".equals(zoneRequested))))
    {
      Zone z = SmartFoxServer.getInstance().getZone(zoneRequested);

      if (z != null)
      {
        this.currentRoom = -1;

        ActionscriptObject response = new ActionscriptObject();
        response.put("_cmd", "zInfo");

        response.put("zName", z.getName());
        response.put("status", String.valueOf(z.getStatus()));
        response.put("count", String.valueOf(z.getCountUpdate()));
        response.put("buddy", String.valueOf(z.hasBuddyList()));
        response.put("empty", String.valueOf(z.getEmptyNames()));
        response.put("login", String.valueOf(z.hasCustomLogin()));
        response.put("totU", String.valueOf(z.getUserCount()));
        response.put("totR", String.valueOf(z.getRoomCount()));
        response.put("maxU", String.valueOf(z.getMaxUsers()));
        response.put("maxR", String.valueOf(z.getMaxRooms()));

        ActionscriptObject modList = new ActionscriptObject();

        response.put("mods", modList);

        if (z.modManager.getIsActive())
        {
          Collection entries = z.modManager.getModList();

          int cnt = 0;

          for (Iterator i = entries.iterator(); i.hasNext(); )
          {
            ActionscriptObject mod = new ActionscriptObject();

            Moderator md = (Moderator)i.next();
            mod.put("name", md.getName());
            mod.put("pass", md.getPwd());

            modList.put(cnt++, mod);
          }

        }

        ActionscriptObject extList = new ActionscriptObject();
        z.getExtManager().getExtensionList(extList);

        if (extList.size() > 0) {
          response.put("ext", extList);
        }

        LinkedList ll = new LinkedList();
        ll.add(who.getChannel());

        sendResponse(response, -1, who, ll);
      }
    }
  }

  private void sendZoneInfo(User who)
  {
    StringBuffer sb = new StringBuffer();

    LinkedList zoneList = SmartFoxServer.getInstance().getAllZones();

    ActionscriptObject response = new ActionscriptObject();
    ActionscriptObject zoneObj = new ActionscriptObject();

    response.put("_cmd", "zoneList");

    for (Iterator i = zoneList.iterator(); i.hasNext(); )
    {
      Zone z = (Zone)i.next();

      sb.append("<node label='").append(z.getName()).append("' type='z'>");

      LinkedList roomList = z.getRoomList();

      int roomCount = 0;

      for (Iterator ii = roomList.iterator(); ii.hasNext(); )
      {
        Room r = (Room)ii.next();
        sb.append("<node label='").append(r.getName()).append("' type='r' id='").append(r.getId()).append("' z='").append(z.getName()).append("' />");
      }

      sb.append("</node>");
    }

    response.put("zoneXml", sb.toString());

    LinkedList ll = new LinkedList();
    ll.add(who.getChannel());

    sendResponse(response, -1, who, ll);
  }

  private void kickUser(ActionscriptObject ao, User who)
  {
    User toKick = SmartFoxServer.getInstance().getUserById(new Integer(ao.getString("id")));
    String msg = ao.getString("msg");

    if ((msg != null) && (toKick != null) && 
      (!(toKick.isBeingKicked)) && (!(toKick.isAdmin())))
    {
      SmartFoxServer.log.info("User [ " + toKick.getName() + " ] is being kicked in " + 6L + " seconds");

      toKick.isBeingKicked = true;

      String kickMsg = (msg.equals("")) ? "You're being kicked" : msg;

      this.sh.sendAdminMessage(who, 0, kickMsg, toKick, "dmn");

      this.taskExecutor.addTask(new KickUser(toKick.getChannel(), System.currentTimeMillis() + 6000L, false));
    }
  }

  private void banishUser(ActionscriptObject ao, User who)
  {
    User toBan = SmartFoxServer.getInstance().getUserById(new Integer(ao.getString("id")));
    String msg = ao.getString("msg");

    int mode = new Integer(ao.getString("mode")).intValue();

    String uName = null;
    String uIp = null;

    if ((msg != null) && (toBan != null) && 
      (!(toBan.isBeingKicked)) && (!(toBan.isAdmin())))
    {
      SmartFoxServer.log.info("User [ " + toBan.getName() + " ] is being banned in " + 6L + " seconds");

      toBan.isBeingKicked = true;

      String kickMsg = (msg.equals("")) ? "You're being banned" : msg;

      this.sh.sendAdminMessage(who, 0, kickMsg, toBan, "dmn");

      if (mode == 0) {
        uName = toBan.getName();
      }
      else if (mode == 1) {
        uIp = toBan.getChannel().socket().getInetAddress().toString().substring(1);
      }
      else if (mode == 2)
      {
        uName = toBan.getName();
        uIp = toBan.getChannel().socket().getInetAddress().toString().substring(1);
      }

      this.taskExecutor.addTask(new BanUserTask(toBan.getChannel(), System.currentTimeMillis() + 6000L, false, uName, uIp));
    }
  }

  private void sendAdminMessage(ActionscriptObject ao, User who)
  {
    String msg = ao.getString("msg");
    String type = ao.getString("t");

    if (type.equals("z"))
    {
      String zName = ao.getString("z");
      Zone z = SmartFoxServer.getInstance().getZone(zName);

      if (z != null) {
        this.sh.sendAdminMessage(who, 0, msg, z, "dmn");
      }

    }
    else if (type.equals("r"))
    {
      String z = ao.getString("z");
      String r = ao.getString("r");

      if ((((r != null) ? 1 : 0) & ((z != null) ? 1 : 0)) != 0)
      {
        Zone zn = SmartFoxServer.getInstance().getZone(z);
        Room rm = zn.getRoom(Integer.parseInt(r));

        if (rm != null)
          this.sh.sendAdminMessage(who, 0, msg, rm, "dmn");

      }

    }
    else if (type.equals("u"))
    {
      Integer i = new Integer(ao.getString("id"));
      User recipient = SmartFoxServer.getInstance().getUserById(i);
      this.sh.sendAdminMessage(who, 0, msg, recipient, "dmn");
    }
    else
    {
      type.equals("a");
    }
  }

  private void stopZone(ActionscriptObject ao, User who)
  {
    String zoneName = ao.getString("z");

    if ((!("".equals(zoneName))) && (who.isAdmin()))
    {
      Zone z = SmartFoxServer.getInstance().getZone(zoneName);

      if (z != null)
      {
        SmartFoxServer.log.warning("Deactivating zone: " + zoneName);
        z.deactivate();
      }
    }
  }

  private void reloadZones(ActionscriptObject ao, User who)
  {
    try
    {
      ConfigReader.getInstance().reload_Zones();
      this.currentRoom = -1;
      sendZoneInfo(who);
    }
    catch (Exception e)
    {
      SmartFoxServer.log.warning("Could not reload zones: " + e.getMessage());
    }
  }

  private void reloadExtension(ActionscriptObject ao, User who)
  {
    String xtName = ao.getString("x");
    String zn = ao.getString("z");
    int rm = (int)ao.getNumber("r");

    ExtensionManager em = null;

    Zone z = SmartFoxServer.getInstance().getZone(zn);

    if (z != null)
    {
      if (rm == -1) {
        em = z.getExtManager();
      }
      else {
        Room r = z.getRoom(rm);
        em = r.getExtManager();
      }

    }

    if (em != null)
    {
      em.reloadExtension(xtName);
      ActionscriptObject obj = new ActionscriptObject();
      obj.put("z", zn);
      obj.put("r", String.valueOf(rm));

      if (rm != -1)
        sendRoomInfo(obj, who);
      else
        sendSpecificZoneInfo(obj, who);
    }
    else {
      SmartFoxServer.log.warning("Extension [ " + xtName + " ]" + " could not be reloaded.");
    }
  }

  public void toggleExtensionStatus(ActionscriptObject ao, User who) {
    String xtName = ao.getString("x");
    String zn = ao.getString("z");
    int rm = (int)ao.getNumber("r");

    Zone z = SmartFoxServer.getInstance().getZone(zn);

    ExtensionManager em = null;

    if (z != null)
    {
      if (rm == -1) {
        em = z.getExtManager();
      }
      else {
        Room r = z.getRoom(rm);
        em = r.getExtManager();
      }

    }

    if (em != null)
    {
      em.toggleStatus(xtName);

      ActionscriptObject obj = new ActionscriptObject();
      obj.put("z", zn);
      obj.put("r", String.valueOf(rm));

      if (rm != -1)
        sendRoomInfo(obj, who);
      else
        sendSpecificZoneInfo(obj, who);
    }
    else {
      SmartFoxServer.log.warning("Invalid Extension-Toggle-Status request: extension doesn't exist!");
    }
  }

  private void toggleZoneStatus(ActionscriptObject ao, User who)
  {
    String zoneName = ao.getString("z");

    if ((!("".equals(zoneName))) && (!("$dmn".equals(zoneName))))
    {
      Zone z = SmartFoxServer.getInstance().getZone(zoneName);

      if (z != null)
      {
        if (z.isActive())
        {
          z.deactivate();
          SmartFoxServer.log.info("Zone [ " + zoneName + " ] deactivated");
        }
        else
        {
          z.activate();
          SmartFoxServer.log.info("Zone [ " + zoneName + " ] activated");
        }

        sendSpecificZoneInfo(ao, who);
      }
    }
  }

  private void sendExtensionsList(User who)
  {
    File xtFile;
    File currDir = new File("./sfsExtensions/");
    FilenameFilter asFilter = new TxtFileFilter(this, "as");
    FilenameFilter pyFilter = new TxtFileFilter(this, "py");

    File[] asEntries = currDir.listFiles(asFilter);
    File[] pyEntries = currDir.listFiles(pyFilter);

    ActionscriptObject response = new ActionscriptObject();
    ActionscriptObject list = new ActionscriptObject();

    response.put("_cmd", "getXt");
    response.put("xt", list);

    int arrId = 0;
    for (int f = 0; f < asEntries.length; ++f)
    {
      xtFile = asEntries[f];
      list.put(arrId++, xtFile.getName());
    }

    for (f = 0; f < pyEntries.length; ++f)
    {
      xtFile = pyEntries[f];
      list.put(arrId++, xtFile.getName());
    }

    LinkedList ll = new LinkedList();
    ll.add(who.getChannel());

    sendResponse(response, -1, who, ll);
  }

  private void loadExtensionFile(ActionscriptObject ao, User who)
  {
    String file = ao.getString("f");

    if ((file != null) && (!(file.equals(""))))
    {
      ActionscriptObject response = new ActionscriptObject();
      response.put("_cmd", "loadXt");
      try
      {
        byte[] fileData = SmartFoxLib.readFileInClassPath("./sfsExtensions/" + file);

        if (fileData.length == 0)
          fileData = "Empty file!".getBytes();

        response.put("xt", SmartFoxLib.fixLineFeeds(fileData));
        response.put("fname", file);

        LinkedList ll = new LinkedList();
        ll.add(who.getChannel());

        sendResponse(response, -1, who, ll);
      }
      catch (IOException ioe)
      {
        SmartFoxServer.log.warning("Could not read the requested extension file: " + file + ". Reason: " + ioe.getMessage());
      }
    }
  }

  private void saveExtensionFile(ActionscriptObject ao, User who)
  {
    ActionscriptObject response = new ActionscriptObject();
    response.put("_cmd", "saveXt");

    String fileName = ao.getString("fn");
    String data = ao.getString("txt");

    boolean success = false;
    try
    {
      if ((data == null) || (fileName == null) || (fileName.equals("")))
        break label168;

      FileWriter fw = new FileWriter("./sfsExtensions/" + fileName);
      fw.write(data.toCharArray());
      fw.close();

      SmartFoxServer.log.info("Extension file [" + fileName + "] was saved succesfully.");
      success = true;
    }
    catch (IOException e)
    {
      SmartFoxServer.log.warning("I/O Exception while remotely saving extension: " + fileName);
      e.printStackTrace();
    }

    if (success)
    {
      label168: response.put("res", "Extension saved succesfully");
    }
    else
    {
      response.put("res", "An error occurred while saving the extension.\nCheck your server logs!");
    }

    LinkedList ll = new LinkedList();
    ll.add(who.getChannel());

    sendResponse(response, -1, who, ll);
  }

  private void destroyRoom(ActionscriptObject ao, User u)
  {
    int roomRequested = Integer.parseInt((String)ao.get("r"));
    String zoneRequested = ao.getString("z");

    if ((roomRequested > -1) && (zoneRequested != null) && (!(zoneRequested.equals(""))))
    {
      Zone z = SmartFoxServer.getInstance().getZone(zoneRequested);
      Room r = z.getRoom(roomRequested);

      if ((r != null) && (r.howManyUsers() == 0))
      {
        z.removeRoom(r, false);
        SmartFoxServer.log.info("Administrator destroys room: " + r.getName());
      }
    }
  }

  private void sendBBStatus(User who)
  {
    DateFormat formatter = DateFormat.getInstance();

    ActionscriptObject response = new ActionscriptObject();
    response.put("_cmd", "bbStatus");
    response.putBool("ws", WebServer.getInstance().getStatus() == 1);

    List bbUsers = SmartFoxServer.getInstance().getBlueBoxedUsers();
    ActionscriptObject uList = new ActionscriptObject();
    response.put("uList", uList);

    int c = 0;

    for (Iterator it = bbUsers.iterator(); it.hasNext(); )
    {
      User user = (User)it.next();

      ActionscriptObject userObj = new ActionscriptObject();
      userObj.putNumber("id", user.getUserId());
      userObj.put("name", user.getName());
      userObj.put("ssid", user.getBlueBoxSessionId());
      userObj.put("zone", user.getZone());
      userObj.put("t1", formatter.format(new Date(user.getLoginTime())));
      userObj.put("t2", formatter.format(new Date(user.getLastMessageTime())));

      uList.put(c++, userObj);
    }

    LinkedList ll = new LinkedList();
    ll.add(who.getChannel());

    sendResponse(response, -1, who, ll);
  }

  private void toggleHttpServerStatus(ActionscriptObject obj, User who)
  {
    int status = (int)obj.getNumber("status");
    try
    {
      if (status == 1) {
        WebServer.newInstance(); break label36:
      }
      WebServer.getInstance().stop();
    }
    catch (Exception e)
    {
      e.printStackTrace();
    }

    label36: sendBBStatus(who);
  }

  private LinkedList getAllAdmins()
  {
    LinkedList ll = new LinkedList();

    if (this.currentAdminId != null)
    {
      User u = SmartFoxServer.getInstance().getUserById(this.currentAdminId);

      if (u != null)
        ll.add(u.getChannel());
    }

    return ll;
  }

  private void disconnectCurrentAdmin()
  {
    if (this.currentAdminId != null)
      SmartFoxServer.getInstance().lostConnection(this.currentAdminId);
  }

  public void addAdminTask(AbstractAdminTask task)
  {
    this.taskExecutor.addTask(task);
  }

  public void logRemoteTrace(String zn, String rm, String msg)
  {
    if (ConfigData.EXTENSION_REMOTE_DEBUGGING)
    {
      ActionscriptObject ao = new ActionscriptObject();
      ao.put("_cmd", "trace");
      ao.put("z", zn);
      ao.put("r", rm);
      ao.put("m", msg);

      sendResponse(ao, -1, null, getAllAdmins());
    }
  }

  private class TxtFileFilter
    implements FilenameFilter
  {
    String xt;

    public TxtFileFilter(, String paramString)
    {
      this.xt = "." + ext;
    }

    public boolean accept(, String name)
    {
      return name.endsWith(this.xt);
    }
  }
}