package it.gotoandplay.smartfoxserver.extensions;

import it.gotoandplay.smartfoxserver.EventWriter;
import it.gotoandplay.smartfoxserver.SmartFoxServer;
import it.gotoandplay.smartfoxserver.controllers.ExtensionHandler;
import it.gotoandplay.smartfoxserver.controllers.SystemHandler;
import it.gotoandplay.smartfoxserver.data.Room;
import it.gotoandplay.smartfoxserver.data.RoomVariable;
import it.gotoandplay.smartfoxserver.data.User;
import it.gotoandplay.smartfoxserver.data.UserVariable;
import it.gotoandplay.smartfoxserver.data.Zone;
import it.gotoandplay.smartfoxserver.data.buddylist.BuddyListManager;
import it.gotoandplay.smartfoxserver.exceptions.CreateRoomException;
import it.gotoandplay.smartfoxserver.exceptions.ExtensionHelperException;
import it.gotoandplay.smartfoxserver.exceptions.LoginException;
import it.gotoandplay.smartfoxserver.lib.ServerEvent;
import it.gotoandplay.smartfoxserver.lib.SmartFoxLib;
import it.gotoandplay.smartfoxserver.util.Attachment;
import java.nio.channels.SelectionKey;
import java.nio.channels.SocketChannel;
import java.util.HashMap;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;
import java.util.logging.Logger;

public class ExtensionHelper
{
  private static ExtensionHelper _instance;
  private static String ERROR_CREATE_ROOM = "Create room failed: ";
  private static String ERROR_JOIN_ROOM = "Join room failed: ";
  public static final int MOD_MESSAGE_TO_USER = 0;
  public static final int MOD_MESSAGE_TO_ROOM = 1;
  public static final int MOD_MESSAGE_TO_ZONE = 2;
  private SmartFoxServer sfs;
  private SystemHandler sysHandler;
  private AdminExtension adminExt;

  private ExtensionHelper()
  {
    this.sfs = SmartFoxServer.getInstance();
    this.sysHandler = this.sfs.getSysHandler();
    this.adminExt = this.sfs.getExtensionHandler().getAdminExtension();
  }

  public static ExtensionHelper instance()
  {
    if (_instance == null)
      _instance = new ExtensionHelper();

    return _instance;
  }

  public static void destroy()
  {
    _instance = null;
  }

  public User canLogin(String nick, String pass, SocketChannel ch, String zone)
    throws LoginException
  {
    return this.sfs.canLogin(nick, pass, ch, zone);
  }

  public void dispatchPublicMessage(String msg, Room r, User u)
  {
    this.sfs.getSysHandler().dispatchPublicMessage(msg, r, u);
  }

  public void dispatchPrivateMessage(String msg, Room r, User sender, User recipient)
  {
    this.sfs.getSysHandler().dispatchPrivateMessage(r, sender, recipient, msg);
  }

  public String getGuestName()
  {
    return this.sfs.getGuestName();
  }

  public String getSecretKey(SocketChannel channel)
  {
    SelectionKey selectionKey = channel.keyFor(SmartFoxServer.getInstance().getReadSelector());

    return ((Attachment)selectionKey.attachment()).getSecretKey();
  }

  public User getUserByChannel(SocketChannel sc)
  {
    return this.sfs.getUserByChannel(sc);
  }

  public User getUserById(int id)
  {
    return this.sfs.getUserById(new Integer(id));
  }

  public void logoutUser(int id, boolean fireClientEvt, boolean fireInternalEvt)
  {
    logoutUser(getUserById(id), fireClientEvt, fireInternalEvt);
  }

  public void logoutUser(User user, boolean fireClientEvt, boolean fireInternalEvt)
  {
    this.sysHandler.handleLogout(user.getChannel(), fireClientEvt, fireInternalEvt);
  }

  public void sendGenericMessage(String message, SocketChannel sender, LinkedList channelList)
  {
    SmartFoxServer.getInstance().getEW().handleEvent(new ServerEvent(message, sender, channelList));
  }

  public Room createRoom(Zone z, Map params, User creator, boolean sendUpdate, boolean broadcastEvent)
    throws ExtensionHelperException
  {
    return createRoom(z, params, creator, null, null, false, sendUpdate, broadcastEvent);
  }

  public Room createRoom(Zone z, Map params, User creator, HashMap roomVars, User varsOwner, boolean setOwnership, boolean sendUpdate, boolean broadcastEvent)
    throws ExtensionHelperException
  {
    label399: label422: boolean isTemp;
    boolean isGame;
    boolean isLimbo;
    int maxUsers;
    int maxSpect;
    if (z == null)
      throw new ExtensionHelperException(ERROR_CREATE_ROOM + "invalid zone");

    if (params == null) {
      throw new ExtensionHelperException(ERROR_CREATE_ROOM + "params is null");
    }

    Room room = null;

    String name = (String)params.get("name");
    String pwd = (String)params.get("pwd");
    String maxU = (String)params.get("maxU");
    String maxS = (String)params.get("maxS");
    String game = (String)params.get("isGame");
    String limbo = (String)params.get("isLimbo");
    String uCount = (String)params.get("uCount");
    String xtName = (String)params.get("xtName");
    String xtClass = (String)params.get("xtClass");
    String lazyInit = (String)params.get("lazyInit");

    if (uCount == null) {
      uCount = "true";
    }

    name = SmartFoxLib.cleanUpString(name, z.roomNameAvoidChars, z.getMaxRoomNameLen());

    if (name == null)
      throw new ExtensionHelperException(ERROR_CREATE_ROOM + "missing room name");

    if (name.equals(""))
      throw new ExtensionHelperException(ERROR_CREATE_ROOM + "empty room name");

    if (z.roomNameAlreadyExist(name)) {
      throw new ExtensionHelperException(ERROR_CREATE_ROOM + "room already exist");
    }

    if (pwd == null) {
      pwd = "";
    }

    try
    {
      maxUsers = Integer.parseInt(maxU);
      if (maxUsers >= 1) break label399;
      maxUsers = 1;
    }
    catch (NumberFormatException nfe)
    {
      throw new ExtensionHelperException(ERROR_CREATE_ROOM + "invalid max number of users");
    }

    try
    {
      maxSpect = Integer.parseInt(maxS);
      if (maxSpect >= 0) break label422;
      maxSpect = 0;
    }
    catch (NumberFormatException nfe)
    {
      maxSpect = 0;
    }

    if ((game != null) && (((game.equalsIgnoreCase("true")) || (game.equalsIgnoreCase("1")))))
      isGame = true;
    else {
      isGame = false;
    }

    if ((limbo != null) && (((limbo.equalsIgnoreCase("true")) || (limbo.equalsIgnoreCase("1")))))
      isLimbo = true;
    else {
      isLimbo = false;
    }

    if (creator == null)
      isTemp = false;
    else {
      isTemp = true;
    }

    if (lazyInit == null) {
      lazyInit = "false";
    }

    try
    {
      room = this.sfs.createRoom(
        z.getName(), 
        name, 
        maxUsers, 
        maxSpect, 
        isTemp, 
        isGame, 
        (pwd.length() > 0) ? 1 : false, 
        pwd);

      if (creator != null)
      {
        room.setCreator(creator.getChannel());
        creator.addCreatedRoom(room);
      }

      if ((uCount != null) && (((uCount.equalsIgnoreCase("false")) || (uCount.equalsIgnoreCase("0")))))
        room.setUcountUpdates(false);
      else {
        room.setUcountUpdates(true);
      }

      room.setLimbo(isLimbo);

      setRoomVariables(room, varsOwner, roomVars, setOwnership, false);

      if ((xtName != null) && (xtClass != null))
      {
        int type = 1;

        if (xtClass.endsWith(".as"))
          type = 2;

        if (xtClass.endsWith(".py")) {
          type = 3;
        }

        if (!(xtName.equals("")))
        {
          this.sfs.getExtensionHandler().createExtension(xtName, 
            xtClass, 
            type, 
            z.getName(), 
            room.getId(), 
            lazyInit.equalsIgnoreCase("false"));
        }

      }

      if (sendUpdate)
      {
        String xmlRes = this.sysHandler.getRoomUpdateXML(room);

        SmartFoxServer.getInstance().getEW().handleEvent(new ServerEvent(xmlRes, null, z.getAllUsersInZone()));
      }

      if (!(broadcastEvent)) break label875;
      this.sysHandler.broadcastRoomAdded(z, room);
    }
    catch (CreateRoomException cre)
    {
      SmartFoxServer.log.warning("Could not create room from the server side: " + cre);
      throw new ExtensionHelperException(cre.getMessage());
    }
    catch (Exception e)
    {
      e.printStackTrace();
      SmartFoxServer.log.warning("Exception while creating room from the server side: " + e);
      throw new ExtensionHelperException("Generic error: " + e.getMessage());
    }

    label875: return room;
  }

  public boolean joinRoom(User usr, int currRoom, int newRoom, boolean leaveRoom, String pword, boolean isSpectator, boolean broadcast)
    throws ExtensionHelperException
  {
    Zone zone = this.sfs.getZone(usr.getZone());
    Room oldRoom = null;
    boolean success = false;
    int oldRoomPlayerIndex = 0;

    if (usr == null) {
      throw new ExtensionHelperException(ERROR_JOIN_ROOM + "user is null");
    }

    if (currRoom > -1)
    {
      oldRoom = zone.getRoom(currRoom);

      if (oldRoom != null)
        oldRoomPlayerIndex = usr.getPlayerIndex(oldRoom);
    }
    else {
      leaveRoom = false;
    }

    success = this.sfs.getSysHandler().startJoinTransaction(usr, 
      oldRoom, 
      zone, 
      newRoom, 
      (leaveRoom) ? 1 : 0, 
      oldRoomPlayerIndex, 
      pword, 
      isSpectator, 
      broadcast);

    return success;
  }

  public void sendModeratorMessage(String message, User sender, int type, int id)
  {
    if ((message != null) && (sender != null))
    {
      if (type == 2)
      {
        this.sysHandler.sendAdminMessage(sender, 0, message, getZone(sender.getZone()), "mod");
      }
      else if (type == 1)
      {
        Zone z = getZone(sender.getZone());
        this.sysHandler.sendAdminMessage(sender, 0, message, z.getRoom(id), "mod");
      }
      else if (type == 0)
      {
        this.sysHandler.sendAdminMessage(sender, 0, message, getUserById(id), "mod");
      }
    }
  }

  public void sendRoomList(SocketChannel sc)
  {
    this.sysHandler.requestRoomList(sc);
  }

  /**
   * @deprecated
   */
  public void setRoomVariable(Room r, User u, String vn, String vt, String vv, boolean pr, boolean ps, boolean setOwnership, boolean broadcastAll)
  {
    this.sfs.getSysHandler().setRoomVariable(r, vn, vt, vv, pr, ps, u, setOwnership);
  }

  public void setRoomVariables(Room r, User u, HashMap vars, boolean setOwnership, boolean broadcastAll)
  {
    if ((r != null) && (vars != null))
    {
      Set varSet = vars.entrySet();

      int changeCount = 0;

      StringBuffer response = this.sysHandler.makeHeader("sys", r.getId(), "rVarsUpdate");
      response.append("<vars>");

      for (Iterator it = varSet.iterator(); it.hasNext(); )
      {
        Map.Entry entry = (Map.Entry)it.next();

        String vName = (String)entry.getKey();
        RoomVariable var = (RoomVariable)entry.getValue();

        if (vName.length() <= 0)
          break label152;
        String temp = this.sysHandler.setRoomVariable(r, 
          vName, 
          var.getType(), 
          var.getValue(), 
          var.isPrivate(), 
          var.isPersistent(), 
          u, 
          setOwnership);
        if (temp == null)
          break label152;
        response.append(temp);
        label152: ++changeCount;
      }

      response.append("</vars>");
      response.append(this.sfs.getSysHandler().closeHeader());

      if ((changeCount > 0) && (broadcastAll))
      {
        this.sfs.getEW().handleEvent(new ServerEvent(response.toString(), u.getChannel(), r.getChannellList()));
      }
    }
  }

  public void setUserVariables(User u, HashMap vars, boolean broadcastAll)
  {
    if ((u != null) && (vars != null))
    {
      Set varSet = vars.entrySet();

      StringBuffer response = new StringBuffer("<vars>");

      for (Iterator it = varSet.iterator(); it.hasNext(); )
      {
        Map.Entry entry = (Map.Entry)it.next();

        String vName = (String)entry.getKey();
        UserVariable var = (UserVariable)entry.getValue();

        if (vName.length() <= 0)
          break label115;
        String temp = this.sfs.getSysHandler().setUserVariable(vName, var.getType(), var.getValue(), u);
        label115: response.append(temp);
      }

      response.append("</vars>");

      response.append("<user id='").append(u.getUserId()).append("' />");

      if (broadcastAll)
      {
        int[] roomIds = u.getRoomsConnected();

        Zone z = this.sfs.getZone(u.getZone());

        for (int i = 0; i < roomIds.length; ++i)
        {
          Room r = z.getRoom(roomIds[i]);

          LinkedList allButMe = r.getChannellList();

          StringBuffer message = this.sfs.getSysHandler().makeHeader("sys", roomIds[i], "uVarsUpdate");
          message.append(response).append(this.sfs.getSysHandler().closeHeader());
          this.sfs.getEW().handleEvent(new ServerEvent(message.toString(), u.getChannel(), allButMe));
        }
      }
    }
  }

  public Zone getZone(String zoneName)
  {
    return this.sfs.getZone(zoneName);
  }

  public void leaveRoom(User u, int roomId, boolean broadcastEvent)
  {
    this.sysHandler.leaveRoom(u, roomId, broadcastEvent);
  }

  public void switchSpectator(User u, int roomId, boolean brodcastEvent)
  {
    this.sysHandler.switchSpectator(u, roomId, brodcastEvent);
  }

  public void disconnectUser(User u)
  {
    if (u != null)
      this.sysHandler.kickUser(u.getChannel());
  }

  public boolean destroyRoom(Zone zone, int roomId)
  {
    boolean result = false;

    if (zone != null)
    {
      Room r = zone.getRoom(roomId);

      if ((r != null) && (r.howManyUsers() == 0))
      {
        zone.removeRoom(r, false);
        result = true;
      }
    }

    return result;
  }

  public void addBuddy(String buddyName, User user)
  {
    this.sysHandler.doAddBuddy(buddyName, user);
  }

  public void removeBuddy(String buddyName, User user)
  {
    if (user != null)
    {
      if ((buddyName != null) && (!(buddyName.equals(""))))
      {
        this.sysHandler.doRemoveBuddy(buddyName, user);
        this.sysHandler.sendBuddyRemovedUpdate(buddyName, user);
      }
      else {
        SmartFoxServer.log.warning("RemoveBuddy failed: empty/null buddy name -> " + buddyName);
      }
    }
    else SmartFoxServer.log.warning("RemoveBuddy failed: user is null");
  }

  public void requestAddBuddyPermission(User sender, String targetUserName, String optionalMessage)
  {
    if (sender != null)
    {
      Zone zone = this.sfs.getZone(sender.getZone());
      User targetBuddy = zone.getUserByName(targetUserName);

      this.sysHandler.askAddBuddyPermission(sender, targetBuddy, optionalMessage, zone.getBuddyManager());
    }
  }

  /**
   * @deprecated
   */
  public void clearBuddyList(User user)
  {
    if (user != null)
    {
      this.sysHandler.doClearBuddyList(user);
      this.sysHandler.sendBuddyListCleared(user);
    }
    else {
      SmartFoxServer.log.warning("RemoveBuddy failed: user is null");
    }
  }

  public void setBuddyVariables(User user, Map<String, String> variables)
  {
    this.sysHandler.doSetBuddyVars(user, variables);
  }

  public void sendBuddyListUpdate(User user)
  {
    if (user != null)
    {
      Zone z = getZone(user.getZone());
      LinkedList recipients = z.getBuddyManager().getClientsToUpdate(user.getName());

      this.sysHandler.sendBuddyListUpdate(recipients, user.getName(), user.getUserId(), true, true);
    }
  }

  public void setBuddyBlockStatus(User user, String buddyName, boolean isBlocked)
  {
    this.sysHandler.setBuddyBlockStatus(user, buddyName, isBlocked);
  }
}