package it.gotoandplay.smartfoxserver.extensions;

import it.gotoandplay.smartfoxserver.SmartFoxServer;
import it.gotoandplay.smartfoxserver.config.ConfigData;
import it.gotoandplay.smartfoxserver.data.User;
import it.gotoandplay.smartfoxserver.events.InternalEventObject;
import it.gotoandplay.smartfoxserver.exceptions.MalformedIncludeException;
import it.gotoandplay.smartfoxserver.extensions.as.AsIncludeParser;
import it.gotoandplay.smartfoxserver.extensions.as.ScriptBlock;
import it.gotoandplay.smartfoxserver.lib.ActionscriptObject;
import it.gotoandplay.smartfoxserver.lib.SmartFoxLib;
import java.io.IOException;
import java.io.StringReader;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Timer;
import java.util.TimerTask;
import java.util.logging.Logger;
import org.json.JSONObject;
import org.mozilla.javascript.Context;
import org.mozilla.javascript.EcmaError;
import org.mozilla.javascript.Function;
import org.mozilla.javascript.JavaScriptException;
import org.mozilla.javascript.RhinoException;
import org.mozilla.javascript.Script;
import org.mozilla.javascript.Scriptable;

public class JavascriptExtension extends AbstractExtension
{
  public static int MAINLIB_LINES = -1;
  public static final String js_path = "./sfsExtensions/";
  private static final String lib_js_path = "./lib/";
  private String scriptFileName = null;
  private Scriptable scope;
  private Function ext_init;
  private Function ext_destroy;
  private Function ext_handleInternalEvent;
  private Function ext_handleRequest;
  private Function ext_handleInternalRequest;
  private List scriptBlocks;

  public void setScriptFile(String fName)
  {
    if ((this.scriptFileName == null) && (fName != null))
    {
      this.scriptFileName = fName;
      loadScript();
    }
  }

  public String getScriptFileName()
  {
    return this.scriptFileName;
  }

  public void init()
  {
    try
    {
      callJavaScriptFunction(this.ext_init, new Object[1]);
    }
    catch (JavaScriptException jse)
    {
      SmartFoxServer.log.warning("JS Exception: " + jse);
    }
    catch (Exception e)
    {
      logASError(e);
    }
  }

  public void destroy()
  {
    try
    {
      callJavaScriptFunction(this.ext_destroy, new Object[1]);
    }
    catch (JavaScriptException jse)
    {
      SmartFoxServer.log.warning("JS Exception: " + jse);
    }
    catch (Exception e)
    {
      logASError(e);
    }
  }

  public void handleInternalEvent(InternalEventObject ieo)
  {
    Object[] jsParams;
    try
    {
      jsParams = new Object[1];
      jsParams[0] = ieo;

      callJavaScriptFunction(this.ext_handleInternalEvent, jsParams);
    }
    catch (JavaScriptException jse)
    {
      SmartFoxServer.log.warning("JS Exception in handleInternalEvent: " + jse);
    }
    catch (EcmaError ee)
    {
      logASError(ee);
    }
    catch (Exception e)
    {
      e.printStackTrace();
    }
  }

  public void handleRequest(String cmd, String[] params, User u, int fromRoom)
  {
    Object[] jsParams;
    try
    {
      jsParams = new Object[5];

      jsParams[0] = cmd;
      jsParams[1] = params;
      jsParams[2] = u;
      jsParams[3] = String.valueOf(fromRoom);
      jsParams[4] = "str";

      callJavaScriptFunction(this.ext_handleRequest, jsParams);
    }
    catch (JavaScriptException jse)
    {
      logASError(jse);
    }
    catch (EcmaError ee)
    {
      logASError(ee);
    }
    catch (Exception e)
    {
      e.printStackTrace();
    }
  }

  public void handleRequest(String cmd, ActionscriptObject ao, User u, int fromRoom)
  {
    Object[] params;
    try
    {
      params = new Object[5];

      params[0] = cmd;
      params[1] = ao;
      params[2] = u;
      params[3] = String.valueOf(fromRoom);
      params[4] = "xml";

      callJavaScriptFunction(this.ext_handleRequest, params);
    }
    catch (Exception e)
    {
      logASError(e);
    }
  }

  public void handleRequest(String cmd, JSONObject jso, User u, int fromRoom)
  {
    Object[] params;
    try
    {
      params = new Object[5];

      params[0] = cmd;
      params[1] = jso;
      params[2] = u;
      params[3] = String.valueOf(fromRoom);
      params[4] = "json";

      callJavaScriptFunction(this.ext_handleRequest, params);
    }
    catch (Exception e)
    {
      logASError(e);
    }
  }

  public Object handleInternalRequest(Object params)
  {
    Object response = null;

    if (this.ext_handleInternalRequest == null)
      return response;

    try
    {
      response = callFunctionWithReturnValue(this.ext_handleInternalRequest, new Object[] { params });
    }
    catch (Exception e)
    {
      logASError(e);
    }

    return response;
  }

  private void loadScript()
  {
    String js_script = null;

    String js_lib = getJsLibrary();

    if (MAINLIB_LINES == -1) {
      MAINLIB_LINES = SmartFoxLib.countCodeLines(js_lib);
    }

    try
    {
      Context cx = Context.enter();

      byte[] scriptFile = SmartFoxLib.readFileInClassPath("./sfsExtensions/" + this.scriptFileName);

      AsIncludeParser parser = new AsIncludeParser();
      js_script = parser.parse(this.scriptFileName, new String(scriptFile));

      this.scriptBlocks = parser.getScriptBlocks();

      StringReader reader = new StringReader(js_lib + js_script);

      this.scope = cx.initStandardObjects(null);

      cx.setOptimizationLevel(ConfigData.RHINO_OPTIMIZATION_LEVEL);

      Scriptable outObj = Context.toObject(System.out, this.scope);

      Scriptable server = Context.toObject(SmartFoxServer.getInstance(), this.scope);

      this.scope.put("__out", this.scope, outObj);
      this.scope.put("__smartFoxServer", this.scope, server);
      this.scope.put("__extension", this.scope, this);

      Script compiledScript = cx.compileReader(reader, this.scriptFileName, 1, null);

      compiledScript.exec(cx, this.scope);

      this.ext_init = ((Function)this.scope.get("init", this.scope));
      this.ext_destroy = ((Function)this.scope.get("destroy", this.scope));
      this.ext_handleRequest = ((Function)this.scope.get("__handleJavaRequest", this.scope));
      this.ext_handleInternalEvent = ((Function)this.scope.get("__handleJavaEvent", this.scope));
      this.ext_handleInternalRequest = ((Function)this.scope.get("__handleInternalRequest", this.scope));
    }
    catch (MalformedIncludeException mie)
    {
      SmartFoxServer.log.severe("Error loading extension. " + mie.getMessage());

      Context.exit();
    }
    catch (IOException ioe)
    {
      SmartFoxServer.log.severe("Can't load Javascript Extension: " + this.scriptFileName);

      Context.exit();
    }
    catch (Exception e)
    {
      logASError(e);

      Context.exit(); } finally { Context.exit();
    }
  }

  private String getJsLibrary()
  {
    String js_script = null;
    try
    {
      byte[] scriptFile = SmartFoxLib.readFileInClassPath("./lib/mainLib.as");

      js_script = new String(scriptFile);
    }
    catch (IOException ioe)
    {
      SmartFoxServer.log.severe("Can't load Javascript Main Library!");
    }

    return js_script;
  }

  private void callJavaScriptFunction(String fName, Object[] params)
    throws Exception
  {
    Function fn = (Function)this.scope.get(fName, this.scope);
    callJavaScriptFunction(fn, params);
  }

  private void callJavaScriptFunction(Function fn, Object[] params)
    throws Exception
  {
    if (fn != null)
    {
      Context cx = Context.enter();
      fn.call(cx, this.scope, this.scope, params);

      Context.exit();
    }
  }

  private Object callFunctionWithReturnValue(Function fn, Object[] params)
    throws Exception
  {
    Context cx = Context.enter();
    Object results = fn.call(cx, this.scope, this.scope, params);
    Context.exit();

    return results;
  }

  public Timer setInterval(String fnName, int startDelay, int interval, Object params)
  {
    Function fn = (Function)this.scope.get(fnName, this.scope);

    Timer t = new Timer();
    t.schedule(new IntervalExecutor(this, fn, this, params), startDelay, interval);

    return t;
  }

  private void logASError(Exception e)
  {
    String msg = e.toString();
    String error = e.getMessage();

    int localLineNumber = -1;
    int globalLineNumber = -1;

    String atLine = null;
    String fileWhereErrorOccurred = null;

    if (e instanceof RhinoException)
    {
      RhinoException re = (RhinoException)e;

      globalLineNumber = re.lineNumber();
      localLineNumber = globalLineNumber - MAINLIB_LINES;
    }

    if (localLineNumber != -1)
    {
      atLine = " Internal: " + localLineNumber + " -- Line number: ";

      if (localLineNumber < 1) {
        atLine = atLine + " (MainLib line: " + globalLineNumber + ")";
      }
      else
      {
        ScriptBlock sb = null;
        HashSet includeFiles = new HashSet();

        for (Iterator i = this.scriptBlocks.iterator(); i.hasNext(); )
        {
          sb = (ScriptBlock)i.next();
          if (sb.filName != this.scriptFileName)
            includeFiles.add(sb.filName);

          if ((localLineNumber < sb.absStart) || (localLineNumber > sb.absEnd))
            break label216;
          fileWhereErrorOccurred = sb.filName;
          label216: break;
        }

        int linesCount = 1;

        for (Iterator i = this.scriptBlocks.iterator(); i.hasNext(); )
        {
          sb = (ScriptBlock)i.next();

          if (!(sb.filName.equals(fileWhereErrorOccurred)))
            break label382;
          if ((localLineNumber >= sb.absStart) && (localLineNumber <= sb.absEnd))
          {
            linesCount += localLineNumber - sb.absStart - includeFiles.size() + 1;
            atLine = atLine + String.valueOf(new StringBuilder(String.valueOf(linesCount)).append(" in file: ").append(sb.filName).toString()); break label382:
          }

          label382: linesCount += sb.absEnd - sb.absStart - 1;
        }

      }

    }
    else
    {
      atLine = " (Line number not available)";
    }

    StringBuffer errorMessage = new StringBuffer("Error in extension [ ");
    errorMessage.append(this.scriptFileName).append(" ]: ").append(error).append(atLine);

    this.adminExtension.logRemoteTrace(getOwnerZone(), getOwnerRoom(), errorMessage.toString());

    SmartFoxServer.log.warning(errorMessage.toString());
  }

  class IntervalExecutor extends TimerTask
  {
    private Function jsFunction;
    private ISmartFoxExtension sfe;
    private JavascriptExtension jse;
    private Object params;

    public IntervalExecutor(, Function paramFunction, ISmartFoxExtension paramISmartFoxExtension, Object paramObject)
    {
      this.jsFunction = paramFunction;
      this.sfe = sfe;
      this.jse = ((JavascriptExtension)sfe);
      this.params = paramObject;
    }

    public void run()
    {
      try
      {
        if (!(this.sfe.isActive())) return;
        JavascriptExtension.access$0(this.this$0, this.jsFunction, new Object[] { this.params });
      }
      catch (Exception e)
      {
        SmartFoxServer.log.warning("SetInterval Error: " + e.getMessage());
        JavascriptExtension.access$1(this.jse, e);
      }
    }
  }
}