package it.gotoandplay.smartfoxserver.extensions;

import it.gotoandplay.smartfoxserver.SmartFoxServer;
import it.gotoandplay.smartfoxserver.data.User;
import it.gotoandplay.smartfoxserver.events.InternalEventObject;
import it.gotoandplay.smartfoxserver.lib.ActionscriptObject;
import it.gotoandplay.smartfoxserver.lib.SmartFoxLib;
import java.io.File;
import java.io.InputStream;
import java.util.Properties;
import java.util.logging.Logger;
import org.json.JSONObject;
import org.python.core.Py;
import org.python.core.PyArray;
import org.python.core.PyException;
import org.python.core.PyInteger;
import org.python.core.PyJavaInstance;
import org.python.core.PyObject;
import org.python.core.PyString;
import org.python.core.PyTraceback;
import org.python.util.PythonInterpreter;

public class PythonExtension extends AbstractExtension
{
  public static final String py_path = "./sfsExtensions/";
  public static int MAINLIB_LINES = -1;
  private PythonInterpreter interp;
  private PyObject pyInstance;
  private String scriptFileName = null;
  private PyObject ext_init;
  private PyObject ext_destroy;
  private PyObject ext_handleInternalEvent;
  private PyObject ext_handleRequest;
  private PyObject ext_handleInternalRequest;

  public PythonExtension()
  {
    File f = new File("./sfsExtensions/");

    Properties props = new Properties();
    props.setProperty("python.path", f.getAbsolutePath());
    PythonInterpreter.initialize(System.getProperties(), props, new String[0]);

    this.interp = new PythonInterpreter();
  }

  public void setScriptFile(String fName)
  {
    if ((this.scriptFileName == null) && (fName != null))
    {
      this.scriptFileName = fName;
      loadScript();
    }
  }

  public String getScriptFileName()
  {
    return this.scriptFileName;
  }

  void loadScript()
  {
    InputStream is = null;
    try
    {
      String mainLib = new String(SmartFoxLib.readFileInClassPath("./lib/mainLib.py"));

      if (MAINLIB_LINES == -1)
        MAINLIB_LINES = SmartFoxLib.countCodeLines(mainLib);

      String pyScript = new String(SmartFoxLib.readFileInClassPath("./sfsExtensions/" + this.scriptFileName));
      mainLib = mainLib + pyScript;

      this.interp.exec(mainLib);

      this.interp.set("__smartFoxServer", new PyJavaInstance(SmartFoxServer.getInstance()));
      this.interp.set("__baseExtension", new PyJavaInstance(this));

      this.interp.exec("_server = _Server()");

      PyObject _server = this.interp.get("_server");

      this.ext_init = this.interp.get("init");
      this.ext_destroy = this.interp.get("destroy");
      this.ext_handleRequest = _server.__getattr__("_handleJavaRequest");
      this.ext_handleInternalEvent = this.interp.get("handleInternalEvent");
      this.ext_handleInternalRequest = this.interp.get("handleInternalRequest");
    }
    catch (Exception e) {
      e.printStackTrace();
    }
  }

  public void init()
  {
    try
    {
      this.ext_init.__call__();
    }
    catch (PyException pe)
    {
      handlePyException(pe);
    }
    catch (Exception e)
    {
      logPyError(e.getMessage());
      e.printStackTrace();
    }
  }

  public void destroy()
  {
    try
    {
      this.ext_destroy.__call__();
    }
    catch (PyException pe)
    {
      handlePyException(pe);
    }
    catch (Exception e)
    {
      logPyError(e.getMessage());
      e.printStackTrace();
    }
  }

  public void handleRequest(String cmd, String[] params, User u, int fromRoom)
  {
    try
    {
      PyObject[] args = { 
        new PyString(cmd), 
        new PyArray(String.class, params), 
        new PyJavaInstance(u), 
        new PyInteger(fromRoom), 
        new PyString("str") };

      this.ext_handleRequest.__call__(args);
    }
    catch (PyException pe)
    {
      handlePyException(pe);
    }
    catch (Exception e)
    {
      logPyError(e.getMessage());
      e.printStackTrace();
    }
  }

  public void handleRequest(String cmd, ActionscriptObject ao, User u, int fromRoom)
  {
    try
    {
      PyObject[] args = { 
        new PyString(cmd), 
        new PyJavaInstance(ao), 
        new PyJavaInstance(u), 
        new PyInteger(fromRoom), 
        new PyString("xml") };

      this.ext_handleRequest.__call__(args);
    }
    catch (PyException pe)
    {
      handlePyException(pe);
    }
    catch (Exception e)
    {
      logPyError(e.getMessage());
      e.printStackTrace();
    }
  }

  public void handleRequest(String cmd, JSONObject jso, User u, int fromRoom)
  {
    try
    {
      PyObject[] args = { 
        new PyString(cmd), 
        new PyJavaInstance(jso), 
        new PyJavaInstance(u), 
        new PyInteger(fromRoom), 
        new PyString("json") };

      this.ext_handleRequest.__call__(args);
    }
    catch (PyException pe)
    {
      handlePyException(pe);
    }
    catch (Exception e)
    {
      logPyError(e.getMessage());
      e.printStackTrace();
    }
  }

  public void handleInternalEvent(InternalEventObject ieo)
  {
    try
    {
      this.ext_handleInternalEvent.__call__(new PyJavaInstance(ieo));
    }
    catch (PyException pe)
    {
      handlePyException(pe);
    }
    catch (Exception e)
    {
      logPyError(e.getMessage());
      e.printStackTrace();
    }
  }

  public Object handleInternalRequest(Object obj)
  {
    try
    {
      if (this.ext_handleInternalRequest == null) break label45;
      return this.ext_handleInternalRequest.__call__(new PyJavaInstance(obj));
    }
    catch (PyException pe)
    {
      handlePyException(pe);
    }
    catch (Exception e)
    {
      logPyError(e.getMessage());
      e.printStackTrace();
    }

    label45: return null;
  }

  private void handlePyException(PyException pe)
  {
    pe.printStackTrace();

    Object candExc = pe.value.__tojava__(Throwable.class);

    if (candExc != Py.NoConversion)
    {
      Throwable wrappedExc = (Throwable)candExc;
      logPyError(wrappedExc.toString() + " --- At line: " + (pe.traceback.tb_lineno - MAINLIB_LINES));
    }
    else
    {
      logPyError(pe.toString() + " --- At line: " + (pe.traceback.tb_lineno - MAINLIB_LINES));
    }
  }

  private void logPyError(String msg)
  {
    StringBuffer errorMessage = new StringBuffer("Error in extension [ ");
    errorMessage.append(this.scriptFileName).append(" ]: ").append(msg);

    this.adminExtension.logRemoteTrace(getOwnerZone(), getOwnerRoom(), errorMessage.toString());

    SmartFoxServer.log.warning(errorMessage.toString());
  }
}