package it.gotoandplay.smartfoxserver.extensions.as;

import it.gotoandplay.smartfoxserver.exceptions.MalformedIncludeException;
import it.gotoandplay.smartfoxserver.lib.SmartFoxLib;
import java.io.IOException;
import java.io.PrintStream;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;
import java.util.Stack;

public class AsIncludeParser
{
  private static final String INCLUDE_TOKEN = "#include";
  private List scriptBlocks;

  public AsIncludeParser()
  {
    this.scriptBlocks = new ArrayList();
  }

  public String parse(String fName, String scriptData)
    throws MalformedIncludeException
  {
    int p;
    StringBuffer sb = new StringBuffer();
    String incFileData = null;
    List lines = SmartFoxLib.getListOfLines(scriptData);

    IncludeFile incf = new IncludeFile(fName, 1, lines.size());
    Stack stack = new Stack();
    stack.push(incf);

    int curPos = 0;

    String line = null;
    ScriptBlock currBlock = new ScriptBlock(incf.fileName, 1, 1, 1, 1);
    this.scriptBlocks.add(currBlock);

    IncludeFile currFile = incf;
    while (true)
    {
      if (currFile.currLine > currFile.totLines)
      {
        sb.append("\n");

        if (currBlock.absEnd > currBlock.absStart)
          currBlock.absEnd -= 1;

        if (currBlock.locEnd > currBlock.locStart) {
          currBlock.locEnd -= 1;
        }

        stack.pop();

        if (stack.empty()) {
          break label576:
        }

        currFile = (IncludeFile)stack.peek();

        currBlock = new ScriptBlock(currFile.fileName, curPos + 1, curPos + 1, 1, 1);
        this.scriptBlocks.add(currBlock);
      }

      line = (String)lines.get(curPos);
      p = line.indexOf("#include");

      if (p != -1)
        break;
      sb.append(line);
      ++curPos;

      currBlock.absEnd += 1;
      currBlock.locEnd += 1;

      currFile.currLine += 1;
    }

    int endOfLine = line.indexOf("\n", p);

    int startOfLine = p + "#include".length();

    if (endOfLine - p > 2)
    {
      String incFile = getIncludeFileName(line, startOfLine, endOfLine);
      try
      {
        incFileData = new String(SmartFoxLib.readFileInClassPath("./sfsExtensions/" + incFile));

        List incLines = SmartFoxLib.getListOfLines(incFileData);

        Object o = lines.remove(curPos);

        lines.addAll(curPos, incLines);

        currFile.currLine += 1;

        currFile = new IncludeFile(incFile, 1, incLines.size());
        stack.push(currFile);

        if (currBlock.absEnd > currBlock.absStart)
          currBlock.absEnd -= 1;

        if (currBlock.locEnd > currBlock.locStart) {
          currBlock.locEnd -= 1;
        }

        currBlock = new ScriptBlock(incFile, curPos + 1, curPos + 1, 1, 1);
        this.scriptBlocks.add(currBlock);
      }
      catch (IOException ioe)
      {
        ioe.printStackTrace();
        throw new MalformedIncludeException("Can't load include file: " + incFile + " -> " + incFile.length());
      }

    }

    throw new MalformedIncludeException("Malformed #include directive");

    label576: return sb.toString();
  }

  public List getScriptBlocks()
  {
    return this.scriptBlocks;
  }

  private void raiseError(int line)
  {
    ScriptBlock sb = null;

    for (Iterator i = this.scriptBlocks.iterator(); i.hasNext(); )
    {
      sb = (ScriptBlock)i.next();

      if ((line < sb.absStart) || (line > sb.absEnd))
        break label89;
      int localLine = line - sb.absStart;
      label89: System.out.println("Error in file: " + sb.filName + " at line: " + (sb.locStart + localLine));
    }
  }

  private void dumpBlocks()
  {
    for (int i = 0; i < this.scriptBlocks.size(); ++i)
    {
      ScriptBlock sb = (ScriptBlock)this.scriptBlocks.get(i);
      System.out.println("\n-------------------------------");
      System.out.println("File: " + sb.filName);
      System.out.println("Abs start: " + sb.absStart);
      System.out.println("Abs end  : " + sb.absEnd);
      System.out.println("Loc start: " + sb.locStart);
      System.out.println("Loc end  : " + sb.locEnd);
    }
  }

  public String getIncludeFileName(String s, int startP, int endP)
  {
    StringBuffer sb = new StringBuffer();

    for (int i = startP; i < endP; ++i)
    {
      char c = s.charAt(i);
      if ((c != ' ') && (c != '"') && (c > ' '))
        sb.append(c);
    }

    return sb.toString();
  }
}