package it.gotoandplay.smartfoxserver.extensions.examples;

import it.gotoandplay.smartfoxserver.SmartFoxServer;
import it.gotoandplay.smartfoxserver.data.Zone;
import it.gotoandplay.smartfoxserver.data.buddylist.StorableBuddyList;
import it.gotoandplay.smartfoxserver.data.buddylist.persistence.AbstractBuddyPersister;
import it.gotoandplay.smartfoxserver.db.DataRow;
import it.gotoandplay.smartfoxserver.db.DbManager;
import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.PrintStream;
import java.util.ArrayList;
import java.util.List;
import java.util.Properties;
import java.util.logging.Logger;

public class BuddyListDatabasePersister extends AbstractBuddyPersister
{
  private static final String CFG_FILE = "javaExtensions/DbPersister.properties";
  private Zone currentZone;
  private Logger log = SmartFoxServer.log;
  private DbManager dbManager;
  private boolean isEnabled = true;
  private String db_tableName;
  private String db_field_id;
  private String db_field_userName;
  private String db_field_buddyList;
  private String db_itemSeparator;

  public void init(Object param)
  {
    this.currentZone = SmartFoxServer.getInstance().getZone(this.zoneName);
    this.dbManager = this.currentZone.dbManager;

    if (this.dbManager == null)
    {
      this.isEnabled = false;
      this.log.severe("DbManager is not setup for Zone: " + this.zoneName + ". Buddylist save/load is disabled!");
    }

    try
    {
      loadConfig();
    }
    catch (IOException exception)
    {
      this.isEnabled = false;
      this.log.severe("Could not retrieve Database Persister configuration! Buddylist save/load is disabled!");
    }
  }

  public StorableBuddyList loadList(String userName)
  {
    if (!(this.isEnabled))
      return null;

    List buddyList = null;

    StringBuilder sqlCmd = new StringBuilder("SELECT * FROM ");
    sqlCmd.append(this.db_tableName);
    sqlCmd.append(" WHERE ").append(this.db_field_userName).append("='").append(userName).append("'");

    System.out.println("DBMANAGER: " + this.dbManager);
    ArrayList results = this.dbManager.executeQuery(sqlCmd.toString());

    if (results.size() == 1)
    {
      DataRow row = (DataRow)results.get(0);

      String buddyData = row.getItem(this.db_field_buddyList);

      if (buddyData != null)
      {
        String[] buddies = buddyData.split("\\" + this.db_itemSeparator);

        buddyList = new ArrayList();

        for (int i = 0; i < buddies.length; ++i)
        {
          buddyList.add(buddies[i]);
        }

      }

    }

    return null;
  }

  public boolean saveList(String userName, StorableBuddyList buddyList)
  {
    return true;
  }

  private void loadConfig()
    throws IOException
  {
    Properties configData = new Properties();
    configData.load(new FileInputStream(new File("javaExtensions/DbPersister.properties")));

    this.db_tableName = configData.getProperty("tableName");
    this.db_field_id = configData.getProperty("field_id");
    this.db_field_userName = configData.getProperty("field_userName");
    this.db_field_buddyList = configData.getProperty("field_buddyList");
    this.db_itemSeparator = configData.getProperty("itemSeparator");
  }
}