package it.gotoandplay.smartfoxserver.extensions.examples;

import it.gotoandplay.smartfoxserver.data.Room;
import it.gotoandplay.smartfoxserver.data.User;
import it.gotoandplay.smartfoxserver.data.Zone;
import it.gotoandplay.smartfoxserver.events.InternalEventObject;
import it.gotoandplay.smartfoxserver.exceptions.ExtensionHelperException;
import it.gotoandplay.smartfoxserver.exceptions.LoginException;
import it.gotoandplay.smartfoxserver.extensions.AbstractExtension;
import it.gotoandplay.smartfoxserver.extensions.ExtensionHelper;
import it.gotoandplay.smartfoxserver.lib.ActionscriptObject;
import java.io.PrintStream;
import java.nio.channels.SocketChannel;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.Map;
import java.util.Random;

public class PixelGame extends AbstractExtension
{
  private static int gX = 30;
  private static int gY = 30;
  private int roomCounter;
  private String currZone;
  private Zone zone;
  private ExtensionHelper helper;

  public void init()
  {
    trace("Pixel game etxension started...");

    this.roomCounter = 0;

    this.currZone = getOwnerZone();
    trace("CurrentZoneName: " + this.currZone);
    this.helper = ExtensionHelper.instance();

    this.zone = this.helper.getZone(this.currZone);

    Object[] rList = this.zone.getRooms();

    for (int i = 0; i < rList.length; ++i)
      createGrid((Room)rList[i]);
  }

  public void handleRequest(String cmd, ActionscriptObject ao, User u, int fromRoom)
  {
  }

  public void handleInternalEvent(InternalEventObject ieo)
  {
    if (ieo.getEventName().equals("loginRequest"))
    {
      boolean ok = false;

      User newUser = null;

      String nick = ieo.getParam("nick");
      String pass = ieo.getParam("pass");
      SocketChannel chan = (SocketChannel)ieo.getObject("chan");

      ActionscriptObject res = new ActionscriptObject();

      if (nick.equals(""))
        nick = getRandomName(12);
      trace("LOGIN " + nick);
      try
      {
        newUser = this.helper.canLogin(nick, pass, chan, this.currZone);

        res.put("_cmd", "logOK");
        res.put("id", String.valueOf(newUser.getUserId()));
        res.put("name", newUser.getName());

        ok = true;
      }
      catch (LoginException le)
      {
        trace("Could not login user: " + nick);

        res.put("_cmd", "logKO");
        res.put("err", le.getMessage());
      }

      LinkedList ll = new LinkedList();
      ll.add(chan);

      sendResponse(res, -1, null, ll);

      if (ok)
        this.helper.sendRoomList(chan);
    }
  }

  public void handleRequest(String cmd, String[] params, User u, int fromRoom)
  {
    if (cmd.equals("jme")) {
      lookForRoom(u);
    }
    else if (cmd.equals("grid")) {
      sendStatus(u, fromRoom);
    }
    else if (cmd.equals("upd"))
      handleUpdate(params, u, fromRoom);
  }

  private void handleUpdate(String[] params, User who, int fromRoom)
  {
    Room r = this.zone.getRoom(fromRoom);

    if (r != null)
    {
      char[][] grid = (char[][])r.properties.get("grid");
      try
      {
        char ch = params[0].charAt(0);
        int px = Integer.parseInt(params[1]);
        int py = Integer.parseInt(params[2]);

        grid[py][px] = ch;

        String[] res = { "upd", params[0], params[1], params[2] };

        LinkedList allButMe = r.getChannellList();
        allButMe.remove(who.getChannel());

        sendResponse(res, fromRoom, who, allButMe);
      }
      catch (NumberFormatException nfe)
      {
        trace("Bad number values in update request!");
      }
      catch (ArrayIndexOutOfBoundsException aioobe)
      {
        trace("Bad grid indexes in update request!");
      }
    }
  }

  private void lookForRoom(User u)
  {
    Object[] rList = this.zone.getRooms();

    Room room = null;

    boolean found = false;

    int i = 0; break label130:

    room = (Room)rList[i];

    if (room.howManyUsers() < room.getMaxUsers());
    try
    {
      trace("Found a room for user >> " + room.getName());
      this.helper.joinRoom(u, -1, room.getId(), false, "", false, true);

      label130: found = true;
    }
    catch (ExtensionHelperException ehe)
    {
      do {
        trace("Oops, user: " + u.getName() + " couldn't join. Reason: " + ehe.getMessage());

        ++i; } while (i < rList.length);
    }

    if (!(found))
    {
      String rName = "AutoRoom_" + this.roomCounter;
      this.roomCounter += 1;

      Room newRoom = makeNewRoom(rName, 4, u);

      if (newRoom != null)
      {
        createGrid(newRoom);
        try
        {
          this.helper.joinRoom(u, -1, newRoom.getId(), false, "", false, true);
        }
        catch (ExtensionHelperException ehe)
        {
          trace("Oops, user: " + u.getName() + " couldn't join. Reason: " + ehe.getMessage());
        }
      }
    }
  }

  private Room makeNewRoom(String name, int maxU, User u)
  {
    Room newRoom = null;

    HashMap map = new HashMap();
    map.put("name", name);
    map.put("pwd", "");
    map.put("maxU", String.valueOf(maxU));
    map.put("maxS", "0");
    map.put("isGame", "false");
    try
    {
      newRoom = this.helper.createRoom(this.zone, map, u, null, null, false, true, true);
    }
    catch (ExtensionHelperException ehe)
    {
      trace("Could not create room: " + name + ", Reason: " + ehe.getMessage());
    }

    return newRoom;
  }

  private void createGrid(Room r)
  {
    char[][] grid = new char[gY][gX];

    for (int y = 0; y < gY; ++y)
    {
      for (int x = 0; x < gX; ++x)
      {
        grid[y][x] = 48;
      }

    }

    r.properties.put("grid", grid);
  }

  private void sendStatus(User who, int roomId)
  {
    Room r = this.zone.getRoom(roomId);

    if (r != null)
    {
      char[][] grid = (char[][])r.properties.get("grid");

      StringBuffer gridData = new StringBuffer();

      for (int y = 0; y < gY; ++y)
      {
        for (int x = 0; x < gX; ++x)
        {
          gridData.append(grid[y][x]);
        }

      }

      String[] res = { "grid", gridData.toString() };

      LinkedList ll = new LinkedList();
      ll.add(who.getChannel());

      sendResponse(res, roomId, who, ll);
    }
  }

  private String getRandomName(int n)
  {
    StringBuffer sb = new StringBuffer();
    Random rnd = new Random();

    for (int i = 0; i < n; ++i)
    {
      int r = rnd.nextInt(26);
      sb.append((char)(97 + r));
    }

    return sb.toString();
  }

  private void dumpGrid(char[][] grid)
  {
    for (int y = 0; y < gY; ++y)
    {
      for (int x = 0; x < gX; ++x)
      {
        System.out.print(grid[y][x]);
      }

      System.out.println("");
    }
  }
}