package it.gotoandplay.smartfoxserver.extensions.stresstest;

import it.gotoandplay.smartfoxserver.SmartFoxServer;
import it.gotoandplay.smartfoxserver.data.Room;
import it.gotoandplay.smartfoxserver.data.User;
import it.gotoandplay.smartfoxserver.data.Zone;
import it.gotoandplay.smartfoxserver.events.InternalEventObject;
import it.gotoandplay.smartfoxserver.exceptions.ExtensionHelperException;
import it.gotoandplay.smartfoxserver.exceptions.LoginException;
import it.gotoandplay.smartfoxserver.extensions.AbstractExtension;
import it.gotoandplay.smartfoxserver.extensions.ExtensionHelper;
import it.gotoandplay.smartfoxserver.lib.ActionscriptObject;
import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.nio.channels.SocketChannel;
import java.util.HashMap;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.Properties;
import java.util.logging.Logger;
import org.json.JSONObject;

public class DynamicRoomTest extends AbstractExtension
{
  private static final String CMD_JOIN_ME = "jme";
  private int usersPerRoom = 10;
  private Zone currZone;
  private SmartFoxServer sfs;
  private ExtensionHelper helper;
  private Logger log;
  private int roomCount = 0;
  private int guestCount = 0;

  public void init()
  {
    this.sfs = SmartFoxServer.getInstance();
    this.log = SmartFoxServer.log;
    this.helper = ExtensionHelper.instance();

    this.currZone = this.sfs.getZone(getOwnerZone());

    this.log.info("Dynamic Stress Test << STARTED >>");

    loadConfig();
  }

  public void destroy()
  {
    this.log.info("Dynamic Stress Test << STOPPED >>");
  }

  public void handleRequest(String cmd, JSONObject jso, User who, int fromRoom)
  {
    try
    {
      if (!(cmd.equals("jme"))) return;

      joinUser(who);
    }
    catch (ExtensionHelperException exc)
    {
      exc.printStackTrace();
    }
  }

  public void handleInternalEvent(InternalEventObject ieo)
  {
    String evt = ieo.getEventName();

    if (evt.equals("loginRequest"))
    {
      try
      {
        handleLogin(ieo);
      }
      catch (Exception e)
      {
        e.printStackTrace();
      }
    }
  }

  private void loadConfig()
  {
    Properties props;
    try
    {
      props = new Properties();
      FileInputStream cfgFile = new FileInputStream(new File("./javaExtensions/dynamic_room_test.properties"));
      props.load(cfgFile);

      this.usersPerRoom = Integer.parseInt(props.getProperty("usersPerRoom"));

      this.log.info("Configuration load successfully. Users/Room = " + this.usersPerRoom);
    }
    catch (IOException ioe)
    {
      this.log.info("Configuration file not found, using default values");
    }
  }

  private void handleLogin(InternalEventObject event) throws ExtensionHelperException
  {
    SocketChannel socket = (SocketChannel)event.getObject("chan");

    String guestNickName = getGuestName();
    try
    {
      User user = this.helper.canLogin(guestNickName, "", socket, getOwnerZone());
      joinUser(user);
    }
    catch (LoginException e)
    {
      e.printStackTrace();
    }
  }

  public String getGuestName()
  {
    return "guest_user_" + (this.guestCount++);
  }

  private void joinUser(User who)
    throws ExtensionHelperException
  {
    if (who == null)
      return;

    LinkedList roomList = this.currZone.getRoomList();
    boolean found = false;

    for (Iterator iter = roomList.iterator(); iter.hasNext(); )
    {
      Room room = (Room)iter.next();

      if (room.howManyUsers() >= room.getMaxUsers())
        break label74;
      this.helper.joinRoom(who, -1, room.getId(), false, "", false, true);
      found = true;
      label74: break;
    }

    if (!(found))
    {
      String roomName = "Test_Room_" + (this.roomCount++);

      Room newRoom = makeNewRoom(roomName, this.usersPerRoom, who);

      if (newRoom != null)
      {
        this.log.info("New room created: " + roomName);
        this.helper.joinRoom(who, -1, newRoom.getId(), false, "", false, true);
      }
      else
      {
        this.log.warning("Failed creating new room: " + roomName);
      }
    }
  }

  private Room makeNewRoom(String rName, int maxUsers, User who)
    throws ExtensionHelperException
  {
    HashMap params = new HashMap();
    params.put("name", rName);
    params.put("pwd", "");
    params.put("maxU", String.valueOf(maxUsers));
    params.put("maxS", String.valueOf("0"));
    params.put("isGame", "false");

    return this.helper.createRoom(this.currZone, params, who, false, true);
  }

  public void handleRequest(String cmd, ActionscriptObject ao, User u, int fromRoom)
  {
  }

  public void handleRequest(String cmd, String[] params, User u, int fromRoom)
  {
  }
}