package it.gotoandplay.smartfoxserver.lib;

import it.gotoandplay.smartfoxserver.SmartFoxServer;
import java.io.PrintStream;
import java.util.Enumeration;
import java.util.HashMap;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.Map;
import java.util.logging.Logger;
import net.n3.nanoxml.IXMLElement;
import net.n3.nanoxml.IXMLParser;
import net.n3.nanoxml.IXMLReader;
import net.n3.nanoxml.StdXMLReader;
import net.n3.nanoxml.XMLElement;
import net.n3.nanoxml.XMLParserFactory;

public class ASObjectSerializer
{
  private static Map asciiTable_e = new HashMap();
  private static Map asciiTable_d;

  static
  {
    asciiTable_e.put(">", "&gt;");
    asciiTable_e.put("<", "&lt;");
    asciiTable_e.put("&", "&amp;");
    asciiTable_e.put("'", "&apos;");
    asciiTable_e.put("\"", "&quot;");

    asciiTable_d = new HashMap();

    asciiTable_d.put("&gt;", ">");
    asciiTable_d.put("&lt;", "<");
    asciiTable_d.put("&amp;", "&");
    asciiTable_d.put("&apos;", "'");
    asciiTable_d.put("&quot;", "\"");
  }

  private static String encodeEntities(String str)
  {
    StringBuffer in_sb = new StringBuffer(str);
    StringBuffer out_sb = new StringBuffer();

    int i = 0;

    while (i < in_sb.length())
    {
      char c = in_sb.charAt(i);

      if ((c == '\t') || (c == '\n') || (c == '\r'))
      {
        out_sb.append(c);
      }
      else if ((c >= ' ') && (c <= '~'))
      {
        Object o = asciiTable_e.get(new String(c));

        if (o != null)
        {
          out_sb.append((String)o);
        }
        else
        {
          out_sb.append(c);
        }

      }
      else
      {
        out_sb.append(c);
      }
      ++i;
    }

    return out_sb.toString();
  }

  public static ActionscriptObject deserialize(String xmlData)
  {
    ActionscriptObject ao = new ActionscriptObject();
    try
    {
      IXMLParser parser = XMLParserFactory.createDefaultXMLParser();

      IXMLReader reader = StdXMLReader.stringReader(xmlData);
      parser.setReader(reader);
      IXMLElement xml = (IXMLElement)parser.parse();

      xml2obj(xml, ao, 0);
    }
    catch (Exception e)
    {
      SmartFoxServer.log.warning(e.toString());
    }

    return ao;
  }

  public static String serialize(ActionscriptObject ao)
  {
    StringBuffer xmlData = new StringBuffer();
    obj2xml(ao, 0, "", xmlData);

    return xmlData.toString();
  }

  private static void xml2obj(IXMLElement node, ActionscriptObject ao, int n)
  {
    Enumeration subNodes = node.enumerateChildren();

    while (subNodes.hasMoreElements())
    {
      String name;
      IXMLElement subNode = (IXMLElement)subNodes.nextElement();

      String subNodeName = subNode.getName();

      if (subNodeName.equals("obj"))
      {
        name = subNode.getAttribute("o", "");

        ActionscriptObject subAO = new ActionscriptObject();

        ao.put(name, subAO);

        xml2obj(subNode, subAO, n + 1);
      }
      else if (subNodeName.equals("var"))
      {
        name = subNode.getAttribute("n", "");
        String type = subNode.getAttribute("t", "");
        String val = subNode.getContent();

        Object varValue = null;

        if (type.equals("b")) {
          varValue = new Boolean(false);
        }
        else if (type.equals("s")) {
          varValue = val;
        }
        else if (type.equals("n")) {
          varValue = new Double(Double.parseDouble(val));
        }

        ao.put(name, varValue);
      }
    }
  }

  private static void obj2xml(ActionscriptObject ao, int depth, String nodeName, StringBuffer xmlData)
  {
    if (depth == 0)
    {
      xmlData.append("<dataObj>");
    }
    else
    {
      xmlData.append("<obj o='").append(nodeName).append("' t='a'>");
    }

    LinkedList keys = new LinkedList(ao.keySet());

    Iterator i = keys.iterator();

    while (i.hasNext())
    {
      String key = (String)i.next();

      Object o = ao.get(key);

      if (o == null)
      {
        xmlData.append("<var n='").append(key).append("' t='x' />");
      } else {
        IXMLElement temp;
        if (o instanceof ActionscriptObject)
        {
          obj2xml((ActionscriptObject)o, depth + 1, key, xmlData);

          xmlData.append("</obj>");

          temp = new XMLElement("obj");
        }
        else if (o instanceof Double)
        {
          double d = ((Double)o).doubleValue();
          String number = new String();

          if (d == Math.rint(d))
            number = ()d;
          else
            number = d;

          xmlData.append("<var n='").append(key).append("' t='n'>").append(number).append("</var>");
        }
        else if (o instanceof String)
        {
          xmlData.append("<var n='").append(key).append("' t='s'>").append(encodeEntities((String)o)).append("</var>");
        }
        else if (o instanceof Boolean)
        {
          boolean b = ((Boolean)o).booleanValue();
          xmlData.append("<var n='").append(key).append("' t='b'>").append((b) ? "1" : "0").append("</var>");
        }
      }

    }

    if (depth == 0)
    {
      xmlData.append("</dataObj>");
    }
  }

  private static void echo(String s, int n)
  {
    echoTabs(n);
    System.out.println(s);
  }

  private static void echoTabs(int n)
  {
    for (int i = 0; i < n; ++i)
      System.out.print("\t");
  }

  public static void dump(ActionscriptObject ao, int depth)
  {
    echo("", 0);

    LinkedList keys = new LinkedList(ao.keySet());

    Iterator i = keys.iterator();

    while (i.hasNext())
    {
      String key = (String)i.next();

      Object o = ao.get(key);

      String output = "";

      if (o == null)
      {
        output = "Null";
      }
      else if (o instanceof ActionscriptObject)
      {
        output = "Object: ";
        dump((ActionscriptObject)o, depth + 1);
      }
      else if (o instanceof Double)
      {
        output = "Number: " + ((Double)o).toString();
      }
      else if (o instanceof String)
      {
        output = "String: " + ((String)o).toString();
      }
      else if (o instanceof Boolean)
      {
        output = "Bool: " + ((Boolean)o).toString();
      }

      echo("[ " + key + " ] > " + output, depth);
    }
  }
}