package it.gotoandplay.smartfoxserver.lib;

import it.gotoandplay.smartfoxserver.db.DataRow;
import java.util.Collection;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;

public class ActionscriptObject
{
  private Map objData;

  public ActionscriptObject()
  {
    this.objData = new HashMap();
  }

  public ActionscriptObject(Collection collection)
  {
    putCollection(this, "", collection);
  }

  public ActionscriptObject(Map map)
  {
    putMap(this, "", map);
  }

  public ActionscriptObject(DataRow row)
  {
    putDataRow(this, "", row);
  }

  public void put(String key, Object o)
  {
    this.objData.put(key, o);
  }

  public void put(int key, Object o)
  {
    this.objData.put(String.valueOf(key), o);
  }

  public void putNumber(int key, double d)
  {
    this.objData.put(String.valueOf(key), new Double(d));
  }

  public void putNumber(String key, double d)
  {
    this.objData.put(key, new Double(d));
  }

  public void putBool(int key, boolean b)
  {
    this.objData.put(String.valueOf(key), new Boolean(b));
  }

  public void putBool(String key, boolean b)
  {
    this.objData.put(key, new Boolean(b));
  }

  public Object get(String key)
  {
    return this.objData.get(key);
  }

  public Object get(int key)
  {
    return this.objData.get(String.valueOf(key));
  }

  public String getString(int key)
  {
    String s = (String)this.objData.get(String.valueOf(key));

    if (s == null)
      return "";

    return s;
  }

  public String getString(String key)
  {
    String s = (String)this.objData.get(key);
    if (s == null)
      return "";

    return s;
  }

  public double getNumber(int key)
  {
    Double var = (Double)this.objData.get(String.valueOf(key));

    if (var == null)
      return 0.0D;

    return var.doubleValue();
  }

  public double getNumber(String key)
  {
    Double var = (Double)this.objData.get(key);

    if (var == null)
      return 0.0D;

    return var.doubleValue();
  }

  public boolean getBool(int key)
  {
    Boolean var = (Boolean)this.objData.get(String.valueOf(key));

    if (var == null)
      return false;

    return var.booleanValue();
  }

  public boolean getBool(String key)
  {
    Boolean var = (Boolean)this.objData.get(key);

    if (var == null)
      return false;

    return var.booleanValue();
  }

  public ActionscriptObject getObj(int key)
  {
    return ((ActionscriptObject)this.objData.get(String.valueOf(key)));
  }

  public ActionscriptObject getObj(String key)
  {
    return ((ActionscriptObject)this.objData.get(key));
  }

  public int size()
  {
    return this.objData.size();
  }

  public Set keySet()
  {
    return this.objData.keySet();
  }

  public Object removeElement(int key)
  {
    return this.objData.remove(String.valueOf(key));
  }

  public Object removeElement(String key)
  {
    return this.objData.remove(key);
  }

  public void putCollection(String key, Collection collection)
  {
    putCollection(new ActionscriptObject(), key, collection);
  }

  public void putCollection(int key, Collection collection)
  {
    putCollection(String.valueOf(key), collection);
  }

  private void putCollection(ActionscriptObject aObj, String key, Collection collection)
  {
    int count = 0;

    if (aObj != this)
      put(key, aObj);

    for (Iterator it = collection.iterator(); it.hasNext(); )
    {
      Object collectionItem = it.next();

      if (collectionItem instanceof Collection) {
        aObj.putCollection(count, (Collection)collectionItem);
      }
      else if (collectionItem instanceof Map) {
        aObj.putMap(count, (Map)collectionItem);
      }
      else if (collectionItem instanceof DataRow) {
        aObj.putDataRow(count, (DataRow)collectionItem);
      }
      else
        aObj.put(count, collectionItem);

      ++count;
    }
  }

  public void putMap(String key, Map map)
  {
    putMap(new ActionscriptObject(), key, map);
  }

  public void putMap(int key, Map map)
  {
    putMap(String.valueOf(key), map);
  }

  private void putMap(ActionscriptObject aObj, String key, Map map)
  {
    if (aObj != this)
      put(key, aObj);

    for (Iterator it = map.entrySet().iterator(); it.hasNext(); )
    {
      Map.Entry entryItem = (Map.Entry)it.next();
      String itemKey = (String)entryItem.getKey();
      Object itemVal = entryItem.getValue();

      if (itemVal instanceof Collection) {
        aObj.putCollection(itemKey, (Collection)itemVal); break label134:
      }
      if (itemVal instanceof Map) {
        aObj.putMap(itemKey, (Map)itemVal); break label134:
      }
      if (itemVal instanceof DataRow) {
        aObj.putDataRow(itemKey, (DataRow)itemVal); break label134:
      }

      label134: aObj.put(itemKey, itemVal);
    }
  }

  public void putDataRow(String key, DataRow row)
  {
    putDataRow(new ActionscriptObject(), key, row);
  }

  public void putDataRow(int key, DataRow row)
  {
    putDataRow(String.valueOf(key), row);
  }

  private void putDataRow(ActionscriptObject aObj, String key, DataRow row)
  {
    if (row.getType() == 0) {
      putCollection(aObj, key, row.getDataAsList());
    }
    else
      putMap(aObj, key, row.getDataAsMap());
  }
}