package it.gotoandplay.smartfoxserver.lib;

import it.gotoandplay.smartfoxserver.SmartFoxServer;
import java.io.PrintStream;
import java.nio.channels.SocketChannel;
import java.util.logging.Logger;

public abstract class ChannelWrap
  implements Runnable, EventHandler
{
  Logger log = SmartFoxServer.log;
  protected ChannelQueue eventQueue;
  protected boolean running = false;
  private Thread[] workers;
  private int spareCount;
  private Object countLock = new Object();
  private String shortname;

  public final void initWrap(int numWorkers)
  {
    this.shortname = super.getClass().getName().substring(super.getClass().getName().lastIndexOf(".") + 1);

    this.eventQueue = new ChannelQueue(this.shortname + "-in");

    this.workers = new Thread[numWorkers];
    for (int i = 0; i < numWorkers; ++i)
    {
      this.workers[i] = new Thread(this, this.shortname + "-" + (i + 1));
      this.workers[i].setDaemon(true);
      this.workers[i].start();
    }
  }

  public void shutdown()
  {
    this.running = false;
    if (this.workers != null)
    {
      for (int i = 0; i < this.workers.length; ++i)
      {
        this.workers[i].interrupt();
      }
    }
  }

  public void handleEvent(SocketChannel chan)
  {
    this.eventQueue.enQueue(chan);
  }

  public int getQueueSize()
  {
    return this.eventQueue.size();
  }

  public void run()
  {
    this.running = true;

    label37: while (this.running)
    {
      try
      {
        SocketChannel chan;
        if ((chan = this.eventQueue.deQueue()) == null) break label37;
        processEvent(chan);
      }
      catch (InterruptedException e)
      {
        System.out.println("Problems in wrap!");
      }
    }
  }

  protected abstract void processEvent(SocketChannel paramSocketChannel);
}