package it.gotoandplay.smartfoxserver.lib;

import it.gotoandplay.smartfoxserver.SmartFoxServer;
import it.gotoandplay.smartfoxserver.config.ConfigData;
import it.gotoandplay.smartfoxserver.exceptions.ConfigurationException;
import it.gotoandplay.smartfoxserver.extensions.ExtensionItem;
import it.gotoandplay.smartfoxserver.py.IPyClass;
import it.gotoandplay.smartfoxserver.py.PyHelper;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.PrintStream;
import java.util.ArrayList;
import java.util.Enumeration;
import java.util.LinkedList;
import java.util.logging.Level;
import net.n3.nanoxml.IXMLElement;
import net.n3.nanoxml.IXMLParser;
import net.n3.nanoxml.IXMLReader;
import net.n3.nanoxml.StdXMLReader;
import net.n3.nanoxml.XMLParserFactory;

public class ConfigReader
{
  public static String xmlFile = "config.xml";
  private static final String licProType = "7BF627CEDE8A99485A92F707FA4E9EEB8FDFB7F88D3ACD8DB8E7A22BDB8A08BD30526C18FD9CB90EDF82E10C4E60BD6A0DB7B79AC898CE6E18F1384ADD";
  private static final String licMaxClient = "3747E67A56DAD9C7A896C6348716A8D9DDD4B1613A61B778D729624EA2FF214E0E7C48D1C0D8523F60184B6222B2913C285BD7C2EBFCBE3BED5D381B50";
  private static final String licAddresses = "3708EF562E2C3837375680FE2746117C3F4614286230072E941E12FFEFD089FADBDCE91D09FF22E77AC12B8A8890EF0E8BA25A502B78F1FA121F";
  private boolean success;
  private TLC tlc;
  private static ConfigReader _instance;
  private IXMLElement sfsXmlConfig;
  private IXMLElement zonesXml;
  private IXMLParser cfgParser;

  private ConfigReader()
  {
    this.success = false;
  }

  public static ConfigReader getInstance()
  {
    if (_instance == null)
      _instance = new ConfigReader();

    return _instance;
  }

  public static IXMLElement loadConfigFile(String xmlFile)
    throws Exception
  {
    IXMLParser parser = XMLParserFactory.createDefaultXMLParser();

    String xmlCfgData = new String(SmartFoxLib.readFileInClassPath(xmlFile));
    IXMLReader reader = StdXMLReader.stringReader(xmlCfgData);

    parser.setReader(reader);

    return ((IXMLElement)parser.parse());
  }

  private IXMLElement loadLicenceFile(String fileName)
    throws Exception
  {
    byte[] licFile = SmartFoxLib.readFileInClassPath(fileName);

    int keyLen = licFile[0];
    byte[] key = new byte[keyLen];
    byte[] data = new byte[licFile.length - 1 - keyLen];

    System.arraycopy(licFile, 1, key, 0, keyLen);

    System.arraycopy(licFile, 1 + keyLen, data, 0, licFile.length - 1 - keyLen);

    LicenceManager lm = new LicenceManager(key);

    String xmlData = new String(lm.decrypt(data));

    IXMLParser xmlParser = XMLParserFactory.createDefaultXMLParser();

    IXMLReader reader = StdXMLReader.stringReader(xmlData);

    xmlParser.setReader(reader);

    return ((IXMLElement)xmlParser.parse());
  }

  public void runHelper()
  {
    try
    {
      String res = (String)SmartFoxServer.getInstance().getPyHelper().getPyClass("taskGenHelper").call("scan", new Object[] { loadLicenceFile("licence.sfl") });

      if (res == null) return;
      throw new ConfigurationException(res);
    }
    catch (IOException ioe)
    {
      System.out.println("\nCould not load Licence File. Aborting.");
      System.exit(-1);
    }
    catch (Exception e)
    {
      System.out.println("\nLicence file is invalid! Aborting.\n" + e);
      e.printStackTrace();
      System.exit(-1);
    }
  }

  private void readCfgFile()
    throws Exception
  {
    this.cfgParser = XMLParserFactory.createDefaultXMLParser();

    String xmlCfgData = new String(
      SmartFoxLib.readFileInClassPath(xmlFile));
    IXMLReader reader = StdXMLReader.stringReader(xmlCfgData);

    this.cfgParser.setReader(reader);
    this.sfsXmlConfig = ((IXMLElement)this.cfgParser.parse());
  }

  public void readServerConfig()
    throws ConfigurationException
  {
    try
    {
      readCfgFile();

      IXMLElement cfgNode = this.sfsXmlConfig.getFirstChildNamed("ServerSetup");

      if (cfgNode != null)
      {
        parse_ServerSetup(cfgNode); return;
      }

      throw new ConfigurationException("<ServerSetup></ServerSetup> block is missing!");
    }
    catch (FileNotFoundException fnfe)
    {
      fnfe.printStackTrace();
      throw new ConfigurationException("Can't open configuration file.");
    }
    catch (NumberFormatException nfe)
    {
      throw new ConfigurationException(nfe.getMessage());
    }
    catch (Exception e)
    {
      e.printStackTrace();
      throw new ConfigurationException(e.getMessage());
    }
  }

  public boolean readZoneConfig()
    throws Exception
  {
    IXMLElement zoneXml;
    try
    {
      zoneXml = this.sfsXmlConfig.getFirstChildNamed("Zones");

      if (zoneXml != null)
      {
        Enumeration cfgNodes = this.sfsXmlConfig.enumerateChildren();

        while (cfgNodes.hasMoreElements())
        {
          IXMLElement cfgNode = (IXMLElement)cfgNodes.nextElement();

          String nodeName = cfgNode.getName();

          if (nodeName.equalsIgnoreCase("Zones"))
          {
            this.zonesXml = cfgNode;
            parse_Zones(this.zonesXml);
          }
        }
        this.success = true; break label146:
      }

      throw new ConfigurationException("<Zones></Zones> block is missing!");
    }
    catch (FileNotFoundException fnfe)
    {
      fnfe.printStackTrace();
      throw new ConfigurationException("Can't open configuration file.");
    }
    catch (NumberFormatException nfe)
    {
      nfe.printStackTrace();
      throw new ConfigurationException(nfe.getMessage());
    }
    catch (Exception e)
    {
      e.printStackTrace();
      throw new ConfigurationException(e.getMessage());
    }
    label146: return this.success;
  }

  private void parse_ServerSetup(IXMLElement node) throws Exception
  {
    Enumeration subNodes = node.enumerateChildren();

    while (subNodes.hasMoreElements())
    {
      IXMLElement option = (IXMLElement)subNodes.nextElement();

      String tagName = option.getFullName();
      String tagValue = option.getContent();
      try
      {
        int val;
        if (tagName.equalsIgnoreCase("ServerIP"))
        {
          ConfigData.SERVER_ADDR = tagValue; break label1402:
        }
        if (tagName.equalsIgnoreCase("ServerPort"))
        {
          ConfigData.SERVER_PORT = Integer.parseInt(tagValue); break label1402:
        }
        if (tagName.equalsIgnoreCase("AutoSendPolicyFile"))
        {
          if (tagValue.equalsIgnoreCase("true")) {
            ConfigData.AUTO_CROSSDOMAIN = true; break label1402:
          }
          ConfigData.AUTO_CROSSDOMAIN = false; break label1402:
        }
        if (tagName.equalsIgnoreCase("PolicyAllowedDomains"))
        {
          parse_AllowedDomains(option); break label1402:
        }
        if (tagName.equalsIgnoreCase("OutQueueThreads"))
        {
          ConfigData.OUT_QUEUE_THREADS = Integer.parseInt(tagValue); break label1402:
        }
        if (tagName.equalsIgnoreCase("ExtHandlerThreads"))
        {
          ConfigData.EXT_HANDLER_THREADS = Integer.parseInt(tagValue); break label1402:
        }
        if (tagName.equalsIgnoreCase("MaxUserIdleTime"))
        {
          ConfigData.MAX_USER_IDLETIME = Integer.parseInt(tagValue); break label1402:
        }
        if (tagName.equalsIgnoreCase("MinMsgTime"))
        {
          ConfigData.MIN_MSG_TIME = Integer.parseInt(tagValue); break label1402:
        }
        if (tagName.equalsIgnoreCase("MaxMsgLen"))
        {
          ConfigData.MAX_MSG_LEN = Integer.parseInt(tagValue); break label1402:
        }
        if (tagName.equalsIgnoreCase("LogMaxSize"))
        {
          ConfigData.LOG_MAX_SIZE = Integer.parseInt(tagValue); break label1402:
        }
        if (tagName.equalsIgnoreCase("LogMaxFiles"))
        {
          ConfigData.LOG_MAX_FILES = Integer.parseInt(tagValue); break label1402:
        }
        if (tagName.equalsIgnoreCase("MaxWriterQueue"))
        {
          ConfigData.MAX_CHANNEL_QUEUE = Integer.parseInt(tagValue); break label1402:
        }
        if (tagName.equalsIgnoreCase("SysHandlerThreads"))
        {
          ConfigData.SYS_HANDLER_THREADS = 1; break label1402:
        }
        if (tagName.equalsIgnoreCase("AdminLogin"))
        {
          SmartFoxServer.getInstance().setAdminLogin(tagValue); break label1402:
        }
        if (tagName.equalsIgnoreCase("AdminPassword"))
        {
          if (tagValue.length() < 6)
            throw new Exception("Admin password must be at least 6 characters long!");

          SmartFoxServer.getInstance().setAdminPassword(tagValue); break label1402:
        }
        if (tagName.equalsIgnoreCase("AdminAllowedAddresses"))
        {
          setAdminAddresses(option); break label1402:
        }
        if (tagName.equalsIgnoreCase("BanCleaning"))
        {
          ConfigData.BAN_CLEAN_MODE = tagValue; break label1402:
        }
        if (tagName.equalsIgnoreCase("BanDuration"))
        {
          ConfigData.BAN_DURATION = 
            Long.parseLong(tagValue);

          break label1402:
        }

        if (tagName.equalsIgnoreCase("BannedLoginMessage"))
        {
          ConfigData.BANNED_LOGIN_MESSAGE = tagValue; break label1402:
        }
        if (tagName.equalsIgnoreCase("AntiFlood"))
        {
          parse_AntiFlood(option); break label1402:
        }
        if (tagName.equalsIgnoreCase("BadWordsFilter"))
        {
          parse_badWordFilter(option); break label1402:
        }
        if (tagName.equalsIgnoreCase("FileLoggingLevel"))
        {
          ConfigData.FILE_LOGGIN_LEVEL = parseLogginLevel(tagValue); break label1402:
        }
        if (tagName.equalsIgnoreCase("ConsoleLoggingLevel"))
        {
          ConfigData.CONSOLE_LOGGIN_LEVEL = parseLogginLevel(tagValue); break label1402:
        }
        if (tagName.equalsIgnoreCase("IpFilter"))
        {
          ConfigData.MAX_USERS_PER_IP = Integer.parseInt(tagValue); break label1402:
        }
        if (tagName.equalsIgnoreCase("Mailer"))
        {
          parse_Mailer(option); break label1402:
        }
        if (tagName.equalsIgnoreCase("ExtensionRemoteDebug"))
        {
          if (!(tagValue.equalsIgnoreCase("false"))) break label1402;
          ConfigData.EXTENSION_REMOTE_DEBUGGING = false; break label1402:
        }
        if (tagName.equalsIgnoreCase("AllowAnonXtRequests"))
        {
          if (tagValue.equalsIgnoreCase("false")) {
            ConfigData.ALLOW_ANON_XT_MESSAGES = false; break label1402:
          }
          ConfigData.ALLOW_ANON_XT_MESSAGES = true; break label1402:
        }
        if (tagName.equalsIgnoreCase("MaxRoomVars"))
        {
          val = -1;
          try { val = Integer.parseInt(tagValue); } catch (NumberFormatException localNumberFormatException1) {
          }
          SmartFoxServer.setMAX_ROOM_VARS(val); break label1402:
        }
        if (tagName.equalsIgnoreCase("MaxUserVars"))
        {
          val = -1;
          try { val = Integer.parseInt(tagValue); } catch (NumberFormatException localNumberFormatException2) {
          }
          SmartFoxServer.setMAX_USER_VARS(val); break label1402:
        }
        if (tagName.equalsIgnoreCase("MaxIncomingQueue"))
        {
          val = -1;
          try { val = Integer.parseInt(tagValue); } catch (NumberFormatException localNumberFormatException3) {
          }
          if (val <= 0) break label1402;
          SmartFoxServer.setMAX_INCOMING_QUEUE(val); break label1402:
        }

        if (tagName.equalsIgnoreCase("WebServer"))
        {
          parse_webServer(option); break label1402:
        }

        if (tagName.equalsIgnoreCase("AutoReloadExtensions"))
        {
          if (!(tagValue.equalsIgnoreCase("true"))) break label1402;
          ConfigData.EXTENSIONS_AUTORELOAD = true; break label1402:
        }

        if (tagName.equalsIgnoreCase("EnableZoneInfo"))
        {
          if (!(tagValue.equalsIgnoreCase("true"))) break label1402;
          ConfigData.ALLOW_ZONE_INFO = true; break label1402:
        }

        if (tagName.equalsIgnoreCase("DeadChannelsPolicy"))
        {
          if (tagValue.equalsIgnoreCase("strict")) {
            ConfigData.DEAD_CHANNELS_POLICY = 1; break label1402:
          }

          ConfigData.DEAD_CHANNELS_POLICY = 0; break label1402:
        }

        if (tagName.equalsIgnoreCase("DisconnectIdleSpectators"))
        {
          if (!(tagValue.equalsIgnoreCase("true"))) break label1402;
          ConfigData.DISCONNECT_IDLE_SPECTATORS = true; break label1402:
        }

        if (tagName.equalsIgnoreCase("RawProtocolSeparator"))
        {
          String separator = tagValue.substring(0, 1);

          if ((separator.equals("<")) || (separator.equals("{")))
            throw new ConfigurationException("The RawProtocolSeparator is invalid. The following characters are forbidden: '<' and '{'");

          ConfigData.STR_DELIMITER = separator; break label1402:
        }

        if (tagName.equalsIgnoreCase("PM_MissingRecipientAlert"))
        {
          if (!(tagValue.equalsIgnoreCase("true"))) break label1402;
          ConfigData.PM_MISSING_RECIPIENT_ALERT = true; break label1402:
        }

        if (tagName.equalsIgnoreCase("ClientMessagQueue"))
        {
          Enumeration children = option.enumerateChildren();

          while (children.hasMoreElements())
          {
            IXMLElement item = (IXMLElement)children.nextElement();

            String nodeName = item.getName();
            String nodeValue = item.getContent();

            if (nodeName.equalsIgnoreCase("QueueSize")) {
              ConfigData.MAX_CHANNEL_QUEUE = Integer.parseInt(nodeValue);
            }
            else if (nodeName.equalsIgnoreCase("MaxAllowedDroppedPackets"))
              ConfigData.MAX_DROPPED_PACKETS = Integer.parseInt(nodeValue);
          }
          break label1402:
        }

        if (tagName.equalsIgnoreCase("buddyListAutoSaveInterval"))
        {
          ConfigData.BUDDY_LIST_AUTOSAVE_INTERVAL = Integer.parseInt(tagValue); break label1402:
        }

        if (tagName.equalsIgnoreCase("buddyListPermissionCleanInterval"))
        {
          ConfigData.BUDDY_LIST_PERMISSION_CLEAN_INTERVAL = Integer.parseInt(tagValue); break label1402:
        }

        if (tagName.equalsIgnoreCase("schedulerResolution"))
        {
          ConfigData.SCHEDULER_RESOLUTION = Integer.parseInt(tagValue); break label1402:
        }

        if (tagName.equalsIgnoreCase("DynamicPropertiesClass"))
        {
          ConfigData.dynamicPropertiesClass = tagValue; break label1402:
        }

        if (tagName.equalsIgnoreCase("H2EnginePort"))
        {
          ConfigData.H2_ARGS[ConfigData.H2_PORT_CONFIG] = tagValue;
          ConfigData.H2_CONN_STRING = "jdbc:h2:tcp://localhost:" + tagValue + "/__sfscore__"; break label1402:
        }

        if ((tagName.equalsIgnoreCase("ExternalPolicyFile")) && (tagValue != null) && (tagValue.length() > 0))
        {
          String dataFile = new String(SmartFoxLib.readFileInClassPath(tagValue));

          if ((dataFile == null) || (dataFile.length() <= 0))
            break label1402;
          SmartFoxServer.getInstance().setCrossDomainPolicy(dataFile);
          ConfigData.EXTERNAL_CROSSDOAMIN_POLICY = true; break label1402:
        }

        if ((!(tagName.equalsIgnoreCase("ASCompilerOptimizationLevel"))) || (tagValue == null))
          break label1402;
        label1402: ConfigData.RHINO_OPTIMIZATION_LEVEL = Integer.parseInt(tagValue);
      }
      catch (NumberFormatException nfe)
      {
        throw new NumberFormatException("Invalid value in <" + tagName + "> section");
      }
      catch (NullPointerException npe)
      {
        throw new Exception("Missing parameter in <" + tagName + "> section");
      }
    }
  }

  private void setAdminAddresses(IXMLElement node)
    throws Exception
  {
    Enumeration addrNodes = node.enumerateChildren();

    while (addrNodes.hasMoreElements())
    {
      IXMLElement item = (IXMLElement)addrNodes.nextElement();

      String ipAddr = item.getContent();

      if (!(ipAddr.equals("")))
        SmartFoxServer.getInstance().setAdminAddresses(ipAddr);
    }
  }

  private void parse_Mailer(IXMLElement node)
  {
    Enumeration nodes = node.enumerateChildren();

    while (nodes.hasMoreElements())
    {
      IXMLElement item = (IXMLElement)nodes.nextElement();

      String nodeName = item.getName();
      String nodeValue = item.getContent();

      if (nodeValue == null)
        nodeValue = "";

      if (nodeName.equalsIgnoreCase("MailHost"))
        MailManager.MAIL_HOST = nodeValue;
      else if (nodeName.equalsIgnoreCase("MailUser"))
        MailManager.MAIL_USER = nodeValue;
      else if (nodeName.equalsIgnoreCase("MailPass"))
        MailManager.MAIL_PASS = nodeValue;
      else if (nodeName.equalsIgnoreCase("SmtpPort"))
        MailManager.MAIL_PORT = nodeValue;
      else if (nodeName.equalsIgnoreCase("WorkerThreads"))
        MailManager.MAIL_THREADS = Integer.parseInt(nodeValue);
    }
  }

  private void parse_webServer(IXMLElement node)
  {
    Enumeration nodes = node.enumerateChildren();

    String isActive = node.getAttribute("active", "false");

    ConfigData.USE_WEBSERVER = isActive.equalsIgnoreCase("true");

    while (nodes.hasMoreElements())
    {
      IXMLElement item = (IXMLElement)nodes.nextElement();

      String nodeName = item.getName();
      String nodeValue = item.getContent();

      if (nodeValue == null)
        nodeValue = "";

      if (nodeName.equalsIgnoreCase("CfgFile"))
      {
        ConfigData.WEBSERVER_CFG_PATH = nodeValue;

        if (!(ConfigData.USE_WEBSERVER)) return;
        SmartFoxServer.getInstance().initWebServer();

        return;
      }
    }
  }

  private Level parseLogginLevel(String level)
    throws Exception
  {
    Level logLevel = null;

    if (level.equalsIgnoreCase("OFF"))
    {
      logLevel = Level.OFF;
    }
    else if (level.equalsIgnoreCase("FINEST"))
    {
      logLevel = Level.FINEST;
    }
    else if (level.equalsIgnoreCase("FINER"))
    {
      logLevel = Level.FINER;
    }
    else if (level.equalsIgnoreCase("FINE"))
    {
      logLevel = Level.FINE;
    }
    else if (level.equalsIgnoreCase("INFO"))
    {
      logLevel = Level.INFO;
    }
    else if (level.equalsIgnoreCase("CONFIG"))
    {
      logLevel = Level.CONFIG;
    }
    else if (level.equalsIgnoreCase("WARNING"))
    {
      logLevel = Level.WARNING;
    }
    else if (level.equalsIgnoreCase("SEVERE"))
    {
      logLevel = Level.SEVERE;
    }
    else
      throw new Exception("Bad Logging Level");

    return logLevel;
  }

  private void parse_Zones(IXMLElement node)
    throws Exception
  {
    System.out.println("\n--- [ Zones & Rooms ] -------------------------------------\n");

    Enumeration subNodes = node.enumerateChildren();

    while (subNodes.hasMoreElements())
    {
      SmartFoxServer.getInstance().setupZone((IXMLElement)subNodes.nextElement(), false);
    }
  }

  private void parse_Extensions(IXMLElement node)
  {
    Enumeration subNodes = node.enumerateChildren();

    while (subNodes.hasMoreElements())
    {
      IXMLElement xt = (IXMLElement)subNodes.nextElement();

      if (xt.getName().equalsIgnoreCase("extension"))
      {
        String xtName = xt.getAttribute("name", "");
        String zoneName = xt.getAttribute("zone", "");

        int xtType = 1;

        if (xt.getAttribute("type", "").equalsIgnoreCase("javascript"))
          xtType = 2;

        String xtClass = xt.getAttribute("className", "");

        SmartFoxServer.getInstance().addExtension(xtName, new ExtensionItem(xtClass, xtType, zoneName));
      }
    }
  }

  private void parse_AllowedDomains(IXMLElement node)
  {
    ConfigData.POLICY_ALLOWED_DOMAINS = new ArrayList();

    Enumeration domains = node.enumerateChildren();

    while (domains.hasMoreElements())
    {
      IXMLElement item = (IXMLElement)domains.nextElement();

      String domain = item.getContent();

      if (!(domain.equals("")))
        ConfigData.POLICY_ALLOWED_DOMAINS.add(domain);
    }
  }

  private void parse_AntiFlood(IXMLElement node)
    throws Exception
  {
    String filter = node.getAttribute("active", "");

    if (filter.equalsIgnoreCase("true"))
    {
      ConfigData.ANTIFLOOD_ACTIVE = true;

      IXMLElement subNode = node.getFirstChildNamed("MinMsgTime");
      ConfigData.ANTIFLOOD_MIN_MSG_TIME = Integer.parseInt(subNode
        .getContent());
      ConfigData.ANTIFLOOD_TOLERANCE = Integer.parseInt(subNode
        .getAttribute("tolerance", ""));

      subNode = node.getFirstChildNamed("MaxRepeatedMessages");
      ConfigData.ANTIFLOOD_MAX_REPEATED = Integer.parseInt(subNode
        .getContent());

      subNode = node.getFirstChildNamed("WarningsBeforeKick");
      ConfigData.ANTIFLOOD_WARNINGS = Integer.parseInt(subNode
        .getContent());

      subNode = node.getFirstChildNamed("WarningMessage");
      ConfigData.ANTIFLOOD_WARNING_MSG = subNode.getContent();

      subNode = node.getFirstChildNamed("KickMessage");
      ConfigData.ANTIFLOOD_KICK_MSG = subNode.getContent();

      subNode = node.getFirstChildNamed("BanMessage");
      ConfigData.ANTIFLOOD_BAN_MSG = subNode.getContent();

      subNode = node.getFirstChildNamed("BanAfter");
      ConfigData.ANTIFLOOD_BAN_AFTER_KICKS = Integer.parseInt(subNode
        .getContent());
      ConfigData.ANTIFLOOD_BAN_AFTER_TIMESPAN = 
        Integer.parseInt(subNode.getAttribute("timeSpan", "")) * 60 * 
        1000;
    }
    else {
      ConfigData.ANTIFLOOD_ACTIVE = false;
    }
  }

  private void parse_badWordFilter(IXMLElement node)
    throws Exception
  {
    String filter = node.getAttribute("active", "");

    if (filter.equalsIgnoreCase("true"))
    {
      ConfigData.BADWORDS_ACTIVE = true;

      IXMLElement subNode = node.getFirstChildNamed("FilterMode");

      if (subNode.getContent().equalsIgnoreCase("filter"))
        ConfigData.BADWORDS_FILTER_MODE = 0;
      else if (subNode.getContent().equalsIgnoreCase("remove"))
        ConfigData.BADWORDS_FILTER_MODE = 1;
      else {
        throw new ConfigurationException(
          "Unrecognized BadWordFilter Mode, available modes are: FILTER, REMOVE");
      }

      subNode = node.getFirstChildNamed("StripCharacters");
      if (subNode != null) {
        BadWordsFilter.punctuation = subNode.getContent();
      }

      subNode = node.getFirstChildNamed("Warnings");
      if (subNode.getContent().equalsIgnoreCase("true"))
        ConfigData.BADWORDS_WARNINGS = true;
      else {
        ConfigData.BADWORDS_WARNINGS = false;
      }

      subNode = node.getFirstChildNamed("FilterRoomNames");
      if (subNode.getContent().equalsIgnoreCase("true"))
        ConfigData.BADWORDS_ROOM_NAMES = true;
      else {
        ConfigData.BADWORDS_ROOM_NAMES = false;
      }

      subNode = node.getFirstChildNamed("FilterUserNames");
      if (subNode.getContent().equalsIgnoreCase("true"))
        ConfigData.BADWORDS_USER_NAMES = true;
      else
        ConfigData.BADWORDS_USER_NAMES = false;

      subNode = node.getFirstChildNamed("Warnings");
      if (subNode.getContent().equalsIgnoreCase("true"))
        ConfigData.BADWORDS_WARNINGS = true;
      else {
        ConfigData.BADWORDS_WARNINGS = false;
      }

      subNode = node.getFirstChildNamed("WarningsBeforeKick");
      ConfigData.BADWORDS_WARNINGS_BEFORE_KICK = 
        Integer.parseInt(subNode.getContent());

      subNode = node.getFirstChildNamed("WarningMessage");
      ConfigData.BADWORDS_WARNING_MSG = subNode.getContent();

      subNode = node.getFirstChildNamed("KickMessage");
      ConfigData.BADWORDS_KICK_MSG = subNode.getContent();

      subNode = node.getFirstChildNamed("BanMessage");
      ConfigData.BADWORDS_BAN_MSG = subNode.getContent();

      subNode = node.getFirstChildNamed("BanAfter");
      ConfigData.BADWORDS_BAN_AFTER_KICKS = Integer.parseInt(subNode
        .getContent());
      ConfigData.BADWORDS_BAN_AFTER_TIMESPAN = 
        Integer.parseInt(subNode.getAttribute("timeSpan", "")) * 60 * 
        1000;

      subNode = node.getFirstChildNamed("BadWordsList");

      if (subNode != null)
      {
        Enumeration words = subNode.enumerateChildren();

        LinkedList tempWordList = new LinkedList();

        while (words.hasMoreElements())
        {
          IXMLElement word = (IXMLElement)words.nextElement();

          if (word.getName().equalsIgnoreCase("badWord"))
          {
            String item = word.getContent();
            tempWordList.add(item);
          }
        }

        BadWordsFilter.getInstance().initBadWordsList(tempWordList);
      }
      else {
        throw new ConfigurationException("The <BadWordList></BadWordList> section is missing!");
      }
    }
  }

  public void reload_Zones()
    throws Exception
  {
    readCfgFile();

    Enumeration cfgNodes = this.sfsXmlConfig.enumerateChildren();

    while (cfgNodes.hasMoreElements())
    {
      IXMLElement cfgNode = (IXMLElement)cfgNodes.nextElement();

      if (cfgNode.getName().equalsIgnoreCase("Zones"))
        this.zonesXml = cfgNode;

    }

    if (this.zonesXml != null)
    {
      Enumeration subNodes = this.zonesXml.enumerateChildren();

      while (subNodes.hasMoreElements())
      {
        IXMLElement zNode = (IXMLElement)subNodes.nextElement();
        String zName = zNode.getAttribute("name", "");

        SmartFoxServer.getInstance().setupZone(zNode, true);
      }
    }
    else {
      throw new ConfigurationException("Zones XML nodes not available!");
    }
  }
}