package it.gotoandplay.smartfoxserver.lib;

import it.gotoandplay.smartfoxserver.EventWriter;
import it.gotoandplay.smartfoxserver.SmartFoxServer;
import it.gotoandplay.smartfoxserver.config.ConfigData;
import it.gotoandplay.smartfoxserver.data.User;
import java.net.Socket;
import java.nio.channels.SocketChannel;
import java.util.HashMap;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;
import java.util.TimerTask;
import java.util.logging.Logger;

public class ConnectionCleanerTask extends TimerTask
{
  private long maxIdleTime;
  private LinkedList channels;
  private LinkedList killableChannels;
  private Iterator it;
  private SmartFoxServer sf;
  private SocketChannel ch;
  private Map NonLoggedUsers;
  private Object o;
  private User u;
  private long lastTime;
  private long currTime;

  public ConnectionCleanerTask(int maxIdleTime)
  {
    this.sf = SmartFoxServer.getInstance();
    this.maxIdleTime = (maxIdleTime * 1000L);
    this.NonLoggedUsers = new HashMap();
  }

  public void run()
  {
    if (ConfigData.DEAD_CHANNELS_POLICY > 0)
      this.sf.getEW().cleanFailedChannels();

    this.channels = this.sf.getChannels();
    this.killableChannels = new LinkedList();

    synchronized (this.channels)
    {
      this.it = this.channels.iterator();

      this.currTime = System.currentTimeMillis();

      label390: while (this.it.hasNext())
      {
        try
        {
          this.ch = ((SocketChannel)this.it.next());

          if ((this.ch == null) || (!(this.ch.isConnected())))
            break label390;
          this.u = this.sf.getUserByChannel(this.ch);

          if (this.u == null)
          {
            this.o = this.NonLoggedUsers.get(this.ch);

            if (this.o != null)
            {
              this.lastTime = ((Long)this.o).longValue();

              if (this.currTime - this.lastTime <= this.maxIdleTime)
                break label390;

              addKillable(this.ch);

              this.NonLoggedUsers.remove(this.ch); break label390:
            }

            this.NonLoggedUsers.put(this.ch, new Long(this.currTime)); break label390:
          }

          this.lastTime = this.u.getLastMessageTime();

          if ((this.currTime - this.lastTime > this.maxIdleTime) && (!(this.u.isAdmin())))
          {
            if ((!(ConfigData.DISCONNECT_IDLE_SPECTATORS)) && (this.u.isSpectator()))
              break label390:

            SmartFoxServer.log.info("Disconneting idle user: " + this.u.getName());

            addKillable(this.ch);
          }

          this.o = this.NonLoggedUsers.get(this.ch);

          if (this.o == null)
            break label390;
          this.NonLoggedUsers.remove(this.ch);
        }
        catch (Exception e)
        {
          SmartFoxServer.log.warning("Connection cleaner error: " + e);
          e.printStackTrace();
        }
      }
    }

    killEmAll();
  }

  private void addKillable(SocketChannel ch)
  {
    this.killableChannels.add(ch);
  }

  private void killEmAll()
  {
    if (this.killableChannels.size() > 0)
    {
      Iterator i = this.killableChannels.iterator();

      while (i.hasNext())
      {
        SocketChannel sch = (SocketChannel)i.next();

        if (sch != null)
        {
          if (sch.socket().isClosed())
            SmartFoxServer.log.fine("IP: " + sch.socket().getInetAddress() + " was found idle and was disconnected.");

          this.sf.lostConnection(sch);
        }

      }

    }

    for (Iterator it = this.NonLoggedUsers.entrySet().iterator(); it.hasNext(); )
    {
      Map.Entry item = (Map.Entry)it.next();

      SocketChannel sc = (SocketChannel)item.getKey();

      if (sc.isConnected())
        break label146;
      label146: it.remove();
    }
  }
}