package it.gotoandplay.smartfoxserver.lib;

import EDU.oswego.cs.dl.util.concurrent.LinkedQueue;
import EDU.oswego.cs.dl.util.concurrent.PooledExecutor;
import it.gotoandplay.smartfoxserver.SmartFoxServer;
import java.util.Date;
import java.util.Properties;
import java.util.logging.Logger;
import javax.mail.Address;
import javax.mail.Message.RecipientType;
import javax.mail.Session;
import javax.mail.Transport;
import javax.mail.internet.InternetAddress;
import javax.mail.internet.MimeMessage;

public class MailManager
{
  public static String MAIL_HOST = "";
  public static String MAIL_USER = "";
  public static String MAIL_PASS = "";
  public static String MAIL_PORT = "25";
  public static int MAIL_THREADS = 1;
  public static final String MAIL_PROTOCOL = "smtp";
  private static MailManager instance;
  private PooledExecutor executor;
  private Session session;

  private MailManager()
  {
    Properties props = new Properties();

    props.setProperty("mail.transport.protocol", "smtp");
    props.setProperty("mail.host", MAIL_HOST);
    props.setProperty("mail.user", MAIL_USER);
    props.setProperty("mail.password", MAIL_PASS);
    props.setProperty("mail.smtp.port", MAIL_PORT);
    props.setProperty("mail.smtp.auth", "true");

    this.session = Session.getDefaultInstance(props, null);

    this.executor = new PooledExecutor(new LinkedQueue());
    this.executor.setKeepAliveTime(-1L);
    this.executor.createThreads(MAIL_THREADS);
  }

  public static MailManager getInstance()
  {
    if (instance == null)
      instance = new MailManager();

    return instance;
  }

  public boolean sendMail(String from, String to, String subject, String mailMsg)
  {
    boolean success = false;
    try
    {
      MimeMessage message = new MimeMessage(this.session);

      message.setContent(mailMsg, "text/html");

      message.setSubject(subject);

      Address fromAddress = new InternetAddress(from);
      Address toAddress = new InternetAddress(to);

      message.setSentDate(new Date());
      message.setFrom(fromAddress);
      message.addRecipient(Message.RecipientType.TO, toAddress);

      this.executor.execute(
        new Runnable(this, message)
      {
        public void run()
        {
          Transport smtp;
          try
          {
            smtp = MailManager.access$0(this.this$0).getTransport("smtp");
            smtp.connect(MailManager.MAIL_HOST, new Integer(MailManager.MAIL_PORT).intValue(), MailManager.MAIL_USER, MailManager.MAIL_PASS);

            smtp.sendMessage(this.val$message, this.val$message.getAllRecipients());
            smtp.close();

            SmartFoxServer.log.fine("Email sent to: " + this.val$message.getAllRecipients()[0]);
          }
          catch (Exception e)
          {
            SmartFoxServer.log.warning("Exception during SMTP transaction: " + e.toString());
          }

        }

      });
      success = true;
    }
    catch (Exception e)
    {
      SmartFoxServer.log.warning("Problems sending email to: " + to + ", subject: " + subject + " -- Error: " + e.toString());
    }

    return success;
  }
}