package it.gotoandplay.smartfoxserver.lib;

import it.gotoandplay.smartfoxserver.SmartFoxServer;
import it.gotoandplay.smartfoxserver.exceptions.BuddyListException;
import java.io.BufferedInputStream;
import java.io.ByteArrayOutputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.InputStream;
import java.io.PrintStream;
import java.net.URL;
import java.net.URLDecoder;
import java.util.ArrayList;
import java.util.List;
import java.util.logging.Logger;

public class SmartFoxLib
{
  static final String avoidChars = "<>'\"&/\\^";
  static final String utfSignature = new String(new char[] { 239, 187, 191 });
  private static final byte UNIX_FEED = 10;
  private static final byte MAC_FEED = 13;

  public static String cleanString(String str)
  {
    StringBuffer result = new StringBuffer();

    for (int i = 0; i < str.length(); ++i)
    {
      char c = str.charAt(i);

      if ((c >= ' ') && ("<>'\"&/\\^".indexOf(c) == -1))
      {
        result.append(c);
      }
    }

    return result.toString();
  }

  public static String cleanUpString(String source, String avoidCh, int maxlen)
  {
    String s;
    if (avoidCh == null) {
      avoidCh = "<>'\"&/\\^";
    }

    StringBuffer result = new StringBuffer();

    for (int i = 0; i < source.length(); ++i)
    {
      char c = source.charAt(i);

      if ((c >= ' ') && (avoidCh.indexOf(c) == -1))
      {
        result.append(c);
      }

    }

    if (result.length() > maxlen)
      s = result.substring(0, maxlen);
    else
      s = result.toString();

    return s;
  }

  public static String escapeQuotes(String str)
  {
    StringBuffer result = new StringBuffer();

    for (int i = 0; i < str.length(); ++i)
    {
      char c = str.charAt(i);

      if (c == '\'')
      {
        result.append('\'').append('\'');
      }
      else if (c == '"')
      {
        result.append('"').append('"');
      }
      else
      {
        result.append(c);
      }
    }
    return result.toString();
  }

  public static InputStream readStreamInClassPath(String fileName) throws IOException
  {
    return ClassLoader.getSystemResourceAsStream(fileName);
  }

  public static byte[] readFileInClassPath(String fileName) throws IOException
  {
    byte[] byteData = (byte[])null;

    InputStream is = ClassLoader.getSystemResourceAsStream(fileName);

    if (is == null)
      throw new FileNotFoundException("File: '" + fileName + "' was not found in the classpath.");

    BufferedInputStream bis = null;
    ByteArrayOutputStream baos = null;
    try
    {
      int read;
      baos = new ByteArrayOutputStream();
      bis = new BufferedInputStream(is);

      label147: label115: while ((read = bis.read()) != -1)
        baos.write(read);
    }
    finally
    {
      try
      {
        if (bis != null)
          bis.close();
        if (baos == null) break label115;
        baos.close();
      }
      catch (Exception localException1) {
      }
      byteData = baos.toByteArray();
    }

    return byteData;
  }

  public static String getFullPath(String resource)
  {
    URL u = ClassLoader.getSystemResource(resource);

    String res = "";
    try
    {
      if (u == null)
        break label83;
      res = URLDecoder.decode(u.getFile(), "UTF-8");

      i = res.indexOf("SmartFoxServer.jar");

      if (i == -1)
        break label83;
      String newRes = res.substring(0, i - 1) + res.substring(i + 19, res.length());
      res = newRes.substring(5);
    }
    catch (java.io.UnsupportedEncodingException i)
    {
    }

    label83: return res;
  }

  public static boolean deleteFile(String fileName)
  {
    File f = new File(fileName);
    System.out.println(f.getAbsolutePath());
    return f.delete();
  }

  public static String fixLineFeeds(byte[] data)
  {
    byte[] temp = (byte[])null;

    if ((data[0] == -17) && (data[1] == -69) && (data[2] == -65))
    {
      System.out.println("UTF8 file!");
      temp = new byte[data.length - 3];

      System.arraycopy(data, 3, temp, 0, temp.length);

      data = temp;
    }

    byte[] fixed = new byte[data.length];
    int p = 0;

    for (int i = 0; i < data.length; ++i)
    {
      byte b = data[i];

      if (b != 13)
      {
        fixed[p] = b;
        ++p;
      }
      else
      {
        fixed[p] = 10;
        ++p;

        if ((i < data.length - 1) && 
          (data[(i + 1)] == 10))
          ++i;

      }

    }

    String s = new String(fixed);

    if (p < data.length)
    {
      s = s.substring(0, p);
    }

    return s;
  }

  public static int countCodeLines(String text)
  {
    int lines = 0;
    int pos = 0;
    while (true)
    {
      pos = text.indexOf("\n", pos);

      if (pos == -1) {
        break;
      }

      ++lines;
      ++pos;
    }

    return lines;
  }

  public static List getListOfLines(String text)
  {
    List l = new ArrayList();

    int pos = 0;
    int lastPos = 0;
    while (true)
    {
      pos = text.indexOf("\n", lastPos);

      if (pos == -1) {
        break;
      }

      l.add(text.substring(lastPos, pos + 1));
      lastPos = pos + 1;
    }

    l.add(text.substring(lastPos));

    return l;
  }

  public static void checkBuddyListFolders(String buddyListPath)
    throws BuddyListException
  {
    File f = new File(buddyListPath);

    if (!(f.isDirectory()))
    {
      if (f.mkdirs())
        SmartFoxServer.log.info("Directory '" + buddyListPath + "' created");
      else
        throw new BuddyListException("Failed creating directory '" + buddyListPath + "'");
    }
  }

  public static String getHexFileName(String name)
  {
    StringBuilder sb = new StringBuilder();

    char[] c = name.toCharArray();

    for (int i = 0; i < c.length; ++i)
    {
      sb.append(Integer.toHexString(c[i]));
    }

    return sb.toString();
  }

  public static byte[] readFile(String fileName)
    throws IOException
  {
    byte[] bytes = (byte[])null;

    BufferedInputStream bis = null;
    try
    {
      File file = new File(fileName);
      bytes = new byte[(int)file.length()];
      bis = new BufferedInputStream(new FileInputStream(file));
      label64: bis.read(bytes);
    }
    finally
    {
      try
      {
        if (bis == null) break label64;
        bis.close();
      }
      catch (Exception localException1) {
      }
    }
    label80: return bytes;
  }

  public static void dumpActiveThreads()
  {
    Thread[] threads = new Thread[Thread.activeCount()];
    Thread.enumerate(threads);

    System.out.println("--->> Active Threads <<-----------");

    for (Thread t : threads)
    {
      System.out.println(t.getName() + " ( daemon = " + t.isDaemon() + ", alive = " + t.isAlive() + " )");
    }
  }
}