package it.gotoandplay.smartfoxserver.lib;

import it.gotoandplay.smartfoxserver.SmartFoxServer;
import it.gotoandplay.smartfoxserver.config.ConfigData;
import java.io.PrintStream;
import java.util.logging.Logger;

public abstract class Wrap
  implements Runnable, EventHandler
{
  Logger log = SmartFoxServer.log;
  protected EventQueue eventQueue;
  protected boolean running = false;
  private Thread[] workers;
  private int spareCount;
  private Object countLock = new Object();
  private String shortname;

  public final void initWrap(int numWorkers)
  {
    this.shortname = super.getClass().getName().substring(super.getClass().getName().lastIndexOf(".") + 1);

    this.eventQueue = new EventQueue(this.shortname + "-in");

    this.workers = new Thread[numWorkers];
    for (int i = 0; i < numWorkers; ++i)
    {
      this.workers[i] = new Thread(this, this.shortname + "-" + (i + 1));
      this.workers[i].setDaemon(true);
      this.workers[i].start();
    }
  }

  public void shutdown()
  {
    this.running = false;
    if (this.workers != null)
    {
      for (int i = 0; i < this.workers.length; ++i)
      {
        this.workers[i].interrupt();
      }
    }
  }

  public void handleEvent(ServerEvent event, boolean isAdmin)
  {
    if (isAdmin) {
      this.eventQueue.adminEnQueue(event);
    }
    else if (this.eventQueue.size() < SmartFoxServer.getMAX_INCOMING_QUEUE())
      this.eventQueue.enQueue(event);
    else
      ConfigData.inComingDroppedMessages += 1L;
  }

  public void handleEvent(ServerEvent event)
  {
    handleEvent(event, false);
  }

  public int getQueueSize()
  {
    return this.eventQueue.size();
  }

  public void run()
  {
    this.running = true;
    label37: while (this.running)
    {
      try
      {
        ServerEvent event;
        if ((event = this.eventQueue.deQueue()) == null) break label37;
        processEvent(event);
      }
      catch (InterruptedException e)
      {
        System.out.println("Problems in wrap!");
      }
    }
  }

  protected abstract void processEvent(ServerEvent paramServerEvent);
}