package it.gotoandplay.smartfoxserver.db;

import it.gotoandplay.smartfoxserver.SmartFoxServer;
import java.io.PrintStream;
import java.sql.Connection;
import java.sql.DriverManager;
import java.sql.ResultSet;
import java.sql.ResultSetMetaData;
import java.sql.SQLException;
import java.util.ArrayList;
import java.util.logging.Logger;
import org.apache.commons.dbcp.ConnectionFactory;
import org.apache.commons.dbcp.DriverManagerConnectionFactory;
import org.apache.commons.dbcp.PoolableConnectionFactory;
import org.apache.commons.dbcp.PoolingDriver;
import org.apache.commons.pool.ObjectPool;
import org.apache.commons.pool.impl.GenericObjectPool;
import org.apache.commons.pool.impl.GenericObjectPool.Config;

public class DbManager
{
  String driverName;
  String connStr;
  String usrName;
  String pword;
  String zoneName;
  String exhaustedAction;
  int maxActive;
  int maxIdle;
  int blockTime;

  public DbManager(String driverName, String connStr, String usrName, String pword, String zoneName, int maxActive, int maxIdle, String exhaustedAction, int blockTime)
  {
    this.driverName = driverName;
    this.connStr = connStr;
    this.usrName = usrName;
    this.pword = pword;
    this.zoneName = zoneName;
    this.exhaustedAction = exhaustedAction;

    this.blockTime = blockTime;
    this.maxActive = maxActive;
    this.maxIdle = maxIdle;

    init();
  }

  private void init()
  {
    try
    {
      Class.forName(this.driverName);

      setupDriver(this.connStr, this.zoneName);
    }
    catch (ClassNotFoundException e)
    {
      System.out.println("Can't load db driver: " + this.driverName);
    }
    catch (Exception e)
    {
      e.printStackTrace();
    }
  }

  private void setupDriver(String connectURI, String connName)
    throws Exception
  {
    GenericObjectPool.Config conf = new GenericObjectPool.Config();

    conf.maxActive = this.maxActive;
    conf.maxIdle = this.maxIdle;

    conf.testOnBorrow = true;
    conf.testOnReturn = true;

    if (this.exhaustedAction.equals("wait"))
    {
      conf.whenExhaustedAction = 1;
      conf.maxWait = this.blockTime;
    }
    else if (this.exhaustedAction.equals("grow"))
    {
      conf.whenExhaustedAction = 2;
    }
    else
    {
      conf.whenExhaustedAction = 0;
    }

    ObjectPool connectionPool = new GenericObjectPool(null, conf);

    ConnectionFactory connectionFactory = new DriverManagerConnectionFactory(connectURI, this.usrName, this.pword);

    PoolableConnectionFactory poolableConnectionFactory = new PoolableConnectionFactory(connectionFactory, connectionPool, null, null, false, true);

    PoolingDriver driver = new PoolingDriver();

    driver.registerPool(connName, connectionPool);
  }

  public Connection getConnection()
  {
    Connection conn = null;
    try
    {
      conn = DriverManager.getConnection("jdbc:apache:commons:dbcp:" + this.zoneName);
    }
    catch (SQLException sqle)
    {
      SmartFoxServer.log.warning("Could not retrieve a database connection: " + sqle);
      sqle.printStackTrace();
    }

    return conn;
  }

  public ArrayList executeQuery(String cmd)
  {
    return executeQuery(cmd, 1); } 
  // ERROR //
  public ArrayList executeQuery(String cmd, int keyType) { // Byte code:
    //   0: aconst_null
    //   1: astore_3
    //   2: aconst_null
    //   3: astore 4
    //   5: aconst_null
    //   6: astore 5
    //   8: new 63	java/lang/StringBuilder
    //   11: dup
    //   12: ldc 160
    //   14: invokespecial 67	java/lang/StringBuilder:<init>	(Ljava/lang/String;)V
    //   17: aload_0
    //   18: getfield 30	it/gotoandplay/smartfoxserver/db/DbManager:zoneName	Ljava/lang/String;
    //   21: invokevirtual 70	java/lang/StringBuilder:append	(Ljava/lang/String;)Ljava/lang/StringBuilder;
    //   24: invokevirtual 74	java/lang/StringBuilder:toString	()Ljava/lang/String;
    //   27: invokestatic 162	java/sql/DriverManager:getConnection	(Ljava/lang/String;)Ljava/sql/Connection;
    //   30: astore_3
    //   31: aload_3
    //   32: invokeinterface 196 1 0
    //   37: astore 4
    //   39: aconst_null
    //   40: astore 6
    //   42: aload 4
    //   44: aload_1
    //   45: invokeinterface 202 2 0
    //   50: astore 6
    //   52: aload_0
    //   53: aload 6
    //   55: iload_2
    //   56: invokespecial 207	it/gotoandplay/smartfoxserver/db/DbManager:buildListFromResult	(Ljava/sql/ResultSet;I)Ljava/util/ArrayList;
    //   59: astore 5
    //   61: goto +110 -> 171
    //   64: astore 7
    //   66: getstatic 167	it/gotoandplay/smartfoxserver/SmartFoxServer:log	Ljava/util/logging/Logger;
    //   69: new 63	java/lang/StringBuilder
    //   72: dup
    //   73: ldc 211
    //   75: invokespecial 67	java/lang/StringBuilder:<init>	(Ljava/lang/String;)V
    //   78: aload_1
    //   79: invokevirtual 70	java/lang/StringBuilder:append	(Ljava/lang/String;)Ljava/lang/StringBuilder;
    //   82: invokevirtual 74	java/lang/StringBuilder:toString	()Ljava/lang/String;
    //   85: invokevirtual 213	java/util/logging/Logger:severe	(Ljava/lang/String;)V
    //   88: aload 7
    //   90: invokevirtual 183	java/sql/SQLException:printStackTrace	()V
    //   93: aload 6
    //   95: invokeinterface 216 1 0
    //   100: goto +5 -> 105
    //   103: astore 9
    //   105: aload 4
    //   107: invokeinterface 221 1 0
    //   112: goto +5 -> 117
    //   115: astore 9
    //   117: aload_3
    //   118: invokeinterface 222 1 0
    //   123: goto +111 -> 234
    //   126: astore 9
    //   128: goto +106 -> 234
    //   131: astore 8
    //   133: aload 6
    //   135: invokeinterface 216 1 0
    //   140: goto +5 -> 145
    //   143: astore 9
    //   145: aload 4
    //   147: invokeinterface 221 1 0
    //   152: goto +5 -> 157
    //   155: astore 9
    //   157: aload_3
    //   158: invokeinterface 222 1 0
    //   163: goto +5 -> 168
    //   166: astore 9
    //   168: aload 8
    //   170: athrow
    //   171: aload 6
    //   173: invokeinterface 216 1 0
    //   178: goto +5 -> 183
    //   181: astore 9
    //   183: aload 4
    //   185: invokeinterface 221 1 0
    //   190: goto +5 -> 195
    //   193: astore 9
    //   195: aload_3
    //   196: invokeinterface 222 1 0
    //   201: goto +33 -> 234
    //   204: astore 9
    //   206: goto +28 -> 234
    //   209: astore 6
    //   211: getstatic 167	it/gotoandplay/smartfoxserver/SmartFoxServer:log	Ljava/util/logging/Logger;
    //   214: new 63	java/lang/StringBuilder
    //   217: dup
    //   218: ldc 223
    //   220: invokespecial 67	java/lang/StringBuilder:<init>	(Ljava/lang/String;)V
    //   223: aload 6
    //   225: invokevirtual 175	java/lang/StringBuilder:append	(Ljava/lang/Object;)Ljava/lang/StringBuilder;
    //   228: invokevirtual 74	java/lang/StringBuilder:toString	()Ljava/lang/String;
    //   231: invokevirtual 213	java/util/logging/Logger:severe	(Ljava/lang/String;)V
    //   234: aload 5
    //   236: areturn
    //
    // Exception table:
    //   from	to	target	type
    //   42	61	64	java/sql/SQLException
    //   93	100	103	Exception
    //   105	112	115	Exception
    //   117	123	126	Exception
    //   42	93	131	finally
    //   133	140	143	Exception
    //   145	152	155	Exception
    //   157	163	166	Exception
    //   171	178	181	Exception
    //   183	190	193	Exception
    //   195	201	204	Exception
    //   8	206	209	java/sql/SQLException } 
  // ERROR //
  public boolean executeCommand(String cmd) { // Byte code:
    //   0: aconst_null
    //   1: astore_2
    //   2: aconst_null
    //   3: astore_3
    //   4: aconst_null
    //   5: astore 4
    //   7: iconst_0
    //   8: istore 5
    //   10: new 63	java/lang/StringBuilder
    //   13: dup
    //   14: ldc 160
    //   16: invokespecial 67	java/lang/StringBuilder:<init>	(Ljava/lang/String;)V
    //   19: aload_0
    //   20: getfield 30	it/gotoandplay/smartfoxserver/db/DbManager:zoneName	Ljava/lang/String;
    //   23: invokevirtual 70	java/lang/StringBuilder:append	(Ljava/lang/String;)Ljava/lang/StringBuilder;
    //   26: invokevirtual 74	java/lang/StringBuilder:toString	()Ljava/lang/String;
    //   29: invokestatic 162	java/sql/DriverManager:getConnection	(Ljava/lang/String;)Ljava/sql/Connection;
    //   32: astore_2
    //   33: aload_2
    //   34: invokeinterface 196 1 0
    //   39: astore_3
    //   40: aload_3
    //   41: aload_1
    //   42: invokeinterface 234 2 0
    //   47: pop
    //   48: iconst_1
    //   49: istore 5
    //   51: goto +84 -> 135
    //   54: astore 6
    //   56: getstatic 167	it/gotoandplay/smartfoxserver/SmartFoxServer:log	Ljava/util/logging/Logger;
    //   59: new 63	java/lang/StringBuilder
    //   62: dup
    //   63: ldc 238
    //   65: invokespecial 67	java/lang/StringBuilder:<init>	(Ljava/lang/String;)V
    //   68: aload_1
    //   69: invokevirtual 70	java/lang/StringBuilder:append	(Ljava/lang/String;)Ljava/lang/StringBuilder;
    //   72: invokevirtual 74	java/lang/StringBuilder:toString	()Ljava/lang/String;
    //   75: invokevirtual 213	java/util/logging/Logger:severe	(Ljava/lang/String;)V
    //   78: aload 6
    //   80: invokevirtual 183	java/sql/SQLException:printStackTrace	()V
    //   83: aload_3
    //   84: invokeinterface 221 1 0
    //   89: goto +5 -> 94
    //   92: astore 8
    //   94: aload_2
    //   95: invokeinterface 222 1 0
    //   100: goto +85 -> 185
    //   103: astore 8
    //   105: goto +80 -> 185
    //   108: astore 7
    //   110: aload_3
    //   111: invokeinterface 221 1 0
    //   116: goto +5 -> 121
    //   119: astore 8
    //   121: aload_2
    //   122: invokeinterface 222 1 0
    //   127: goto +5 -> 132
    //   130: astore 8
    //   132: aload 7
    //   134: athrow
    //   135: aload_3
    //   136: invokeinterface 221 1 0
    //   141: goto +5 -> 146
    //   144: astore 8
    //   146: aload_2
    //   147: invokeinterface 222 1 0
    //   152: goto +33 -> 185
    //   155: astore 8
    //   157: goto +28 -> 185
    //   160: astore 6
    //   162: getstatic 167	it/gotoandplay/smartfoxserver/SmartFoxServer:log	Ljava/util/logging/Logger;
    //   165: new 63	java/lang/StringBuilder
    //   168: dup
    //   169: ldc 223
    //   171: invokespecial 67	java/lang/StringBuilder:<init>	(Ljava/lang/String;)V
    //   174: aload 6
    //   176: invokevirtual 175	java/lang/StringBuilder:append	(Ljava/lang/Object;)Ljava/lang/StringBuilder;
    //   179: invokevirtual 74	java/lang/StringBuilder:toString	()Ljava/lang/String;
    //   182: invokevirtual 213	java/util/logging/Logger:severe	(Ljava/lang/String;)V
    //   185: iload 5
    //   187: ireturn
    //
    // Exception table:
    //   from	to	target	type
    //   40	51	54	java/sql/SQLException
    //   83	89	92	Exception
    //   94	100	103	Exception
    //   40	83	108	finally
    //   110	116	119	Exception
    //   121	127	130	Exception
    //   135	141	144	Exception
    //   146	152	155	Exception
    //   10	157	160	java/sql/SQLException } 
  private ArrayList buildListFromResult(ResultSet rs, int keyType) throws SQLException { ArrayList list = new ArrayList();

    ResultSetMetaData metaData = rs.getMetaData();
    int cols = metaData.getColumnCount();

    while (rs.next())
    {
      int c = 1;

      DataRow row = new DataRow(keyType);
      row.addItem("");

      while (c <= cols)
      {
        String colName = metaData.getColumnName(c);

        if (keyType == 0)
          row.addItem(rs.getString(c));
        else
          row.addItem(colName, rs.getString(c));

        ++c;
      }

      list.add(row);
    }

    return list;
  }

  private void printDriverStats()
    throws Exception
  {
    PoolingDriver driver = (PoolingDriver)DriverManager.getDriver("jdbc:apache:commons:dbcp:");
    ObjectPool connectionPool = driver.getConnectionPool(this.zoneName);

    System.out.println("NumActive: " + connectionPool.getNumActive());
    System.out.println("NumIdle: " + connectionPool.getNumIdle());
  }

  private void shutdownDriver()
    throws Exception
  {
    PoolingDriver driver = (PoolingDriver)DriverManager.getDriver("jdbc:apache:commons:dbcp:");
    driver.closePool(this.zoneName);
  }

  public void shutDown()
  {
    try
    {
      shutdownDriver();
    }
    catch (Exception e)
    {
      e.printStackTrace();
    }

    SmartFoxServer.log.fine("Database connection [ " + this.zoneName + " ] shutting down");
  }
}