package it.gotoandplay.smartfoxserver.util.scheduling;

import it.gotoandplay.smartfoxserver.lib.IService;
import java.io.PrintStream;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.concurrent.Executor;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;

public class Scheduler
  implements IService, Runnable
{
  private long SLEEP_TIME;
  private Executor taskExecutor;
  private LinkedList<ScheduledTask> taskList;
  private LinkedList<ScheduledTask> addList;
  private boolean running;

  public Scheduler()
  {
    this.SLEEP_TIME = 750L;

    this.running = false;

    init(null);
  }

  public Scheduler(long interval)
  {
    this.SLEEP_TIME = interval;
  }

  public void init(Object param)
  {
    this.taskList = new LinkedList();
    this.addList = new LinkedList();
  }

  public void destroy(Object param)
  {
    stopService();
  }

  public void startService()
  {
    this.running = true;
    this.taskExecutor = Executors.newSingleThreadExecutor();
    this.taskExecutor.execute(this);
  }

  public void stopService()
  {
    this.running = false;
    ((ExecutorService)this.taskExecutor).shutdown();

    this.taskExecutor = null;
  }

  public void run()
  {
    while (this.running)
    {
      try
      {
        executeTasks();
        Thread.sleep(this.SLEEP_TIME);
      }
      catch (InterruptedException localInterruptedException)
      {
      }
      catch (Exception e)
      {
        e.printStackTrace();
      }
    }
  }

  public void addScheduledTask(Task task, int interval, boolean loop, ITaskHandler callback)
  {
    this.addList.add(new ScheduledTask(this, task, interval, loop, callback));
  }

  /**
   * @deprecated
   */
  public void removeScheduledTask(Task task)
  {
    synchronized (this.taskList)
    {
      for (Iterator it = this.taskList.iterator(); it.hasNext(); )
      {
        ScheduledTask t = (ScheduledTask)it.next();

        if (t.getTask() != task)
          break label47;
        it.remove();
        label47: break;
      }
    }
  }

  private void executeTasks()
  {
    long now = System.currentTimeMillis();

    synchronized (this.taskList)
    {
      for (??? = this.taskList.iterator(); ???.hasNext(); )
      {
        ScheduledTask t = (ScheduledTask)???.next();

        if (!(t.task.active))
        {
          ???.remove();
          break label178:
        }

        if (now < t.expiry)
          break label178;

        try
        {
          t.callback.doTask(t.task);
        }
        catch (Exception e)
        {
          System.err.println("Error: " + e);
          System.err.println("Handler: " + t.callback);
          e.printStackTrace();
        }

        if (t.loop)
        {
          t.expiry += t.interval * 1000; break label178:
        }

        label178: ???.remove();
      }

      if (this.addList.size() > 0)
      {
        synchronized (this.taskList)
        {
          this.taskList.addAll(this.addList);
          this.addList.clear();
        }
      }
    }
  }

  private class ScheduledTask
  {
    long expiry;
    int interval;
    boolean loop;
    ITaskHandler callback;
    Task task;

    public ScheduledTask(, Task paramTask, int paramInt, boolean paramBoolean, ITaskHandler paramITaskHandler)
    {
      this.task = paramTask;
      this.interval = paramInt;
      this.expiry = (System.currentTimeMillis() + paramInt * 1000);
      this.callback = paramITaskHandler;
      this.loop = paramBoolean;
    }

    public int getInterval()
    {
      return this.interval;
    }

    public Task getTask()
    {
      return this.task;
    }

    public ITaskHandler getCallback()
    {
      return this.callback;
    }

    public long getExpiry()
    {
      return this.expiry;
    }

    public boolean isLooping()
    {
      return this.loop;
    }
  }
}